@@ -1,7 +1,5 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-//#define DEBUGTAG "AddressSpace"
-
 #include "AddressSpace.h"
 
 #include <limits.h>
@@ -16,26 +14,19 @@
 #include "preload/preload_interface.h"
 
 #include "AutoRemoteSyscalls.h"
-#include "log.h"
+#include "MonitoredSharedMemory.h"
 #include "RecordSession.h"
+#include "RecordTask.h"
 #include "Session.h"
-#include "task.h"
+#include "Task.h"
+#include "log.h"
 
-using namespace rr;
 using namespace std;
 
+namespace rr {
+
 /*static*/ const uint8_t AddressSpace::breakpoint_insn;
 
-void HasTaskSet::insert_task(Task* t) {
-  LOG(debug) << "adding " << t->tid << " to task set " << this;
-  tasks.insert(t);
-}
-
-void HasTaskSet::erase_task(Task* t) {
-  LOG(debug) << "removing " << t->tid << " from task group " << this;
-  tasks.erase(t);
-}
-
 /**
  * Advance *str to skip leading blank characters.
  */
@@ -53,18 +44,14 @@ static const char* trim_leading_blanks(const char* str) {
  */
 class KernelMapIterator {
 public:
-  KernelMapIterator(Task* t) : t(t) {
-    char maps_path[PATH_MAX];
-    sprintf(maps_path, "/proc/%d/maps", t->tid);
-    ASSERT(t, (maps_file = fopen(maps_path, "r"))) << "Failed to open "
-                                                   << maps_path;
-    ++*this;
-  }
+  KernelMapIterator(Task* t) : tid(t->tid) { init(); }
+  KernelMapIterator(pid_t tid) : tid(tid) { init(); }
   ~KernelMapIterator() {
     if (maps_file) {
       fclose(maps_file);
     }
   }
+
   // It's very important to keep in mind that btrfs files can have the wrong
   // device number!
   const KernelMapping& current(string* raw_line = nullptr) {
@@ -77,7 +64,15 @@ public:
   void operator++();
 
 private:
-  Task* t;
+  void init() {
+    char maps_path[PATH_MAX];
+    sprintf(maps_path, "/proc/%d/maps", tid);
+    if (!(maps_file = fopen(maps_path, "r")))
+      FATAL() << "Failed to open " << maps_path;
+    ++*this;
+  }
+
+  pid_t tid;
   FILE* maps_file;
   string raw_line;
   KernelMapping km;
@@ -99,8 +94,8 @@ void KernelMapIterator::operator++() {
                              " %x:%x %" SCNu64 " %n",
                        &start, &end, flags, &offset, &dev_major, &dev_minor,
                        &inode, &chars_scanned);
-  ASSERT(t, 8 /*number of info fields*/ == nparsed ||
-                7 /*num fields if name is blank*/ == nparsed);
+  assert(8 /*number of info fields*/ == nparsed ||
+         7 /*num fields if name is blank*/ == nparsed);
 
   // trim trailing newline, if any
   int last_char = strlen(line) - 1;
@@ -120,9 +115,9 @@ void KernelMapIterator::operator++() {
     // that being correct yet.
     char proc_exe[PATH_MAX];
     char exe[PATH_MAX];
-    snprintf(proc_exe, sizeof(proc_exe), "/proc/%d/exe", t->tid);
+    snprintf(proc_exe, sizeof(proc_exe), "/proc/%d/exe", tid);
     readlink(proc_exe, exe, sizeof(exe));
-    FATAL() << "Sorry, tracee " << t->tid << " has x86-64 image " << exe
+    FATAL() << "Sorry, tracee " << tid << " has x86-64 image " << exe
             << " and that's not supported with a 32-bit rr.";
   }
 #endif
@@ -136,10 +131,9 @@ void KernelMapIterator::operator++() {
                      f, offset);
 }
 
-KernelMapping AddressSpace::read_kernel_mapping(Task* t,
-                                                remote_ptr<void> addr) {
+static KernelMapping read_kernel_mapping(pid_t tid, remote_ptr<void> addr) {
   MemoryRange range(addr, 1);
-  for (KernelMapIterator it(t); !it.at_end(); ++it) {
+  for (KernelMapIterator it(tid); !it.at_end(); ++it) {
     const KernelMapping& km = it.current();
     if (km.contains(range)) {
       return km;
@@ -148,6 +142,15 @@ KernelMapping AddressSpace::read_kernel_mapping(Task* t,
   return KernelMapping();
 }
 
+KernelMapping AddressSpace::read_kernel_mapping(Task* t,
+                                                remote_ptr<void> addr) {
+  return rr::read_kernel_mapping(t->tid, addr);
+}
+
+KernelMapping AddressSpace::read_local_kernel_mapping(uint8_t* addr) {
+  return rr::read_kernel_mapping(getpid(), remote_ptr<void>((uintptr_t)addr));
+}
+
 /**
  * Cat the /proc/[t->tid]/maps file to stdout, line by line.
  */
@@ -159,6 +162,19 @@ static void print_process_mmap(Task* t) {
   }
 }
 
+AddressSpace::Mapping::Mapping(const KernelMapping& map,
+                               const KernelMapping& recorded_map,
+                               EmuFile::shr_ptr emu_file, void* local_addr,
+                               shared_ptr<MonitoredSharedMemory>&& monitored)
+    : map(map),
+      recorded_map(recorded_map),
+      emu_file(emu_file),
+      local_addr(static_cast<uint8_t*>(local_addr)),
+      monitored_shared_memory(move(monitored)),
+      flags(FLAG_NONE) {}
+
+AddressSpace::Mapping::~Mapping() {}
+
 AddressSpace::~AddressSpace() { session_->on_destroy(this); }
 
 void AddressSpace::after_clone() { allocate_watchpoints(); }
@@ -205,7 +221,7 @@ remote_code_ptr AddressSpace::find_syscall_instruction(Task* t) {
 }
 
 static string find_rr_page_file(Task* t) {
-  string path = exe_directory() + "rr_page_";
+  string path = exe_directory() + "../bin/rr_page_";
   switch (t->arch()) {
     case x86:
       path += "32";
@@ -258,9 +274,76 @@ void AddressSpace::map_rr_page(Task* t) {
   map(rr_page_start(), rr_page_size(), prot, flags, 0, file_name, fstat.st_dev,
       fstat.st_ino);
 
-  traced_syscall_ip_ = rr_page_traced_syscall_ip(t->arch());
-  privileged_traced_syscall_ip_ =
-      rr_page_privileged_traced_syscall_ip(t->arch());
+  traced_syscall_ip_ = rr_page_syscall_entry_point(
+      TRACED, UNPRIVILEGED, RECORDING_AND_REPLAY, t->arch());
+  privileged_traced_syscall_ip_ = rr_page_syscall_entry_point(
+      TRACED, PRIVILEGED, RECORDING_AND_REPLAY, t->arch());
+}
+
+/**
+ * Must match generate_rr_page.py
+ */
+static const AddressSpace::SyscallType entry_points[] = {
+  { AddressSpace::TRACED, AddressSpace::UNPRIVILEGED,
+    AddressSpace::RECORDING_AND_REPLAY },
+  { AddressSpace::TRACED, AddressSpace::PRIVILEGED,
+    AddressSpace::RECORDING_AND_REPLAY },
+  { AddressSpace::UNTRACED, AddressSpace::UNPRIVILEGED,
+    AddressSpace::RECORDING_AND_REPLAY },
+  { AddressSpace::UNTRACED, AddressSpace::UNPRIVILEGED,
+    AddressSpace::REPLAY_ONLY },
+  { AddressSpace::UNTRACED, AddressSpace::UNPRIVILEGED,
+    AddressSpace::RECORDING_ONLY },
+  { AddressSpace::UNTRACED, AddressSpace::PRIVILEGED,
+    AddressSpace::RECORDING_AND_REPLAY },
+  { AddressSpace::UNTRACED, AddressSpace::PRIVILEGED,
+    AddressSpace::REPLAY_ONLY },
+  { AddressSpace::UNTRACED, AddressSpace::PRIVILEGED,
+    AddressSpace::RECORDING_ONLY },
+};
+
+static remote_code_ptr ip_from_index(size_t i) {
+  return remote_code_ptr(RR_PAGE_ADDR + RR_PAGE_SYSCALL_STUB_SIZE * i +
+                         RR_PAGE_SYSCALL_INSTRUCTION_END);
+}
+
+remote_code_ptr AddressSpace::rr_page_syscall_exit_point(Traced traced,
+                                                         Privileged privileged,
+                                                         Enabled enabled) {
+  for (auto& e : entry_points) {
+    if (e.traced == traced && e.privileged == privileged &&
+        e.enabled == enabled) {
+      return ip_from_index(&e - entry_points);
+    }
+  }
+  return nullptr;
+}
+
+const AddressSpace::SyscallType* AddressSpace::rr_page_syscall_from_exit_point(
+    remote_code_ptr ip) {
+  for (size_t i = 0; i < array_length(entry_points); ++i) {
+    if (ip_from_index(i) == ip) {
+      return &entry_points[i];
+    }
+  }
+  return nullptr;
+}
+
+remote_code_ptr AddressSpace::rr_page_syscall_entry_point(Traced traced,
+                                                          Privileged privileged,
+                                                          Enabled enabled,
+                                                          SupportedArch arch) {
+  remote_code_ptr ip = rr_page_syscall_exit_point(traced, privileged, enabled);
+  return ip.is_null() ? remote_code_ptr()
+                      : ip.decrement_by_syscall_insn_length(arch);
+}
+
+vector<AddressSpace::SyscallType> AddressSpace::rr_page_syscalls() {
+  vector<SyscallType> result;
+  for (auto& e : entry_points) {
+    result.push_back(e);
+  }
+  return result;
 }
 
 template <typename Arch> static vector<uint8_t> read_auxv_arch(Task* t) {
@@ -339,15 +422,15 @@ SupportedArch AddressSpace::arch() const {
   return (*task_set().begin())->arch();
 }
 
-TrapType AddressSpace::get_breakpoint_type_for_retired_insn(
+BreakpointType AddressSpace::get_breakpoint_type_for_retired_insn(
     remote_code_ptr ip) {
   remote_code_ptr addr = ip.decrement_by_bkpt_insn_length(SupportedArch::x86);
   return get_breakpoint_type_at_addr(addr);
 }
 
-TrapType AddressSpace::get_breakpoint_type_at_addr(remote_code_ptr addr) {
+BreakpointType AddressSpace::get_breakpoint_type_at_addr(remote_code_ptr addr) {
   auto it = breakpoints.find(addr);
-  return it == breakpoints.end() ? TRAP_NONE : it->second.type();
+  return it == breakpoints.end() ? BKPT_NONE : it->second.type();
 }
 
 bool AddressSpace::is_breakpoint_in_private_read_only_memory(
@@ -378,6 +461,11 @@ void AddressSpace::replace_breakpoints_with_original_values(
   }
 }
 
+bool AddressSpace::is_breakpoint_instruction(Task* t, remote_code_ptr ip) {
+  bool ok = true;
+  return t->read_mem(ip.to_data_ptr<uint8_t>(), &ok) == breakpoint_insn && ok;
+}
+
 static void remove_range(set<MemoryRange>& ranges, const MemoryRange& range) {
   auto start = ranges.lower_bound(range);
   auto end = start;
@@ -411,7 +499,8 @@ KernelMapping AddressSpace::map(remote_ptr<void> addr, size_t num_bytes,
                                 int prot, int flags, off64_t offset_bytes,
                                 const string& fsname, dev_t device, ino_t inode,
                                 const KernelMapping* recorded_map,
-                                TraceWriter::MappingOrigin origin) {
+                                EmuFile::shr_ptr emu_file, void* local_addr,
+                                shared_ptr<MonitoredSharedMemory>&& monitored) {
   LOG(debug) << "mmap(" << addr << ", " << num_bytes << ", " << HEX(prot)
              << ", " << HEX(flags) << ", " << HEX(offset_bytes);
   num_bytes = ceil_page_size(num_bytes);
@@ -432,7 +521,8 @@ KernelMapping AddressSpace::map(remote_ptr<void> addr, size_t num_bytes,
   unmap_internal(addr, num_bytes);
 
   const KernelMapping& actual_recorded_map = recorded_map ? *recorded_map : m;
-  map_and_coalesce(m, actual_recorded_map);
+  map_and_coalesce(m, actual_recorded_map, emu_file, local_addr,
+                   move(monitored));
 
   if ((prot & PROT_EXEC) &&
       (fsname.find(SYSCALLBUF_LIB_FILENAME) != string::npos ||
@@ -454,23 +544,29 @@ template <typename Arch> void AddressSpace::at_preload_init_arch(Task* t) {
   auto params = t->read_mem(
       remote_ptr<rrcall_init_preload_params<Arch> >(t->regs().arg1()));
 
-  ASSERT(t, t->session().as_record()->use_syscall_buffer() ==
-                params.syscallbuf_enabled)
-      << "Tracee thinks syscallbuf is "
-      << (params.syscallbuf_enabled ? "en" : "dis")
-      << "abled, but tracer thinks "
-      << (t->session().as_record()->use_syscall_buffer() ? "en" : "dis")
-      << "abled";
+  if (t->session().is_recording()) {
+    ASSERT(t, t->session().as_record()->use_syscall_buffer() ==
+                  params.syscallbuf_enabled)
+        << "Tracee thinks syscallbuf is "
+        << (params.syscallbuf_enabled ? "en" : "dis")
+        << "abled, but tracer thinks "
+        << (t->session().as_record()->use_syscall_buffer() ? "en" : "dis")
+        << "abled";
+  }
 
   if (!params.syscallbuf_enabled) {
     return;
   }
 
-  monkeypatch_state->patch_at_preload_init(t);
+  syscallbuf_enabled_ = true;
+
+  if (t->session().is_recording()) {
+    monkeypatch_state->patch_at_preload_init(static_cast<RecordTask*>(t));
+  }
 }
 
 void AddressSpace::at_preload_init(Task* t) {
-  ASSERT(t, syscallbuf_lib_start_)
+  ASSERT(t, !syscallbuf_lib_start_.is_null())
       << "should have found preload library already";
   RR_ARCH_FUNCTION(at_preload_init_arch, t->arch(), t);
 }
@@ -483,6 +579,11 @@ const AddressSpace::Mapping& AddressSpace::mapping_of(
   assert(it->second.map.contains(range));
   return it->second;
 }
+uint32_t& AddressSpace::mapping_flags_of(remote_ptr<void> addr) {
+  return const_cast<AddressSpace::Mapping&>(
+             static_cast<const AddressSpace*>(this)->mapping_of(addr))
+      .flags;
+}
 
 bool AddressSpace::has_mapping(remote_ptr<void> addr) const {
   if (addr + page_size() < addr) {
@@ -504,7 +605,7 @@ void AddressSpace::protect(remote_ptr<void> addr, size_t num_bytes, int prot) {
     LOG(debug) << "  protecting (" << rem << ") ...";
 
     Mapping m = move(mm);
-    mem.erase(m.map);
+    remove_from_map(m.map);
 
     // PROT_GROWSDOWN means that if this is a grows-down segment
     // (which for us means "stack") then the change should be
@@ -524,11 +625,13 @@ void AddressSpace::protect(remote_ptr<void> addr, size_t num_bytes, int prot) {
     // If the first segment we protect underflows the
     // region, remap the underflow region with previous
     // prot.
+    auto monitored = m.monitored_shared_memory;
     if (m.map.start() < new_start) {
       Mapping underflow(
           m.map.subrange(m.map.start(), rem.start()),
-          m.recorded_map.subrange(m.recorded_map.start(), rem.start()));
-      mem[underflow.map] = underflow;
+          m.recorded_map.subrange(m.recorded_map.start(), rem.start()),
+          m.emu_file, m.local_addr, move(monitored));
+      add_to_map(underflow);
     }
     // Remap the overlapping region with the new prot.
     remote_ptr<void> new_end = min(rem.end(), m.map.end());
@@ -536,17 +639,29 @@ void AddressSpace::protect(remote_ptr<void> addr, size_t num_bytes, int prot) {
     int new_prot = prot & (PROT_READ | PROT_WRITE | PROT_EXEC);
     Mapping overlap(
         m.map.subrange(new_start, new_end).set_prot(new_prot),
-        m.recorded_map.subrange(new_start, new_end).set_prot(new_prot));
-    mem[overlap.map] = overlap;
+        m.recorded_map.subrange(new_start, new_end).set_prot(new_prot),
+        m.emu_file,
+        m.local_addr ? m.local_addr + (new_start - m.map.start()) : 0,
+        m.monitored_shared_memory
+            ? m.monitored_shared_memory->subrange(new_start - m.map.start(),
+                                                  new_end - new_start)
+            : nullptr);
+    add_to_map(overlap);
     last_overlap = overlap.map;
 
     // If the last segment we protect overflows the
     // region, remap the overflow region with previous
     // prot.
     if (rem.end() < m.map.end()) {
-      Mapping overflow(m.map.subrange(rem.end(), m.map.end()),
-                       m.recorded_map.subrange(rem.end(), m.map.end()));
-      mem[overflow.map] = overflow;
+      Mapping overflow(
+          m.map.subrange(rem.end(), m.map.end()),
+          m.recorded_map.subrange(rem.end(), m.map.end()), m.emu_file,
+          m.local_addr ? m.local_addr + (rem.end() - m.map.start()) : 0,
+          m.monitored_shared_memory
+              ? m.monitored_shared_memory->subrange(rem.end() - m.map.start(),
+                                                    m.map.end() - rem.end())
+              : nullptr);
+      add_to_map(overflow);
     }
   };
   for_each_in_range(addr, num_bytes, protector, ITERATE_CONTIGUOUS);
@@ -580,6 +695,7 @@ void AddressSpace::remap(remote_ptr<void> old_addr, size_t old_num_bytes,
              << new_addr << ", " << new_num_bytes << ")";
 
   auto mr = mapping_of(old_addr);
+  assert(!mr.monitored_shared_memory);
   const KernelMapping& m = mr.map;
 
   old_num_bytes = ceil_page_size(old_num_bytes);
@@ -587,6 +703,7 @@ void AddressSpace::remap(remote_ptr<void> old_addr, size_t old_num_bytes,
   if (0 == new_num_bytes) {
     return;
   }
+  new_num_bytes = ceil_page_size(new_num_bytes);
 
   auto it = dont_fork.lower_bound(MemoryRange(old_addr, old_num_bytes));
   if (it != dont_fork.end() && it->start() < old_addr + old_num_bytes) {
@@ -601,10 +718,12 @@ void AddressSpace::remap(remote_ptr<void> old_addr, size_t old_num_bytes,
 
   remote_ptr<void> new_end = new_addr + new_num_bytes;
   map_and_coalesce(m.set_range(new_addr, new_end),
-                   mr.recorded_map.set_range(new_addr, new_end));
+                   mr.recorded_map.set_range(new_addr, new_end), mr.emu_file,
+                   nullptr, nullptr);
 }
 
-void AddressSpace::remove_breakpoint(remote_code_ptr addr, TrapType type) {
+void AddressSpace::remove_breakpoint(remote_code_ptr addr,
+                                     BreakpointType type) {
   auto it = breakpoints.find(addr);
   if (it == breakpoints.end() || it->second.unref(type) > 0) {
     return;
@@ -612,7 +731,7 @@ void AddressSpace::remove_breakpoint(remote_code_ptr addr, TrapType type) {
   destroy_breakpoint(it);
 }
 
-bool AddressSpace::add_breakpoint(remote_code_ptr addr, TrapType type) {
+bool AddressSpace::add_breakpoint(remote_code_ptr addr, BreakpointType type) {
   auto it = breakpoints.find(addr);
   if (it == breakpoints.end()) {
     uint8_t overwritten_data;
@@ -641,6 +760,22 @@ void AddressSpace::remove_all_breakpoints() {
   }
 }
 
+void AddressSpace::suspend_breakpoint_at(remote_code_ptr addr) {
+  auto it = breakpoints.find(addr);
+  if (it != breakpoints.end()) {
+    Task* t = *task_set().begin();
+    t->write_mem(addr.to_data_ptr<uint8_t>(), it->second.overwritten_data);
+  }
+}
+
+void AddressSpace::restore_breakpoint_at(remote_code_ptr addr) {
+  auto it = breakpoints.find(addr);
+  if (it != breakpoints.end()) {
+    Task* t = *task_set().begin();
+    t->write_mem(addr.to_data_ptr<uint8_t>(), breakpoint_insn);
+  }
+}
+
 int AddressSpace::access_bits_of(WatchType type) {
   switch (type) {
     case WATCH_EXEC:
@@ -791,22 +926,31 @@ void AddressSpace::unmap_internal(remote_ptr<void> addr, ssize_t num_bytes) {
     LOG(debug) << "  unmapping (" << rem << ") ...";
 
     Mapping m = move(mm);
-    mem.erase(m.map);
+    remove_from_map(m.map);
+
     LOG(debug) << "  erased (" << m.map << ") ...";
 
     // If the first segment we unmap underflows the unmap
     // region, remap the underflow region.
+    auto monitored = m.monitored_shared_memory;
     if (m.map.start() < rem.start()) {
       Mapping underflow(m.map.subrange(m.map.start(), rem.start()),
-                        m.recorded_map.subrange(m.map.start(), rem.start()));
-      mem[underflow.map] = underflow;
+                        m.recorded_map.subrange(m.map.start(), rem.start()),
+                        m.emu_file, m.local_addr, move(monitored));
+      add_to_map(underflow);
     }
     // If the last segment we unmap overflows the unmap
     // region, remap the overflow region.
     if (rem.end() < m.map.end()) {
-      Mapping overflow(m.map.subrange(rem.end(), m.map.end()),
-                       m.recorded_map.subrange(rem.end(), m.map.end()));
-      mem[overflow.map] = overflow;
+      Mapping overflow(
+          m.map.subrange(rem.end(), m.map.end()),
+          m.recorded_map.subrange(rem.end(), m.map.end()), m.emu_file,
+          m.local_addr ? m.local_addr + (rem.end() - m.map.start()) : 0,
+          m.monitored_shared_memory
+              ? m.monitored_shared_memory->subrange(rem.end() - m.map.start(),
+                                                    m.map.end() - rem.end())
+              : nullptr);
+      add_to_map(overflow);
     }
   };
   for_each_in_range(addr, num_bytes, unmapper);
@@ -983,15 +1127,6 @@ static void assert_segments_match(Task* t, const KernelMapping& input_m,
   }
 }
 
-KernelMapping AddressSpace::fix_stack_segment_start(
-    const MemoryRange& mapping, remote_ptr<void> new_start) {
-  auto it = mem.find(mapping);
-  it->first.update_start(new_start);
-  it->second.map.update_start(new_start);
-  it->second.recorded_map.update_start(new_start);
-  return it->second.map;
-}
-
 KernelMapping AddressSpace::vdso() const {
   assert(!vdso_start_addr.is_null());
   return mapping_of(vdso_start_addr).map;
@@ -1039,18 +1174,18 @@ AddressSpace::AddressSpace(Task* t, const string& exe, uint32_t exec_count)
       session_(&t->session()),
       monkeypatch_state(t->session().is_recording() ? new Monkeypatcher()
                                                     : nullptr),
-      child_mem_fd(-1),
+      syscallbuf_enabled_(false),
       first_run_event_(0) {
   // TODO: this is a workaround of
   // https://github.com/mozilla/rr/issues/1113 .
-  if (session_->can_validate()) {
+  if (session_->done_initial_exec()) {
     populate_address_space(t);
     assert(!vdso_start_addr.is_null());
   } else {
     // Find the location of the VDSO in the just-spawned process. This will
     // match the VDSO in rr itself since we haven't execed yet. So, speed
     // things up by searching rr's own VDSO for a syscall instruction.
-    size_t rr_vdso_len;
+    size_t rr_vdso_len = 0;
     remote_ptr<void> rr_vdso = find_rr_vdso(t, &rr_vdso_len);
     // Here we rely on the VDSO location in the spawned tracee being the same
     // as in rr itself.
@@ -1075,6 +1210,7 @@ AddressSpace::AddressSpace(Session* session, const AddressSpace& o,
       brk_end(o.brk_end),
       is_clone(true),
       mem(o.mem),
+      monitored_mem(o.monitored_mem),
       session_(session),
       vdso_start_addr(o.vdso_start_addr),
       monkeypatch_state(o.monkeypatch_state
@@ -1084,6 +1220,7 @@ AddressSpace::AddressSpace(Session* session, const AddressSpace& o,
       privileged_traced_syscall_ip_(o.privileged_traced_syscall_ip_),
       syscallbuf_lib_start_(o.syscallbuf_lib_start_),
       syscallbuf_lib_end_(o.syscallbuf_lib_end_),
+      syscallbuf_enabled_(o.syscallbuf_enabled_),
       saved_auxv_(o.saved_auxv_),
       first_run_event_(0) {
   for (auto& it : o.breakpoints) {
@@ -1216,6 +1353,25 @@ vector<WatchConfig> AddressSpace::all_watchpoints() {
   return get_watchpoints_internal(ALL_WATCHPOINTS);
 }
 
+bool AddressSpace::has_any_watchpoint_changes() {
+  for (auto& kv : watchpoints) {
+    if (kv.second.changed) {
+      return true;
+    }
+  }
+  return false;
+}
+
+bool AddressSpace::has_exec_watchpoint_fired(remote_code_ptr addr) {
+  for (auto& kv : watchpoints) {
+    if (kv.second.changed && kv.second.exec_count > 0 &&
+        kv.first.start() == addr.to_data_ptr<void>()) {
+      return true;
+    }
+  }
+  return false;
+}
+
 bool AddressSpace::allocate_watchpoints() {
   Task::DebugRegs regs = get_watch_configs(SETTING_TASK_STATE);
 
@@ -1241,6 +1397,15 @@ bool AddressSpace::allocate_watchpoints() {
   return false;
 }
 
+static inline void assert_coalesceable(const AddressSpace::Mapping& lower,
+                                       const AddressSpace::Mapping& higher) {
+  assert(lower.emu_file == higher.emu_file);
+  assert(lower.flags == higher.flags);
+  assert((lower.local_addr == 0 && higher.local_addr == 0) ||
+         lower.local_addr + lower.map.size() == higher.local_addr);
+  assert(!lower.monitored_shared_memory && !higher.monitored_shared_memory);
+}
+
 void AddressSpace::coalesce_around(MemoryMap::iterator it) {
   auto first_kv = it;
   while (mem.begin() != first_kv) {
@@ -1251,6 +1416,7 @@ void AddressSpace::coalesce_around(MemoryMap::iterator it) {
       first_kv = next;
       break;
     }
+    assert_coalesceable(first_kv->second, next->second);
   }
   auto last_kv = it;
   while (true) {
@@ -1262,6 +1428,7 @@ void AddressSpace::coalesce_around(MemoryMap::iterator it) {
       last_kv = prev;
       break;
     }
+    assert_coalesceable(prev->second, last_kv->second);
   }
   assert(last_kv != mem.end());
   if (first_kv == last_kv) {
@@ -1270,9 +1437,13 @@ void AddressSpace::coalesce_around(MemoryMap::iterator it) {
   }
 
   Mapping new_m(first_kv->second.map.extend(last_kv->first.end()),
-                first_kv->second.recorded_map.extend(last_kv->first.end()));
+                first_kv->second.recorded_map.extend(last_kv->first.end()),
+                first_kv->second.emu_file, first_kv->second.local_addr);
+  new_m.flags = first_kv->second.flags;
   LOG(debug) << "  coalescing " << new_m.map;
 
+  // monitored-memory currently isn't coalescable so we don't need to
+  // adjust monitored_mem
   mem.erase(first_kv, ++last_kv);
 
   auto ins = mem.insert(MemoryMap::value_type(new_m.map, new_m));
@@ -1280,6 +1451,8 @@ void AddressSpace::coalesce_around(MemoryMap::iterator it) {
 }
 
 void AddressSpace::destroy_breakpoint(BreakpointMap::const_iterator it) {
+  if (task_set().empty())
+    return;
   Task* t = *task_set().begin();
   t->write_mem(it->first.to_data_ptr<uint8_t>(), it->second.overwritten_data);
   breakpoints.erase(it);
@@ -1324,18 +1497,24 @@ void AddressSpace::for_each_in_range(
   }
 }
 
-void AddressSpace::map_and_coalesce(const KernelMapping& m,
-                                    const KernelMapping& recorded_map) {
+void AddressSpace::map_and_coalesce(
+    const KernelMapping& m, const KernelMapping& recorded_map,
+    EmuFile::shr_ptr emu_file, void* local_addr,
+    shared_ptr<MonitoredSharedMemory>&& monitored) {
   LOG(debug) << "  mapping " << m;
 
-  auto ins = mem.insert(MemoryMap::value_type(m, Mapping(m, recorded_map)));
+  if (monitored) {
+    monitored_mem.insert(m.start());
+  }
+  auto ins = mem.insert(MemoryMap::value_type(
+      m, Mapping(m, recorded_map, emu_file, local_addr, move(monitored))));
   coalesce_around(ins.first);
 
   update_watchpoint_values(m.start(), m.end());
 }
 
 static bool could_be_stack(const KernelMapping& km) {
-  // On 4.1.6-200.fc22.x86_64 we observe that during exec of the exec_stub
+  // On 4.1.6-200.fc22.x86_64 we observe that during exec of the rr_exec_stub
   // during replay, when the process switches from 32-bit to 64-bit, the 64-bit
   // registers seem truncated to 32 bits during the initial PTRACE_GETREGS so
   // our sp looks wrong and /proc/<pid>/maps doesn't identify the region as
@@ -1376,17 +1555,109 @@ void AddressSpace::populate_address_space(Task* t) {
     if (is_stack) {
       ++found_stacks;
       flags |= MAP_GROWSDOWN;
-      // MAP_GROWSDOWN segments really occupy one additional page before
-      // the start address shown by /proc/<pid>/maps --- unless that page
-      // is already occupied by another mapping.
-      if (!has_mapping(start - page_size())) {
-        start -= page_size();
+      if (uses_invisible_guard_page()) {
+        // MAP_GROWSDOWN segments really occupy one additional page before
+        // the start address shown by /proc/<pid>/maps --- unless that page
+        // is already occupied by another mapping.
+        if (!has_mapping(start - page_size())) {
+          start -= page_size();
+        }
       }
     }
 
     map(start, km.end() - start, km.prot(), flags, km.file_offset_bytes(),
-        km.fsname(), check_device(t, km), km.inode(), nullptr,
-        TraceWriter::EXEC_MAPPING);
+        km.fsname(), check_device(t, km), km.inode(), nullptr);
   }
   ASSERT(t, found_stacks == 1);
 }
+
+static int random_addr_bits(SupportedArch arch) {
+  switch (arch) {
+    default:
+      assert(0 && "Unknown architecture");
+    case x86:
+      return 32;
+    // Current x86-64 systems have only 48 bits of virtual address space,
+    // and only the bottom half is usable by user space
+    case x86_64:
+      return 47;
+  }
+}
+
+static MemoryRange adjust_range_for_stack_growth(const KernelMapping& km) {
+  remote_ptr<void> start = km.start();
+  if (km.flags() & MAP_GROWSDOWN) {
+    start = min(start, km.end() - AddressSpace::chaos_mode_min_stack_size());
+  }
+  return MemoryRange(start, km.end());
+}
+
+remote_ptr<void> AddressSpace::chaos_mode_find_free_memory(Task* t,
+                                                           size_t len) {
+  remote_ptr<void> addr;
+  // Half the time, try to allocate at a completely random address. The other
+  // half of the time, we'll try to allocate immediately before or after a
+  // randomly chosen existing mapping.
+  if (random() % 2) {
+    int bits = random_addr_bits(t->arch());
+    // Some of these addresses will not be mappable. That's fine, the
+    // kernel will fall back to a valid address if the hint is not valid.
+    uint64_t r = ((uint64_t)(uint32_t)random() << 32) | (uint32_t)random();
+    addr = floor_page_size(remote_ptr<void>(r & ((uint64_t(1) << bits) - 1)));
+  } else {
+    ASSERT(t, !mem.empty());
+    int map_index = random() % mem.size();
+    int map_count = 0;
+    for (const auto& m : maps()) {
+      if (map_count == map_index) {
+        addr = m.map.start();
+        break;
+      }
+      ++map_count;
+    }
+  }
+
+  // If there's a collision (which there always will be in the second case
+  // above), either move the mapping forwards or backwards in memory until it
+  // fits. Choose the direction randomly.
+  int direction = (random() % 2) ? 1 : -1;
+  while (true) {
+    Maps m = maps_starting_at(addr);
+    if (m.begin() == m.end()) {
+      return addr;
+    }
+    MemoryRange range = adjust_range_for_stack_growth(m.begin()->map);
+    if (range.start() >= addr + len) {
+      // No overlap with an existing mapping; we're good!
+      return addr;
+    }
+    if (direction == -1) {
+      addr = range.start() - len;
+    } else {
+      addr = range.end();
+    }
+  }
+}
+
+remote_ptr<void> AddressSpace::find_free_memory(size_t required_space,
+                                                remote_ptr<void> after) {
+  auto maps = maps_starting_at(after);
+  auto current = maps.begin();
+  while (current != maps.end()) {
+    auto next = current;
+    ++next;
+    if (next == maps.end()) {
+      if (current->map.end() + required_space >= current->map.end()) {
+        break;
+      }
+    } else {
+      if (current->map.end() + required_space <= next->map.start()) {
+        break;
+      }
+    }
+    current = next;
+  }
+  return current->map.end();
+}
+
+} // namespace rr
