@@ -1,23 +1,25 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-//#define DEBUGTAG "RemoteSyscalls"
-
 #include "AutoRemoteSyscalls.h"
 
 #include <limits.h>
+#include <linux/net.h>
+#include <sys/socket.h>
+#include <sys/un.h>
 
 #include "rr/rr.h"
 
-#include "kernel_metadata.h"
-#include "log.h"
 #include "ReplaySession.h"
 #include "Session.h"
-#include "task.h"
+#include "Task.h"
+#include "kernel_metadata.h"
+#include "log.h"
 #include "util.h"
 
-using namespace rr;
 using namespace std;
 
+namespace rr {
+
 /**
  * The ABI of the socketcall syscall is a nightmare; the first arg to
  * the kernel is the sub-operation, and the second argument is a
@@ -27,7 +29,7 @@ template <typename Arch> struct socketcall_args {
   typename Arch::signed_long args[3];
 } __attribute__((packed));
 
-void AutoRestoreMem::init(const uint8_t* mem, ssize_t num_bytes) {
+void AutoRestoreMem::init(const void* mem, ssize_t num_bytes) {
   ASSERT(remote.task(), !remote.regs().sp().is_null())
       << "Memory parameters were disabled";
 
@@ -72,6 +74,19 @@ AutoRemoteSyscalls::AutoRemoteSyscalls(Task* t,
   } else {
     initial_regs.set_sp(remote_ptr<void>());
   }
+  // We need to make sure to clear any breakpoints or other alterations of
+  // the syscall instruction we're using. Note that the tracee may have set its
+  // own breakpoints or otherwise modified the instruction, so suspending our
+  // own breakpoint is insufficient.
+  std::vector<uint8_t> syscall = rr::syscall_instruction(t->arch());
+  replaced_bytes =
+      t->read_mem(initial_regs.ip().to_data_ptr<uint8_t>(), syscall.size());
+  if (replaced_bytes == syscall) {
+    replaced_bytes.clear();
+  } else {
+    t->write_mem(initial_regs.ip().to_data_ptr<uint8_t>(), syscall.data(),
+                 syscall.size());
+  }
 }
 
 static bool is_usable_area(const KernelMapping& km) {
@@ -80,7 +95,7 @@ static bool is_usable_area(const KernelMapping& km) {
 }
 
 void AutoRemoteSyscalls::maybe_fix_stack_pointer() {
-  if (!t->session().can_validate()) {
+  if (!t->session().done_initial_exec()) {
     return;
   }
 
@@ -109,6 +124,11 @@ void AutoRemoteSyscalls::maybe_fix_stack_pointer() {
 AutoRemoteSyscalls::~AutoRemoteSyscalls() { restore_state_to(t); }
 
 void AutoRemoteSyscalls::restore_state_to(Task* t) {
+  if (!replaced_bytes.empty()) {
+    t->write_mem(remote_ptr<uint8_t>(
+                     t->vm()->traced_syscall_ip().to_data_ptr<uint8_t>()),
+                 replaced_bytes.data(), replaced_bytes.size());
+  }
   initial_regs.set_ip(initial_ip);
   initial_regs.set_sp(initial_sp);
   // Restore stomped registers.
@@ -141,7 +161,7 @@ void AutoRemoteSyscalls::syscall_helper(SyscallWaiting wait, int syscallno,
 void AutoRemoteSyscalls::wait_syscall(int syscallno) {
   ASSERT(t, pending_syscallno == syscallno || syscallno < 0);
 
-  // Wait for syscall-exit trap.
+  // Wait for syscall-exit trap (or PTRACE_EVENT_* trap).
   t->wait();
   pending_syscallno = -1;
 
@@ -195,6 +215,10 @@ static int create_bind_and_listen_socket(const char* path) {
     FATAL() << "Failed to bind listen socket";
   }
 
+  if (chmod(path, 0666)) {
+    FATAL() << "Failed to make listening socket world-writeable";
+  }
+
   if (listen(listen_sock, 1)) {
     FATAL() << "Failed to mark listening for listen socket";
   }
@@ -234,8 +258,15 @@ static void child_connect_socket(AutoRemoteSyscalls& remote,
   auto tmp_buf_end = buf_end;
   auto remote_dot = allocate(&tmp_buf_end, remote_buf, 2);
   remote.task()->write_mem(remote_dot.cast<char>(), ".", 2);
-  *cwd_fd = remote.infallible_syscall(syscall_number_for_open(Arch::arch()),
-                                      remote_dot, O_PATH | O_DIRECTORY);
+  // This may fail because the tracee has done a setuid and is no longer
+  // able to access its own CWD. There doesn't seem to be a way to save
+  // and restore the CWD in this situation --- and connectat() doesn't exist yet
+  // so we have no way to avoid having to set the CWD if we're in a chrooted
+  // environment. We'll hack around such a failure later by putting the child
+  // into a new directory that it can't access.
+  *cwd_fd = remote.syscall(syscall_number_for_open(Arch::arch()), remote_dot,
+                           O_PATH | O_DIRECTORY);
+  ASSERT(remote.task(), *cwd_fd >= 0 || *cwd_fd == -EACCES);
   remote.infallible_syscall(Arch::fchdir, RR_RESERVED_ROOT_DIR_FD);
 
   auto remote_addr = allocate<typename Arch::sockaddr_un>(&buf_end, remote_buf);
@@ -258,6 +289,31 @@ static void child_connect_socket(AutoRemoteSyscalls& remote,
                         callregs);
 }
 
+template <typename Arch>
+static void restore_cwd(AutoRemoteSyscalls& remote, int cwd_fd) {
+  if (cwd_fd >= 0) {
+    remote.infallible_syscall(Arch::fchdir, cwd_fd);
+    remote.infallible_syscall(Arch::close, cwd_fd);
+  } else {
+    // We can't get back to the original directory. (Note that we may not
+    // be able to access it ourselves from outside the tracee, if it's in a
+    // mount namespace created by the tracee.)
+    // Fake it by making a world-accessible directory, chdir the tracee into it,
+    // then make it inaccessible.
+    char path[PATH_MAX];
+    sprintf(path, "/tmp/rr-tracee-fd-inaccessible-dir-%d-%ld",
+            remote.task()->tid, random());
+    int ret = mkdir(path, 0777);
+    ASSERT(remote.task(), ret >= 0);
+    AutoRestoreMem mem(remote, path);
+    remote.infallible_syscall(Arch::chdir, mem.get());
+    ret = chmod(path, 0000);
+    ASSERT(remote.task(), ret >= 0);
+    ret = rmdir(path);
+    ASSERT(remote.task(), ret >= 0);
+  }
+}
+
 template <typename Arch>
 static void child_sendmsg(AutoRemoteSyscalls& remote,
                           AutoRestoreMem& remote_buf,
@@ -316,13 +372,12 @@ static void child_sendmsg(AutoRemoteSyscalls& remote,
 }
 
 static int recvmsg_socket(int sock) {
-  char cmsgbuf[CMSG_SPACE(sizeof(int))];
-
   char received_data;
   struct iovec msgdata;
   msgdata.iov_base = &received_data;
   msgdata.iov_len = 1;
 
+  char cmsgbuf[CMSG_SPACE(sizeof(int))];
   struct msghdr msg;
   memset(&msg, 0, sizeof(msg));
   msg.msg_control = cmsgbuf;
@@ -373,6 +428,20 @@ template <typename Arch> ScopedFd AutoRemoteSyscalls::retrieve_fd_arch(int fd) {
   if (sock < 0) {
     FATAL() << "Failed to create parent socket";
   }
+
+  // Verify that the child's PID is correct. This is a world-writeable socket
+  // so anyone could connect, and this check is vital.
+  struct ucred cred;
+  cred.pid = -1;
+  socklen_t credlen = sizeof(cred);
+  if (getsockopt(sock, SOL_SOCKET, SO_PEERCRED, &cred, &credlen) < 0) {
+    FATAL() << "Failed SO_PEERCRED";
+  }
+  ASSERT(t, credlen == sizeof(cred));
+  if (cred.pid != task()->real_tgid()) {
+    FATAL() << "Some other process connected to our socket!!!";
+  }
+
   // Complete child's connect() syscall
   wait_syscall();
   int child_syscall_result = t->regs().syscall_result_signed();
@@ -380,8 +449,7 @@ template <typename Arch> ScopedFd AutoRemoteSyscalls::retrieve_fd_arch(int fd) {
     FATAL() << "Failed to connect() in tracee; err="
             << errno_name(-child_syscall_result);
   }
-  infallible_syscall(Arch::fchdir, cwd_fd);
-  infallible_syscall(Arch::close, cwd_fd);
+  restore_cwd<Arch>(*this, cwd_fd);
 
   // Listening socket not needed anymore
   close(listen_sock);
@@ -428,6 +496,24 @@ remote_ptr<void> AutoRemoteSyscalls::infallible_mmap_syscall(
   return ret;
 }
 
+int64_t AutoRemoteSyscalls::infallible_lseek_syscall(int fd, int64_t offset,
+                                                     int whence) {
+  switch (arch()) {
+    case x86: {
+      AutoRestoreMem mem(*this, &offset, sizeof(int64_t));
+      infallible_syscall(syscall_number_for__llseek(arch()), fd, offset >> 32,
+                         offset, mem.get(), whence);
+      return t->read_mem(mem.get().cast<int64_t>());
+    }
+    case x86_64:
+      return infallible_syscall(syscall_number_for_lseek(arch()), fd, offset,
+                                whence);
+    default:
+      ASSERT(task(), false) << "Unknown arch";
+      return -1;
+  }
+}
+
 void AutoRemoteSyscalls::check_syscall_result(int syscallno) {
   long ret = t->regs().syscall_result_signed();
   if (-4096 < ret && ret < 0) {
@@ -441,3 +527,5 @@ void AutoRemoteSyscalls::check_syscall_result(int syscallno) {
                      << " failed with errno " << errno_name(-ret) << extra_msg;
   }
 }
+
+} // namespace rr
