@@ -9,16 +9,18 @@
 
 #include "AddressSpace.h"
 #include "Command.h"
+#include "TraceStream.h"
 #include "kernel_metadata.h"
 #include "main.h"
-#include "TraceStream.h"
 #include "util.h"
 
 using namespace std;
 
+namespace rr {
+
 class DumpCommand : public Command {
 public:
-  virtual int run(std::vector<std::string>& args);
+  virtual int run(vector<string>& args);
 
 protected:
   DumpCommand(const char* name, const char* help) : Command(name, help) {}
@@ -32,6 +34,7 @@ DumpCommand DumpCommand::singleton(
     "  Event specs can be either an event number like `127', or a range\n"
     "  like `1000-5000'.  By default, all events are dumped.\n"
     "  -b, --syscallbuf           dump syscallbuf contents\n"
+    " --g, --generic              dump generic data\n"
     "  -m, --recorded-metadata    dump recorded data metadata\n"
     "  -p, --mmaps                dump mmap data\n"
     "  -r, --raw                  dump trace frames in a more easily\n"
@@ -41,6 +44,7 @@ DumpCommand DumpCommand::singleton(
 
 struct DumpFlags {
   bool dump_syscallbuf;
+  bool dump_generic;
   bool dump_recorded_data_metadata;
   bool dump_mmaps;
   bool raw_dump;
@@ -48,23 +52,26 @@ struct DumpFlags {
 
   DumpFlags()
       : dump_syscallbuf(false),
+        dump_generic(false),
         dump_recorded_data_metadata(false),
         dump_mmaps(false),
         raw_dump(false),
         dump_statistics(false) {}
 };
 
-static bool parse_dump_arg(std::vector<std::string>& args, DumpFlags& flags) {
+static bool parse_dump_arg(vector<string>& args, DumpFlags& flags) {
   if (parse_global_option(args)) {
     return true;
   }
 
-  static const OptionSpec options[] = { { 'b', "syscallbuf", NO_PARAMETER },
-                                        { 'm', "recorded-metadata",
-                                          NO_PARAMETER },
-                                        { 'p', "mmaps", NO_PARAMETER },
-                                        { 'r', "raw", NO_PARAMETER },
-                                        { 's', "statistics", NO_PARAMETER } };
+  static const OptionSpec options[] = {
+    { 'b', "syscallbuf", NO_PARAMETER },
+    { 'g', "generic", NO_PARAMETER },
+    { 'm', "recorded-metadata", NO_PARAMETER },
+    { 'p', "mmaps", NO_PARAMETER },
+    { 'r', "raw", NO_PARAMETER },
+    { 's', "statistics", NO_PARAMETER }
+  };
   ParsedOption opt;
   if (!Command::parse_option(args, options, &opt)) {
     return false;
@@ -74,6 +81,9 @@ static bool parse_dump_arg(std::vector<std::string>& args, DumpFlags& flags) {
     case 'b':
       flags.dump_syscallbuf = true;
       break;
+    case 'g':
+      flags.dump_generic = true;
+      break;
     case 'm':
       flags.dump_recorded_data_metadata = true;
       break;
@@ -165,7 +175,8 @@ static void dump_events_matching(TraceReader& trace, const DumpFlags& flags,
       while (true) {
         TraceReader::MappedData data;
         bool found;
-        KernelMapping km = trace.read_mapped_region(&data, &found);
+        KernelMapping km =
+            trace.read_mapped_region(&data, &found, TraceReader::DONT_VALIDATE);
         if (!found) {
           break;
         }
@@ -184,10 +195,14 @@ static void dump_events_matching(TraceReader& trace, const DumpFlags& flags,
             prot_flags[3] = 's';
           }
           fprintf(out, "  { map_file:\"%s\", addr:%p, length:%p, "
-                       "prot_flags:\"%s\", file_offset:0x%llx }\n",
+                       "prot_flags:\"%s\", file_offset:0x%llx, "
+                       "data_file:\"%s\", data_offset:0x%llx, "
+                       "file_size:0x%llx }\n",
                   km.fsname().c_str(), (void*)km.start().as_int(),
                   (void*)km.size(), prot_flags,
-                  (long long)km.file_offset_bytes());
+                  (long long)km.file_offset_bytes(), data.file_name.c_str(),
+                  (long long)data.data_offset_bytes,
+                  (long long)data.file_size_bytes);
         }
       }
 
@@ -198,6 +213,10 @@ static void dump_events_matching(TraceReader& trace, const DumpFlags& flags,
                   (void*)data.data.size());
         }
       }
+      vector<uint8_t> buf;
+      while (flags.dump_generic && trace.read_generic_for_frame(frame, buf)) {
+        fprintf(out, "  { length:%p }\n", (void*)buf.size());
+      }
       if (!flags.raw_dump) {
         fprintf(out, "}\n");
       }
@@ -207,7 +226,8 @@ static void dump_events_matching(TraceReader& trace, const DumpFlags& flags,
       }
       while (true) {
         TraceReader::MappedData data;
-        KernelMapping km = trace.read_mapped_region(&data);
+        KernelMapping km = trace.read_mapped_region(&data, nullptr,
+                                                    TraceReader::DONT_VALIDATE);
         if (km.size() == 0) {
           break;
         }
@@ -219,8 +239,8 @@ static void dump_events_matching(TraceReader& trace, const DumpFlags& flags,
 static void dump_statistics(const TraceReader& trace, FILE* out) {
   uint64_t uncompressed = trace.uncompressed_bytes();
   uint64_t compressed = trace.compressed_bytes();
-  fprintf(stdout, "// Uncompressed bytes %" PRIu64 ", compressed bytes %" PRIu64
-                  ", ratio %.2fx\n",
+  fprintf(out, "// Uncompressed bytes %" PRIu64 ", compressed bytes %" PRIu64
+               ", ratio %.2fx\n",
           uncompressed, compressed, double(uncompressed) / compressed);
 }
 
@@ -248,7 +268,7 @@ static void dump(const string& trace_dir, const DumpFlags& flags,
   }
 }
 
-int DumpCommand::run(std::vector<std::string>& args) {
+int DumpCommand::run(vector<string>& args) {
   DumpFlags flags;
 
   while (parse_dump_arg(args, flags)) {
@@ -263,3 +283,5 @@ int DumpCommand::run(std::vector<std::string>& args) {
   dump(trace_dir, flags, args, stdout);
   return 0;
 }
+
+} // namespace rr
