@@ -8,12 +8,14 @@
 
 #include "rr/rr.h"
 
-#include "log.h"
 #include "Session.h"
-#include "task.h"
+#include "Task.h"
+#include "log.h"
 
 using namespace std;
 
+namespace rr {
+
 void FdTable::add_monitor(int fd, FileMonitor* monitor) {
   // In the future we could support multiple monitors on an fd, but we don't
   // need to yet.
@@ -30,6 +32,32 @@ bool FdTable::allow_close(int fd) {
   return it->second->allow_close();
 }
 
+bool FdTable::emulate_ioctl(int fd, RecordTask* t, uint64_t* result) {
+  auto it = fds.find(fd);
+  if (it == fds.end()) {
+    return false;
+  }
+  return it->second->emulate_ioctl(t, result);
+}
+
+bool FdTable::emulate_fcntl(int fd, RecordTask* t, uint64_t* result) {
+  auto it = fds.find(fd);
+  if (it == fds.end()) {
+    return false;
+  }
+  return it->second->emulate_fcntl(t, result);
+}
+
+bool FdTable::emulate_read(int fd, RecordTask* t,
+                           const std::vector<FileMonitor::Range>& ranges,
+                           int64_t offset, uint64_t* result) {
+  auto it = fds.find(fd);
+  if (it == fds.end()) {
+    return false;
+  }
+  return it->second->emulate_read(t, ranges, offset, result);
+}
+
 Switchable FdTable::will_write(Task* t, int fd) {
   auto it = fds.find(fd);
   if (it == fds.end()) {
@@ -39,10 +67,11 @@ Switchable FdTable::will_write(Task* t, int fd) {
 }
 
 void FdTable::did_write(Task* t, int fd,
-                        const std::vector<FileMonitor::Range>& ranges) {
+                        const std::vector<FileMonitor::Range>& ranges,
+                        int64_t offset) {
   auto it = fds.find(fd);
   if (it != fds.end()) {
-    it->second->did_write(t, ranges);
+    it->second->did_write(t, ranges, offset);
   }
 }
 
@@ -83,10 +112,11 @@ void FdTable::update_syscallbuf_fds_disabled(int fd) {
     }
     vms_updated.insert(vm);
 
-    if (!t->syscallbuf_fds_disabled_child.is_null() &&
-        fd < SYSCALLBUF_FDS_DISABLED_SIZE) {
+    if (!t->preload_globals.is_null() && fd < SYSCALLBUF_FDS_DISABLED_SIZE) {
       bool disable = is_fd_monitored_in_any_task(vm, fd);
-      t->write_mem(t->syscallbuf_fds_disabled_child + fd, (char)disable);
+      t->write_mem(
+          REMOTE_PTR_FIELD(t->preload_globals, syscallbuf_fds_disabled[0]) + fd,
+          (char)disable);
     }
   }
 }
@@ -94,7 +124,7 @@ void FdTable::update_syscallbuf_fds_disabled(int fd) {
 void FdTable::init_syscallbuf_fds_disabled(Task* t) {
   ASSERT(t, has_task(t));
 
-  if (t->syscallbuf_fds_disabled_child.is_null()) {
+  if (t->preload_globals.is_null()) {
     return;
   }
 
@@ -114,8 +144,8 @@ void FdTable::init_syscallbuf_fds_disabled(Task* t) {
     }
   }
 
-  t->write_mem(t->syscallbuf_fds_disabled_child, disabled,
-               SYSCALLBUF_FDS_DISABLED_SIZE);
+  t->write_mem(REMOTE_PTR_FIELD(t->preload_globals, syscallbuf_fds_disabled[0]),
+               disabled, SYSCALLBUF_FDS_DISABLED_SIZE);
 }
 
 static bool is_fd_open(Task* t, int fd) {
@@ -145,3 +175,5 @@ void FdTable::update_for_cloexec(Task* t, TraceTaskEvent& event) {
     did_close(fd);
   }
 }
+
+} // namespace rr
