@@ -13,6 +13,11 @@ class Task;
 
 #include "util.h"
 
+namespace rr {
+
+class RecordTask;
+class Registers;
+
 class FileMonitor {
 public:
   typedef std::shared_ptr<FileMonitor> shr_ptr;
@@ -36,7 +41,7 @@ public:
    * write a blocking write. This ensures that writes are performed in the order
    * of will_write notifications.
    */
-  virtual Switchable will_write(Task* t) { return ALLOW_SWITCH; }
+  virtual Switchable will_write(Task*) { return ALLOW_SWITCH; }
   /**
    * Notification that task |t| wrote to the file descriptor.
    * Due to races, if will_write did not return PREVENT_SWITCH, it's possible
@@ -47,7 +52,34 @@ public:
     size_t length;
     Range(remote_ptr<void> data, size_t length) : data(data), length(length) {}
   };
-  virtual void did_write(Task* t, const std::vector<Range>& ranges) {}
+  virtual void did_write(Task*, const std::vector<Range>&, int64_t) {}
+
+  /**
+   * Return true if the ioctl should be fully emulated. If so the result
+   * is stored in the last parameter.
+   * Only called during recording.
+   */
+  virtual bool emulate_ioctl(RecordTask*, uint64_t*) { return false; }
+
+  /**
+   * Return true if the fcntl should should be fully emulated. If so the
+   * result is stored in the last parameter.
+   * Only called during recording.
+   */
+  virtual bool emulate_fcntl(RecordTask*, uint64_t*) { return false; }
+
+  /**
+   * Return true if the read should should be fully emulated. If so the
+   * result is stored in the last parameter. The emulation should write to the
+   * task's memory ranges.
+   * Only called during recording.
+   */
+  virtual bool emulate_read(RecordTask*, const std::vector<Range>&, int64_t,
+                            uint64_t*) {
+    return false;
+  }
 };
 
+} // namespace rr
+
 #endif /* RR_FILE_MONITOR_H_ */
