@@ -5,6 +5,8 @@
 
 #include <string>
 
+namespace rr {
+
 class GdbCommand;
 class GdbServer;
 class Task;
@@ -27,7 +29,11 @@ public:
   static std::string process_command(GdbServer& gdb_server, Task* t,
                                      std::string payload);
 
+  static GdbCommand* command_for_name(const std::string& name);
+
 private:
 };
 
+} // namespace rr
+
 #endif
