@@ -1,7 +1,5 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-//#define DEBUGTAG "GdbConnection"
-
 #define REVERSE_EXECUTION
 
 /**
@@ -19,6 +17,7 @@
 #include <limits.h>
 #include <netinet/in.h>
 #include <poll.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <sys/mman.h>
 #include <sys/socket.h>
@@ -28,11 +27,15 @@
 #include <sstream>
 #include <vector>
 
-#include "log.h"
 #include "GdbCommandHandler.h"
 #include "ReplaySession.h"
 #include "ScopedFd.h"
 #include "StringVectorToCharArray.h"
+#include "log.h"
+
+using namespace std;
+
+namespace rr {
 
 static const char INTERRUPT_CHAR = '\x03';
 
@@ -44,8 +47,6 @@ static const char INTERRUPT_CHAR = '\x03';
   LOG(info)
 #endif
 
-using namespace std;
-
 const GdbThreadId GdbThreadId::ANY(0, 0);
 const GdbThreadId GdbThreadId::ALL(-1, -1);
 
@@ -60,7 +61,7 @@ static bool request_needs_immediate_response(const GdbRequest* req) {
 }
 
 GdbConnection::GdbConnection(pid_t tgid, const Features& features)
-    : tgid(tgid), no_ack(false), inlen(0), outlen(0), features_(features) {
+    : tgid(tgid), cpu_features_(0), no_ack(false), features_(features) {
 #ifndef REVERSE_EXECUTION
   features_.reverse_execution = false;
 #endif
@@ -121,6 +122,19 @@ struct DebuggerParams {
   short port;
 };
 
+static void push_default_gdb_options(vector<string>& vec) {
+  vec.push_back("-l");
+  vec.push_back("-1");
+}
+
+static void push_target_remote_cmd(vector<string>& vec, unsigned short port) {
+  vec.push_back("-ex");
+  stringstream ss;
+  ss << "target extended-remote :";
+  ss << port;
+  vec.push_back(ss.str());
+}
+
 unique_ptr<GdbConnection> GdbConnection::await_client_connection(
     unsigned short desired_port, ProbePort probe, pid_t tgid,
     const string& exe_image, const Features& features,
@@ -137,38 +151,32 @@ unique_ptr<GdbConnection> GdbConnection::await_client_connection(
     ssize_t nwritten = write(*client_params_fd, &params, sizeof(params));
     assert(nwritten == sizeof(params));
   } else {
-    fprintf(stderr, "Launch gdb with\n"
-                    "  gdb %s\n"
-                    "and attach to the rr debug server with:\n"
-                    "  target remote :%d\n",
-            exe_image.c_str(), port);
+    vector<string> options;
+    push_default_gdb_options(options);
+    push_target_remote_cmd(options, port);
+    cerr << "Launch gdb with \n"
+         << "  gdb ";
+    for (auto& opt : options) {
+      cerr << "'" << opt << "' ";
+    }
+    cerr << exe_image << "\n";
   }
   LOG(debug) << "limiting debugger traffic to tgid " << tgid;
   dbg->await_debugger(listen_fd);
   return dbg;
 }
 
-static string create_gdb_command_file(const string& macros) {
-  char tmp[] = "/tmp/rr-gdb-commands-XXXXXX";
-  // This fd is just leaked. That's fine since we only call this once
-  // per rr invocation at the moment.
-  int fd = mkstemp(tmp);
-  unlink(tmp);
-
-  ssize_t len = macros.size();
-  int written = write(fd, macros.c_str(), len);
-  if (written != len) {
-    FATAL() << "Failed to write gdb command file";
+static string to_string(const vector<string>& args) {
+  stringstream ss;
+  for (auto& a : args) {
+    ss << "'" << a << "' ";
   }
-
-  stringstream procfile;
-  procfile << "/proc/" << getpid() << "/fd/" << fd;
-  return procfile.str();
+  return ss.str();
 }
 
-void GdbConnection::launch_gdb(ScopedFd& params_pipe_fd, const string& macros,
-                               const string& gdb_command_file_path,
-                               const string& gdb_binary_file_path) {
+void GdbConnection::launch_gdb(ScopedFd& params_pipe_fd,
+                               const string& gdb_binary_file_path,
+                               const vector<string>& gdb_options) {
   DebuggerParams params;
   ssize_t nread;
   while (true) {
@@ -183,12 +191,6 @@ void GdbConnection::launch_gdb(ScopedFd& params_pipe_fd, const string& macros,
   }
   assert(nread == sizeof(params));
 
-  stringstream attach_cmd;
-  attach_cmd << "target extended-remote " << connection_addr << ":"
-             << params.port;
-  LOG(debug) << "launching " << gdb_binary_file_path << " with command '"
-             << attach_cmd.str() << "'";
-
   vector<string> args;
   args.push_back(gdb_binary_file_path);
   // The gdb protocol uses the "vRun" packet to reload
@@ -207,20 +209,12 @@ void GdbConnection::launch_gdb(ScopedFd& params_pipe_fd, const string& macros,
   // the gdb reload models don't quite match up, we'll
   // work around it on the rr side by disabling the
   // remote-reply timeout.
-  args.push_back("-l");
-  args.push_back("-1");
+  push_default_gdb_options(args);
+  args.insert(args.end(), gdb_options.begin(), gdb_options.end());
+  push_target_remote_cmd(args, params.port);
   args.push_back(params.exe_image);
-  if (gdb_command_file_path.length() > 0) {
-    args.push_back("-x");
-    args.push_back(gdb_command_file_path);
-  }
-  if (macros.size()) {
-    string gdb_command_file = create_gdb_command_file(macros);
-    args.push_back("-x");
-    args.push_back(gdb_command_file);
-  }
-  args.push_back("-ex");
-  args.push_back(attach_cmd.str());
+
+  LOG(debug) << "launching " << to_string(args);
 
   StringVectorToCharArray c_args(args);
   execvp(gdb_binary_file_path.c_str(), c_args.get());
@@ -260,7 +254,8 @@ void GdbConnection::read_data_once() {
   /* Wait until there's data, instead of busy-looping on
    * EAGAIN. */
   poll_incoming(sock_fd, -1 /* wait forever */);
-  nread = read(sock_fd, inbuf + inlen, sizeof(inbuf) - inlen);
+  uint8_t buf[4096];
+  nread = read(sock_fd, buf, sizeof(buf));
   if (0 == nread) {
     LOG(info) << "(gdb closed debugging socket, exiting)";
     exit(0);
@@ -268,37 +263,32 @@ void GdbConnection::read_data_once() {
   if (nread <= 0) {
     FATAL() << "Error reading from gdb";
   }
-  inlen += nread;
-  assert("Impl dynamic alloc if this fails (or double inbuf size)" &&
-         inlen < int(sizeof(inbuf)));
+  inbuf.insert(inbuf.end(), buf, buf + nread);
 }
 
 void GdbConnection::write_flush() {
-  ssize_t write_index = 0;
+  size_t write_index = 0;
 
-#ifdef DEBUGTAG
-  outbuf[outlen] = '\0';
-  LOG(debug) << "write_flush: '" << outbuf << "'";
-#endif
-  while (write_index < outlen) {
+  outbuf.push_back(0);
+  LOG(debug) << "write_flush: '" << outbuf.data() << "'";
+  outbuf.pop_back();
+
+  while (write_index < outbuf.size()) {
     ssize_t nwritten;
 
     poll_outgoing(sock_fd, -1 /*wait forever*/);
-    nwritten = write(sock_fd, outbuf + write_index, outlen - write_index);
+    nwritten = write(sock_fd, outbuf.data() + write_index,
+                     outbuf.size() - write_index);
     if (nwritten < 0) {
       FATAL() << "Error writing to gdb";
     }
     write_index += nwritten;
   }
-  outlen = 0;
+  outbuf.clear();
 }
 
 void GdbConnection::write_data_raw(const uint8_t* data, ssize_t len) {
-  assert("Impl dynamic alloc if this fails (or double outbuf size)" &&
-         (outlen + len) < int(sizeof(inbuf)));
-
-  memcpy(outbuf + outlen, data, len);
-  outlen += len;
+  outbuf.insert(outbuf.end(), data, data + len);
 }
 
 void GdbConnection::write_hex(unsigned long hex) {
@@ -398,28 +388,25 @@ static string decode_ascii_encoded_hex_str(const char* encoded) {
 }
 
 bool GdbConnection::skip_to_packet_start() {
-  uint8_t* p = nullptr;
-  int i;
-
+  ssize_t end = -1;
   /* XXX we want memcspn() here ... */
-  for (i = 0; i < inlen; ++i) {
+  for (size_t i = 0; i < inbuf.size(); ++i) {
     if (inbuf[i] == '$' || inbuf[i] == INTERRUPT_CHAR) {
-      p = &inbuf[i];
+      end = i;
       break;
     }
   }
 
-  if (!p) {
+  if (end < 0) {
     /* Discard all read bytes, which we don't care
      * about. */
-    inlen = 0;
+    inbuf.clear();
     return false;
   }
   /* Discard bytes up to start-of-packet. */
-  memmove(inbuf, p, inlen - (p - inbuf));
-  inlen -= (p - inbuf);
+  inbuf.erase(inbuf.begin(), inbuf.begin() + end);
 
-  parser_assert(1 <= inlen);
+  parser_assert(1 <= inbuf.size());
   parser_assert('$' == inbuf[0] || INTERRUPT_CHAR == inbuf[0]);
   return true;
 }
@@ -429,14 +416,11 @@ bool GdbConnection::sniff_packet() {
     /* We've already seen a (possibly partial) packet. */
     return true;
   }
-  parser_assert(0 == inlen);
+  parser_assert(inbuf.empty());
   return poll_incoming(sock_fd, 0 /*don't wait*/);
 }
 
 void GdbConnection::read_packet() {
-  uint8_t* p;
-  size_t checkedlen;
-
   /* Read and discard bytes until we see the start of a
    * packet.
    *
@@ -458,16 +442,23 @@ void GdbConnection::read_packet() {
   }
 
   /* Read until we see end-of-packet. */
-  for (checkedlen = 0; !(p = (uint8_t*)memchr(inbuf + checkedlen, '#', inlen));
-       checkedlen = inlen) {
+  size_t checkedlen = 0;
+  while (true) {
+    uint8_t* p = (uint8_t*)memchr(inbuf.data() + checkedlen, '#',
+                                  inbuf.size() - checkedlen);
+    if (p) {
+      packetend = p - inbuf.data();
+      break;
+    }
+    checkedlen = inbuf.size();
     read_data_once();
   }
-  packetend = (p - inbuf);
+
   /* NB: we're ignoring the gdb packet checksums here too.  If
    * gdb is corrupted enough to garble a checksum over TCP, it's
    * not really clear why asking for the packet again might make
    * the bug go away. */
-  parser_assert('$' == inbuf[0] && packetend < inlen);
+  parser_assert('$' == inbuf[0] && packetend < inbuf.size());
 
   /* Acknowledge receipt of the packet. */
   if (!no_ack) {
@@ -517,43 +508,136 @@ static GdbThreadId parse_threadid(const char* str, char** endptr) {
   return t;
 }
 
-bool GdbConnection::xfer(const char* name, char* args) {
-  LOG(debug) << "gdb asks us to transfer " << name << "(" << args << ")";
-
-  if (!strcmp(name, "auxv")) {
-    parser_assert(!strncmp(args, "read::", sizeof("read::") - 1));
-
-    req = GdbRequest(DREQ_GET_AUXV);
-    req.target = query_thread;
-    return true;
+void GdbConnection::write_xfer_response(const void* data, size_t size,
+                                        uint64_t offset, uint64_t len) {
+  if (offset > size) {
+    write_packet("E01");
+    return;
   }
-  if (name == strstr(name, "siginfo")) {
-    if (args == strstr(args, "read")) {
-      req = GdbRequest(DREQ_READ_SIGINFO);
-      req.target = query_thread;
-      args += strlen("read");
-      parser_assert(':' == *args++);
-      parser_assert(':' == *args++);
+  if (offset == size) {
+    write_packet("l");
+    return;
+  }
+  if (offset + len < size) {
+    write_binary_packet("m", static_cast<const uint8_t*>(data) + offset, len);
+    return;
+  }
+  write_binary_packet("l", static_cast<const uint8_t*>(data) + offset,
+                      size - offset);
+}
 
-      req.mem().addr = strtoul(args, &args, 16);
-      parser_assert(',' == *args++);
-
-      req.mem().len = strtoul(args, &args, 16);
-      parser_assert('\0' == *args);
-
-      return true;
+static string read_target_desc(const char* file_name) {
+  string path = exe_directory() + "../share/rr/" + string(file_name);
+  stringstream ss;
+  FILE* f = fopen(path.c_str(), "r");
+  while (true) {
+    int ch = getc(f);
+    if (ch == EOF) {
+      break;
     }
-    if (args == strstr(args, "write")) {
-      req = GdbRequest(DREQ_WRITE_SIGINFO);
-      req.target = query_thread;
-      return true;
-    }
-    UNHANDLED_REQ() << "Unhandled 'siginfo' request: " << args;
+    ss << (char)ch;
+  }
+  fclose(f);
+  return ss.str();
+}
+
+static const char* target_description_name(uint32_t cpu_features) {
+  // This doesn't scale, but it's what gdb does...
+  switch (cpu_features) {
+    case 0:
+      return "i386-linux.xml";
+    case GdbConnection::CPU_64BIT:
+      return "amd64-linux.xml";
+    case GdbConnection::CPU_AVX:
+      return "i386-avx-linux.xml";
+    case GdbConnection::CPU_64BIT | GdbConnection::CPU_AVX:
+      return "amd64-avx-linux.xml";
+    default:
+      FATAL() << "Unknown features";
+      return nullptr;
+  }
+}
+
+bool GdbConnection::xfer(const char* name, char* args) {
+  const char* mode = args;
+  args = strchr(args, ':');
+  parser_assert(args);
+  *args++ = '\0';
+
+  if (strcmp(mode, "read") && strcmp(mode, "write")) {
+    write_packet("");
     return false;
   }
 
-  UNHANDLED_REQ() << "Unhandled gdb xfer request: " << name << "(" << args
-                  << ")";
+  const char* annex = args;
+  args = strchr(args, ':');
+  parser_assert(args);
+  *args++ = '\0';
+
+  uint64_t offset = strtoul(args, &args, 16);
+
+  uint64_t len = 0;
+  if (!strcmp(mode, "read")) {
+    parser_assert(',' == *args++);
+    len = strtoul(args, &args, 16);
+    parser_assert(!*args);
+  } else {
+    parser_assert(*args == ':');
+    ++args;
+  }
+
+  LOG(debug) << "gdb asks us to transfer " << name << " mode=" << mode
+             << ", annex=" << annex << ", offset=" << offset << " len=" << len;
+
+  if (!strcmp(name, "auxv")) {
+    if (strcmp(annex, "")) {
+      write_packet("E00");
+      return false;
+    }
+    if (strcmp(mode, "read")) {
+      write_packet("");
+      return false;
+    }
+
+    req = GdbRequest(DREQ_GET_AUXV);
+    req.target = query_thread;
+    // XXX handle offset/len here!
+    return true;
+  }
+
+  if (!strcmp(name, "siginfo")) {
+    if (strcmp(annex, "")) {
+      write_packet("E00");
+      return false;
+    }
+    if (!strcmp(mode, "read")) {
+      req = GdbRequest(DREQ_READ_SIGINFO);
+      req.target = query_thread;
+      req.mem().addr = offset;
+      req.mem().len = len;
+      return true;
+    }
+
+    req = GdbRequest(DREQ_WRITE_SIGINFO);
+    req.target = query_thread;
+    return true;
+  }
+
+  if (!strcmp(name, "features")) {
+    if (strcmp(mode, "read")) {
+      write_packet("");
+      return false;
+    }
+
+    string target_desc =
+        read_target_desc((strcmp(annex, "") && strcmp(annex, "target.xml"))
+                             ? annex
+                             : target_description_name(cpu_features_));
+    write_xfer_response(target_desc.c_str(), target_desc.size(), offset, len);
+    return false;
+  }
+
+  write_packet("");
   return false;
 }
 
@@ -667,8 +751,10 @@ bool GdbConnection::query(char* payload) {
     LOG(debug) << "gdb supports " << args;
 
     stringstream supported;
-    supported << "PacketSize=" << sizeof(outbuf);
-    supported << ";QStartNoAckMode+"
+    // Encourage gdb to use very large packets since we support any packet size
+    supported << "PacketSize=1048576"
+                 ";QStartNoAckMode+"
+                 ";qXfer:features:read+"
                  ";qXfer:auxv:read+"
                  ";qXfer:siginfo:read+"
                  ";qXfer:siginfo:write+"
@@ -710,9 +796,9 @@ bool GdbConnection::query(char* payload) {
   if (!strcmp(name, "Xfer")) {
     name = args;
     args = strchr(args, ':');
-    if (args) {
-      *args++ = '\0';
-    }
+    parser_assert(args);
+    *args++ = '\0';
+
     return xfer(name, args);
   }
   if (!strcmp(name, "Search")) {
@@ -728,7 +814,8 @@ bool GdbConnection::query(char* payload) {
       parser_assert(';' == *args++);
       req.mem().len = strtoul(args, &args, 16);
       parser_assert(';' == *args++);
-      read_binary_data((const uint8_t*)args, inbuf + packetend, req.mem().data);
+      read_binary_data((const uint8_t*)args, inbuf.data() + packetend,
+                       req.mem().data);
 
       LOG(debug) << "gdb searching memory (addr=" << HEX(req.mem().addr)
                  << ", len=" << req.mem().len << ")";
@@ -943,31 +1030,40 @@ bool GdbConnection::process_vpacket(char* payload) {
   return false;
 }
 
-bool GdbConnection::process_packet() {
-  char request;
-  char* payload = nullptr;
-  bool ret;
+static string to_string(const vector<uint8_t>& bytes, size_t max_len) {
+  stringstream ss;
+  for (size_t i = 0; i < bytes.size(); ++i) {
+    if (i >= max_len) {
+      ss << "...";
+      break;
+    }
+    char buf[3];
+    sprintf(buf, "%02x", bytes[i]);
+    ss << buf;
+  }
+  return ss.str();
+}
 
+bool GdbConnection::process_packet() {
   parser_assert(INTERRUPT_CHAR == inbuf[0] ||
                 ('$' == inbuf[0] &&
-                 (((uint8_t*)memchr(inbuf, '#', inlen) - inbuf) == packetend)));
+                 (uint8_t*)memchr(inbuf.data(), '#', inbuf.size()) ==
+                     inbuf.data() + packetend));
 
   if (INTERRUPT_CHAR == inbuf[0]) {
-    request = INTERRUPT_CHAR;
-  } else {
-    request = inbuf[1];
-    payload = (char*)&inbuf[2];
-    inbuf[packetend] = '\0';
+    LOG(debug) << "gdb requests interrupt";
+    req = GdbRequest(DREQ_INTERRUPT);
+    inbuf.erase(inbuf.begin());
+    return true;
   }
 
+  char request = inbuf[1];
+  char* payload = (char*)&inbuf[2];
+  inbuf[packetend] = '\0';
   LOG(debug) << "raw request " << request << payload;
 
+  bool ret;
   switch (request) {
-    case INTERRUPT_CHAR:
-      LOG(debug) << "gdb requests interrupt";
-      req = GdbRequest(DREQ_INTERRUPT);
-      ret = true;
-      break;
     case 'b':
       ret = process_bpacket(payload);
       break;
@@ -1071,12 +1167,13 @@ bool GdbConnection::process_packet() {
       parser_assert(',' == *payload++);
       req.mem().len = strtoul(payload, &payload, 16);
       parser_assert(':' == *payload++);
-      read_binary_data((const uint8_t*)payload, inbuf + packetend,
+      read_binary_data((const uint8_t*)payload, inbuf.data() + packetend,
                        req.mem().data);
       parser_assert(req.mem().len == req.mem().data.size());
 
       LOG(debug) << "gdb setting memory (addr=" << HEX(req.mem().addr)
-                 << ", len=" << req.mem().len << ")";
+                 << ", len=" << req.mem().len
+                 << ", data=" << to_string(req.mem().data, 32) << ")";
 
       ret = true;
       break;
@@ -1140,9 +1237,10 @@ bool GdbConnection::process_packet() {
       UNHANDLED_REQ() << "Unhandled gdb request '" << inbuf[1] << "'";
       ret = false;
   }
-  /* Erase the newly processed packet from the input buffer. */
-  memmove(inbuf, inbuf + packetend, inlen - packetend);
-  inlen = (inlen - packetend);
+  /* Erase the newly processed packet from the input buffer. The checksum
+   * after the '#' will be skipped later as we look for the next packet start.
+   */
+  inbuf.erase(inbuf.begin(), inbuf.begin() + packetend + 1);
 
   /* If we processed the request internally, consume it. */
   if (!ret) {
@@ -1492,15 +1590,14 @@ void GdbConnection::reply_get_reg(const GdbRegisterValue& reg) {
   consume_request();
 }
 
-void GdbConnection::reply_get_regs(const GdbRegisterFile& file) {
-  size_t n_regs = file.total_registers();
-  char buf[n_regs * 2 * GdbRegisterValue::MAX_SIZE + 1];
+void GdbConnection::reply_get_regs(const vector<GdbRegisterValue>& file) {
+  char buf[file.size() * 2 * GdbRegisterValue::MAX_SIZE + 1];
 
   assert(DREQ_GET_REGS == req.type);
 
   size_t offset = 0;
-  for (auto it = file.regs.begin(), end = file.regs.end(); it != end; ++it) {
-    offset += print_reg_value(*it, &buf[offset]);
+  for (auto& reg : file) {
+    offset += print_reg_value(reg, &buf[offset]);
   }
   write_packet(buf);
 
@@ -1606,3 +1703,5 @@ void GdbConnection::reply_rr_cmd(const std::string& text) {
 
   consume_request();
 }
+
+} // namespace rr
