@@ -16,6 +16,8 @@
 #include "ReplaySession.h"
 #include "ReplayTimeline.h"
 
+namespace rr {
+
 /**
  * Descriptor for task within a task group.  Note: on linux, we can
  * uniquely identify any thread by its |tid| (ignoring pid
@@ -58,18 +60,6 @@ struct GdbRegisterValue {
   bool defined;
 };
 
-/**
- * Represents the register file, indexed by |DbgRegister| values
- * above.
- */
-struct GdbRegisterFile {
-  std::vector<GdbRegisterValue> regs;
-
-  GdbRegisterFile(size_t n_regs) : regs(n_regs){};
-
-  size_t total_registers() const { return regs.size(); }
-};
-
 enum GdbRequestType {
   DREQ_NONE = 0,
 
@@ -301,9 +291,9 @@ public:
    * |params_pipe_fd|.  Optionally, pre-define in the gdb client the set
    * of macros defined in |macros| if nonnull.
    */
-  static void launch_gdb(ScopedFd& params_pipe_fd, const std::string& macros,
-                         const std::string& gdb_command_file_path,
-                         const std::string& gdb_binary_file_path);
+  static void launch_gdb(ScopedFd& params_pipe_fd,
+                         const std::string& gdb_binary_file_path,
+                         const std::vector<std::string>& gdb_options);
 
   /**
    * Call this when the target of |req| is needed to fulfill the
@@ -412,7 +402,7 @@ public:
    * Send |file| back to the debugger host.  |file| may contain
    * undefined register values.
    */
-  void reply_get_regs(const GdbRegisterFile& file);
+  void reply_get_regs(const std::vector<GdbRegisterValue>& file);
 
   /**
    * Pass |ok = true| iff the requested register was successfully set.
@@ -488,6 +478,13 @@ public:
 
   const Features& features() { return features_; }
 
+  enum {
+    CPU_64BIT = 0x1,
+    CPU_AVX = 0x2,
+  };
+  void set_cpu_features(uint32_t features) { cpu_features_ = features; }
+  uint32_t cpu_features() const { return cpu_features_; }
+
 private:
   GdbConnection(pid_t tgid, const Features& features);
 
@@ -512,6 +509,8 @@ private:
   void write_binary_packet(const char* pfx, const uint8_t* data,
                            ssize_t num_bytes);
   void write_hex_bytes_packet(const uint8_t* bytes, size_t len);
+  void write_xfer_response(const void* data, size_t size, uint64_t offset,
+                           uint64_t len);
   /**
    * Consume bytes in the input buffer until start-of-packet ('$') or
    * the interrupt character is seen.  Does not block.  Return true if
@@ -571,17 +570,17 @@ private:
   // multi-exe-image debugging scenarios, so we pretend only
   // this task group exists when interfacing with gdb
   pid_t tgid;
+  uint32_t cpu_features_;
   // true when "no-ack mode" enabled, in which we don't have
   // to send ack packets back to gdb.  This is a huge perf win.
   bool no_ack;
   ScopedFd sock_fd;
-  /* XXX probably need to dynamically size these */
-  uint8_t inbuf[32768];  /* buffered input from gdb */
-  ssize_t inlen;         /* length of valid data */
-  ssize_t packetend;     /* index of '#' character */
-  uint8_t outbuf[32768]; /* buffered output for gdb */
-  ssize_t outlen;
+  std::vector<uint8_t> inbuf;  /* buffered input from gdb */
+  size_t packetend;            /* index of '#' character */
+  std::vector<uint8_t> outbuf; /* buffered output for gdb */
   Features features_;
 };
 
+} // namespace rr
+
 #endif /* RR_GDB_CONNECTION_H_ */
