@@ -8,6 +8,8 @@
 
 #include <vector>
 
+namespace rr {
+
 class Task;
 
 /**
@@ -39,4 +41,6 @@ private:
   std::vector<std::vector<uint8_t> > bytecode_variants;
 };
 
+} // namespace rr
+
 #endif // RR_GDB_EXPRESSION_H_
