@@ -1,7 +1,5 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-//#define DEBUGTAG "GdbServer"
-
 #include "GdbServer.h"
 
 #include <assert.h>
@@ -12,47 +10,35 @@
 
 #include <limits>
 #include <map>
-#include <string>
 #include <sstream>
+#include <string>
 #include <vector>
 
 #include "BreakpointCondition.h"
 #include "GdbCommandHandler.h"
 #include "GdbExpression.h"
-#include "kernel_metadata.h"
-#include "log.h"
 #include "ReplaySession.h"
 #include "ScopedFd.h"
-#include "task.h"
+#include "Task.h"
+#include "TaskGroup.h"
+#include "kernel_metadata.h"
+#include "log.h"
 #include "util.h"
 
-using namespace rr;
 using namespace std;
 
-/**
- * 32-bit writes to DBG_COMMAND_MAGIC_ADDRESS by the debugger trigger
- * rr commands.
- */
-static const uint32_t DBG_COMMAND_MAGIC_ADDRESS = 29298; // 'rr'
+namespace rr {
 
-/**
- * The high-order byte of the 32-bit value indicates the specific command
- * message. Not-understood command messages are ignored.
- */
-static const uint32_t DBG_COMMAND_MSG_MASK = 0xFF000000;
-/**
- * Create a checkpoint of the current state whose index is given by the
- * command parameter. If there is already a checkpoint with that index, it
- * is deleted first.
- */
-static const int DBG_COMMAND_MSG_CREATE_CHECKPOINT = 0x01000000;
-/**
- * Delete the checkpoint of the current state whose index is given by the
- * command parameter.
- */
-static const int DBG_COMMAND_MSG_DELETE_CHECKPOINT = 0x02000000;
-
-static const uint32_t DBG_COMMAND_PARAMETER_MASK = 0x00FFFFFF;
+GdbServer::GdbServer(std::unique_ptr<GdbConnection>& dbg, Task* t)
+    : dbg(std::move(dbg)),
+      debuggee_tguid(t->task_group()->tguid()),
+      last_continue_tuid(t->tuid()),
+      last_query_tuid(t->tuid()),
+      final_event(UINT32_MAX),
+      stop_reason(0),
+      stop_replaying_to_target(false),
+      interrupt_pending(false),
+      emergency_debug_session(&t->session()) {}
 
 // Special-sauce macros defined by rr when launching the gdb client,
 // which implement functionality outside of the gdb remote protocol.
@@ -62,78 +48,80 @@ static const string& gdb_rr_macros() {
 
   if (s.empty()) {
     stringstream ss;
-    ss
-        // TODO define `document' sections for these
-        << "define checkpoint\n"
-        << "  init-if-undefined $_next_checkpoint_index = 1\n"
-        /* Ensure the command echoes the checkpoint number, not the encoded
-         * message
-           */
-        << "  p (*(int*)" << DBG_COMMAND_MAGIC_ADDRESS << " = "
-        << DBG_COMMAND_MSG_CREATE_CHECKPOINT << " | $_next_checkpoint_index), "
-        << "$_next_checkpoint_index++\n"
-        << "end\n"
-        << "define delete checkpoint\n"
-        << "  p (*(int*)" << DBG_COMMAND_MAGIC_ADDRESS << " = "
-        << DBG_COMMAND_MSG_DELETE_CHECKPOINT << " | $arg0), $arg0\n"
-        << "end\n"
-        << "define restart\n"
-        << "  run c$arg0\n"
-        << "end\n"
-        // In gdb version "Fedora 7.8.1-30.fc21", a raw "run" command
-        // issued before any user-generated resume-execution command
-        // results in gdb hanging just after the inferior hits an internal
-        // gdb breakpoint.  This happens outside of rr, with gdb
-        // controlling gdbserver, as well.  We work around that by
-        // ensuring *some* resume-execution command has been issued before
-        // restarting the session.  But, only if the inferior hasn't
-        // already finished execution ($_thread != 0).  If it has and we
-        // issue the "stepi" command, then gdb refuses to restart
-        // execution.
-        << "define hook-run\n"
-        << "  if $_thread != 0 && !$suppress_run_hook\n"
-        << "    stepi\n"
-        << "  end\n"
-        << "end\n"
-        << "define hookpost-continue\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-step\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-stepi\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-next\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-nexti\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-finish\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-reverse-continue\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-reverse-step\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-reverse-stepi\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-reverse-finish\n"
-        << "  set $suppress_run_hook = 1\n"
-        << "end\n"
-        << "define hookpost-run\n"
-        << "  set $suppress_run_hook = 0\n"
-        << "end\n"
-        << "handle SIGURG stop\n"
-        << "set prompt (rr) \n" << GdbCommandHandler::gdb_macros()
-        // Try both "set target-async" and "maint set target-async" since
-        // that changed recently.
-        << "set target-async 0\n"
-        << "maint set target-async 0\n";
+    ss << "define restart\n"
+       << "  run c$arg0\n"
+       << "end\n"
+       // In gdb version "Fedora 7.8.1-30.fc21", a raw "run" command
+       // issued before any user-generated resume-execution command
+       // results in gdb hanging just after the inferior hits an internal
+       // gdb breakpoint.  This happens outside of rr, with gdb
+       // controlling gdbserver, as well.  We work around that by
+       // ensuring *some* resume-execution command has been issued before
+       // restarting the session.  But, only if the inferior hasn't
+       // already finished execution ($_thread != 0).  If it has and we
+       // issue the "stepi" command, then gdb refuses to restart
+       // execution.
+       << "define hook-run\n"
+       << "  if $_thread != 0 && !$suppress_run_hook\n"
+       << "    stepi\n"
+       << "  end\n"
+       << "end\n"
+       << "define hookpost-continue\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-step\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-stepi\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-next\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-nexti\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-finish\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-reverse-continue\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-reverse-step\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-reverse-stepi\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-reverse-finish\n"
+       << "  set $suppress_run_hook = 1\n"
+       << "end\n"
+       << "define hookpost-run\n"
+       << "  set $suppress_run_hook = 0\n"
+       << "end\n"
+       << "set unwindonsignal on\n"
+       << "handle SIGURG stop\n"
+       << "set prompt (rr) \n"
+       << GdbCommandHandler::gdb_macros()
+       // Try both "set target-async" and "maint set target-async" since
+       // that changed recently.
+       << "python-interactive\n"
+       << "import re\n"
+       << "m = re.compile("
+       << "'.* ([0-9]+)\\.([0-9]+)(\\.([0-9]+))?.*'"
+       << ").match(gdb.execute('show version', False, True))\n"
+       << "ver = int(m.group(1))*10000 + int(m.group(2))*100\n"
+       << "if m.group(4):\n"
+       << "    ver = ver + int(m.group(4))\n"
+       << "\n"
+       << "if ver == 71100:\n"
+       << "    gdb.write("
+       << "'This version of gdb (7.11.0) has known bugs that break rr. "
+       << "Install 7.11.1 or later.\\n', gdb.STDERR)\n"
+       << "\n"
+       << "if ver < 71101:\n"
+       << "    gdb.execute('set target-async 0')\n"
+       << "    gdb.execute('maint set target-async 0')\n";
     s = ss.str();
   }
   return s;
@@ -170,7 +158,7 @@ GdbRegisterValue GdbServer::get_reg(const Registers& regs,
   GdbRegisterValue reg;
   memset(&reg, 0, sizeof(reg));
   reg.name = which;
-  reg.size = ::get_reg(regs, extra_regs, &reg.value[0], which, &reg.defined);
+  reg.size = rr::get_reg(regs, extra_regs, &reg.value[0], which, &reg.defined);
   return reg;
 }
 
@@ -214,56 +202,44 @@ static WatchType watchpoint_type(GdbRequestType req) {
 }
 
 static void maybe_singlestep_for_event(Task* t, GdbRequest* req) {
+  if (!t->session().is_replaying()) {
+    return;
+  }
+  auto rt = static_cast<ReplayTask*>(t);
   if (trace_instructions_up_to_event(
-          t->replay_session().current_trace_frame().time())) {
+          rt->session().current_trace_frame().time())) {
     fputs("Stepping: ", stderr);
     t->regs().print_register_file_compact(stderr);
     fprintf(stderr, " ticks:%" PRId64 "\n", t->tick_count());
     *req = GdbRequest(DREQ_CONT);
     req->suppress_debugger_stop = true;
-    req->cont().actions.push_back(GdbContAction(
-        ACTION_STEP, get_threadid(t->replay_session(), t->tuid())));
+    req->cont().actions.push_back(
+        GdbContAction(ACTION_STEP, get_threadid(t->session(), t->tuid())));
   }
 }
 
-bool GdbServer::maybe_process_magic_command(const GdbRequest& req) {
-  if (!(req.mem().addr == DBG_COMMAND_MAGIC_ADDRESS && req.mem().len == 4)) {
-    return false;
-  }
-  uint32_t cmd;
-  memcpy(&cmd, req.mem().data.data(), sizeof(cmd));
-  uintptr_t param = cmd & DBG_COMMAND_PARAMETER_MASK;
-  switch (cmd & DBG_COMMAND_MSG_MASK) {
-    case DBG_COMMAND_MSG_CREATE_CHECKPOINT: {
-      if (timeline.can_add_checkpoint()) {
-        checkpoints[param] =
-            Checkpoint(timeline, last_continue_tuid, Checkpoint::EXPLICIT);
-      }
-      break;
-    }
-    case DBG_COMMAND_MSG_DELETE_CHECKPOINT: {
-      auto it = checkpoints.find(param);
-      if (it != checkpoints.end()) {
-        timeline.remove_explicit_checkpoint(it->second.mark);
-        checkpoints.erase(it);
-      }
-      break;
-    }
-    default:
-      return false;
-  }
-  dbg->reply_set_mem(true);
-  return true;
-}
-
 void GdbServer::dispatch_regs_request(const Registers& regs,
                                       const ExtraRegisters& extra_regs) {
-  size_t n_regs = regs.total_registers();
-  GdbRegisterFile file(n_regs);
-  for (size_t i = 0; i < n_regs; ++i) {
-    file.regs[i] = get_reg(regs, extra_regs, GdbRegister(i));
+  GdbRegister end;
+  // Send values for all the registers we sent XML register descriptions for.
+  // Those descriptions are controlled by GdbConnection::cpu_features().
+  bool have_AVX = dbg->cpu_features() & GdbConnection::CPU_AVX;
+  switch (regs.arch()) {
+    case x86:
+      end = have_AVX ? DREG_YMM7H : DREG_ORIG_EAX;
+      break;
+    case x86_64:
+      end = have_AVX ? DREG_64_YMM15H : DREG_ORIG_RAX;
+      break;
+    default:
+      FATAL() << "Unknown architecture";
+      break;
   }
-  dbg->reply_get_regs(file);
+  vector<GdbRegisterValue> rs;
+  for (GdbRegister r = GdbRegister(0); r < end; r = GdbRegister(r + 1)) {
+    rs.push_back(get_reg(regs, extra_regs, r));
+  }
+  dbg->reply_get_regs(rs);
 }
 
 class GdbBreakpointCondition : public BreakpointCondition {
@@ -435,9 +411,6 @@ void GdbServer::dispatch_debugger_request(Session& session,
         dbg->reply_set_mem(true);
         return;
       }
-      if (maybe_process_magic_command(req)) {
-        return;
-      }
       // We only allow the debugger to write memory if the
       // memory will be written to an diversion session.
       // Arbitrary writes to replay sessions cause
@@ -507,12 +480,13 @@ void GdbServer::dispatch_debugger_request(Session& session,
           << "Debugger setting bad breakpoint insn";
       // Mirror all breakpoint/watchpoint sets/unsets to the target process
       // if it's not part of the timeline (i.e. it's a diversion).
-      Task* replay_task = timeline.current_session().find_task(target->tuid());
+      ReplayTask* replay_task =
+          timeline.current_session().find_task(target->tuid());
       bool ok = timeline.add_breakpoint(replay_task, req.watch().addr,
                                         breakpoint_condition(req));
       if (ok && &session != &timeline.current_session()) {
         bool diversion_ok =
-            target->vm()->add_breakpoint(req.watch().addr, TRAP_BKPT_USER);
+            target->vm()->add_breakpoint(req.watch().addr, BKPT_USER);
         ASSERT(target, diversion_ok);
       }
       dbg->reply_watchpoint_request(ok);
@@ -522,7 +496,8 @@ void GdbServer::dispatch_debugger_request(Session& session,
     case DREQ_SET_RD_WATCH:
     case DREQ_SET_WR_WATCH:
     case DREQ_SET_RDWR_WATCH: {
-      Task* replay_task = timeline.current_session().find_task(target->tuid());
+      ReplayTask* replay_task =
+          timeline.current_session().find_task(target->tuid());
       bool ok = timeline.add_watchpoint(
           replay_task, req.watch().addr, req.watch().kind,
           watchpoint_type(req.type), breakpoint_condition(req));
@@ -535,10 +510,11 @@ void GdbServer::dispatch_debugger_request(Session& session,
       return;
     }
     case DREQ_REMOVE_SW_BREAK: {
-      Task* replay_task = timeline.current_session().find_task(target->tuid());
+      ReplayTask* replay_task =
+          timeline.current_session().find_task(target->tuid());
       timeline.remove_breakpoint(replay_task, req.watch().addr);
       if (&session != &timeline.current_session()) {
-        target->vm()->remove_breakpoint(req.watch().addr, TRAP_BKPT_USER);
+        target->vm()->remove_breakpoint(req.watch().addr, BKPT_USER);
       }
       dbg->reply_watchpoint_request(true);
       return;
@@ -547,7 +523,8 @@ void GdbServer::dispatch_debugger_request(Session& session,
     case DREQ_REMOVE_RD_WATCH:
     case DREQ_REMOVE_WR_WATCH:
     case DREQ_REMOVE_RDWR_WATCH: {
-      Task* replay_task = timeline.current_session().find_task(target->tuid());
+      ReplayTask* replay_task =
+          timeline.current_session().find_task(target->tuid());
       timeline.remove_watchpoint(replay_task, req.watch().addr,
                                  req.watch().kind, watchpoint_type(req.type));
       if (&session != &timeline.current_session()) {
@@ -587,6 +564,7 @@ bool GdbServer::diverter_process_debugger_requests(
     switch (req->type) {
       case DREQ_RESTART:
       case DREQ_DETACH:
+      case DREQ_RR_CMD:
         diversion_refcount = 0;
         return false;
 
@@ -870,7 +848,7 @@ void GdbServer::try_lazy_reverse_singlesteps(GdbRequest& req) {
 
   ReplayTimeline::Mark now;
   bool need_seek = false;
-  Task* t = timeline.current_session().current_task();
+  ReplayTask* t = timeline.current_session().current_task();
   while (t && req.type == DREQ_CONT &&
          req.cont().run_direction == RUN_BACKWARD &&
          req.cont().actions.size() == 1 &&
@@ -990,7 +968,6 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
     }
   }
 
-  auto interrupt_check = [&]() { return dbg->sniff_packet(); };
   if (req.cont().run_direction == RUN_FORWARD) {
     if (is_in_exec(timeline) &&
         timeline.current_session().current_task()->task_group()->tguid() ==
@@ -1003,8 +980,7 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
       RunCommand command = compute_run_command_from_actions(
           timeline.current_session().current_task(), req, &signal_to_deliver);
       // Ignore gdb's |signal_to_deliver|; we just have to follow the replay.
-      result =
-          timeline.replay_step_forward(command, target.event, interrupt_check);
+      result = timeline.replay_step_forward(command, target.event);
     }
     if (result.status == REPLAY_EXITED) {
       return handle_exited_state(last_resume_request);
@@ -1030,6 +1006,7 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
       return false;
     };
 
+    auto interrupt_check = [&]() { return dbg->sniff_packet(); };
     switch (command) {
       case RUN_CONTINUE:
         result = timeline.reverse_continue(stop_filter, interrupt_check);
@@ -1063,7 +1040,7 @@ GdbServer::ContinueOrStop GdbServer::debug_one_step(
 bool GdbServer::at_target() {
   // Don't launch the debugger for the initial rr fork child.
   // No one ever wants that to happen.
-  if (!timeline.current_session().can_validate()) {
+  if (!timeline.current_session().done_initial_exec()) {
     return false;
   }
   Task* t = timeline.current_session().current_task();
@@ -1130,12 +1107,13 @@ void GdbServer::activate_debugger() {
   // session.  The cloned tasks will look like children
   // of the clonees, so this scheme prevents |pstree|
   // output from getting /too/ far out of whack.
+  const char* where = "???";
   if (timeline.can_add_checkpoint()) {
     debugger_restart_checkpoint =
-        Checkpoint(timeline, last_continue_tuid, Checkpoint::EXPLICIT);
+        Checkpoint(timeline, last_continue_tuid, Checkpoint::EXPLICIT, where);
   } else {
-    debugger_restart_checkpoint =
-        Checkpoint(timeline, last_continue_tuid, Checkpoint::NOT_EXPLICIT);
+    debugger_restart_checkpoint = Checkpoint(timeline, last_continue_tuid,
+                                             Checkpoint::NOT_EXPLICIT, where);
   }
 }
 
@@ -1186,9 +1164,7 @@ void GdbServer::restart_session(const GdbRequest& req) {
   // Note that we don't reset the target pid; we intentionally keep targeting
   // the same process no matter what is running when we hit the event.
   target.event = req.restart().param;
-  if (final_event >= 0) {
-    target.event = min(final_event - 1, target.event);
-  }
+  target.event = min(final_event - 1, target.event);
   timeline.seek_to_before_event(target.event);
   do {
     ReplayResult result =
@@ -1206,6 +1182,44 @@ void GdbServer::restart_session(const GdbRequest& req) {
   activate_debugger();
 }
 
+static uint32_t get_cpu_features(SupportedArch arch) {
+  uint32_t cpu_features;
+  switch (arch) {
+    case x86:
+      cpu_features = 0;
+      break;
+    case x86_64:
+      cpu_features = GdbConnection::CPU_64BIT;
+      break;
+    default:
+      FATAL() << "Unknown architecture";
+      return 0;
+  }
+
+  unsigned int AVX_cpuid_flags = AVX_FEATURE_FLAG | OSXSAVE_FEATURE_FLAG;
+  auto cpuid_data = cpuid(CPUID_GETFEATURES, 0);
+  // We're assuming here that AVX support on the system making the recording
+  // is the same as the AVX support during replay. But if that's not true,
+  // rr is totally broken anyway.
+  if ((cpuid_data.ecx & AVX_cpuid_flags) == AVX_cpuid_flags) {
+    cpu_features |= GdbConnection::CPU_AVX;
+  }
+
+  return cpu_features;
+}
+
+static unique_ptr<GdbConnection> await_connection(
+    Task* t, unsigned short port, GdbConnection::ProbePort probe,
+    const GdbConnection::Features& features,
+    ScopedFd* client_params_fd = nullptr) {
+  auto result = GdbConnection::await_client_connection(
+      port, probe, t->tgid(), t->vm()->exe_image(), features, client_params_fd);
+
+  result->set_cpu_features(get_cpu_features(t->arch()));
+
+  return result;
+}
+
 void GdbServer::serve_replay(const ConnectionFlags& flags) {
   do {
     ReplayResult result =
@@ -1225,9 +1239,8 @@ void GdbServer::serve_replay(const ConnectionFlags& flags) {
   auto probe = flags.dbg_port > 0 ? GdbConnection::DONT_PROBE
                                   : GdbConnection::PROBE_PORT;
   Task* t = timeline.current_session().current_task();
-  dbg = GdbConnection::await_client_connection(
-      port, probe, t->tgid(), t->vm()->exe_image(), GdbConnection::Features(),
-      flags.debugger_params_write_pipe);
+  dbg = await_connection(t, port, probe, GdbConnection::Features(),
+                         flags.debugger_params_write_pipe);
   if (flags.debugger_params_write_pipe) {
     flags.debugger_params_write_pipe->close();
   }
@@ -1247,11 +1260,33 @@ void GdbServer::serve_replay(const ConnectionFlags& flags) {
   LOG(debug) << "debugger server exiting ...";
 }
 
+static string create_gdb_command_file(const string& macros) {
+  char tmp[] = "/tmp/rr-gdb-commands-XXXXXX";
+  // This fd is just leaked. That's fine since we only call this once
+  // per rr invocation at the moment.
+  int fd = mkstemp(tmp);
+  unlink(tmp);
+
+  ssize_t len = macros.size();
+  int written = write(fd, macros.c_str(), len);
+  if (written != len) {
+    FATAL() << "Failed to write gdb command file";
+  }
+
+  stringstream procfile;
+  procfile << "/proc/" << getpid() << "/fd/" << fd;
+  return procfile.str();
+}
+
 void GdbServer::launch_gdb(ScopedFd& params_pipe_fd,
-                           const string& gdb_command_file_path,
-                           const string& gdb_binary_file_path) {
-  GdbConnection::launch_gdb(params_pipe_fd, gdb_rr_macros(),
-                            gdb_command_file_path, gdb_binary_file_path);
+                           const string& gdb_binary_file_path,
+                           const vector<string>& gdb_options) {
+  auto macros = gdb_rr_macros();
+  string gdb_command_file = create_gdb_command_file(macros);
+  vector<string> options = gdb_options;
+  options.push_back("-x");
+  options.push_back(gdb_command_file);
+  GdbConnection::launch_gdb(params_pipe_fd, gdb_binary_file_path, options);
 }
 
 void GdbServer::emergency_debug(Task* t) {
@@ -1272,11 +1307,12 @@ void GdbServer::emergency_debug(Task* t) {
   // b) some gdb versions will fail if the user doesn't turn off async
   // mode (and we don't want to require users to do that)
   features.reverse_execution = false;
-  unique_ptr<GdbConnection> dbg = GdbConnection::await_client_connection(
-      t->tid, GdbConnection::PROBE_PORT, t->tgid(), t->vm()->exe_image(),
-      features);
+  unique_ptr<GdbConnection> dbg =
+      await_connection(t, t->tid, GdbConnection::PROBE_PORT, features);
 
   GdbServer(dbg, t).process_debugger_requests();
 }
 
 string GdbServer::init_script() { return gdb_rr_macros(); }
+
+} // namespace rr
