@@ -14,8 +14,16 @@
 #include "ScopedFd.h"
 #include "TraceFrame.h"
 
+namespace rr {
+
 class GdbServer {
-  friend class GdbCommand;
+  // Not ideal but we can't inherit friend from GdbCommand
+  friend std::string invoke_checkpoint(GdbServer&, Task*,
+                                       const std::vector<std::string>&);
+  friend std::string invoke_delete_checkpoint(GdbServer&, Task*,
+                                              const std::vector<std::string>&);
+  friend std::string invoke_info_checkpoints(GdbServer&, Task*,
+                                             const std::vector<std::string>&);
 
 public:
   struct Target {
@@ -46,7 +54,7 @@ public:
   GdbServer(std::shared_ptr<ReplaySession> session,
             const ReplaySession::Flags& flags, const Target& target)
       : target(target),
-        final_event(-1),
+        final_event(UINT32_MAX),
         stop_reason(0),
         in_debuggee_end_state(false),
         stop_replaying_to_target(false),
@@ -64,8 +72,8 @@ public:
    * the pipe passed to serve_replay_with_debugger).
    */
   static void launch_gdb(ScopedFd& params_pipe_fd,
-                         const std::string& gdb_command_file_path,
-                         const std::string& gdb_binary_file_path);
+                         const std::string& gdb_binary_file_path,
+                         const std::vector<std::string>& gdb_options);
 
   /**
    * Start a debugging connection for |t| and return when there are no
@@ -97,27 +105,13 @@ public:
                                   GdbRegister which);
 
 private:
-  GdbServer(std::unique_ptr<GdbConnection>& dbg, Task* t)
-      : dbg(std::move(dbg)),
-        debuggee_tguid(t->task_group()->tguid()),
-        last_continue_tuid(t->tuid()),
-        last_query_tuid(t->tuid()),
-        final_event(-1),
-        stop_reason(0),
-        stop_replaying_to_target(false),
-        interrupt_pending(false),
-        emergency_debug_session(&t->session()) {}
+  GdbServer(std::unique_ptr<GdbConnection>& dbg, Task* t);
 
   Session& current_session() {
     return timeline.is_running() ? timeline.current_session()
                                  : *emergency_debug_session;
   }
 
-  /**
-   * If |req| is a magic-write command, interpret it and return true.
-   * Otherwise, do nothing and return false.
-   */
-  bool maybe_process_magic_command(const GdbRequest& req);
   void dispatch_regs_request(const Registers& regs,
                              const ExtraRegisters& extra_regs);
   enum ReportState { REPORT_NORMAL, REPORT_THREADS_DEAD };
@@ -219,8 +213,9 @@ private:
 
   struct Checkpoint {
     enum Explicit { EXPLICIT, NOT_EXPLICIT };
-    Checkpoint(ReplayTimeline& timeline, TaskUid last_continue_tuid, Explicit e)
-        : last_continue_tuid(last_continue_tuid), is_explicit(e) {
+    Checkpoint(ReplayTimeline& timeline, TaskUid last_continue_tuid, Explicit e,
+               const std::string& where)
+        : last_continue_tuid(last_continue_tuid), is_explicit(e), where(where) {
       if (e == EXPLICIT) {
         mark = timeline.add_explicit_checkpoint();
       } else {
@@ -231,6 +226,7 @@ private:
     ReplayTimeline::Mark mark;
     TaskUid last_continue_tuid;
     Explicit is_explicit;
+    std::string where;
   };
   // |debugger_restart_mark| is the point where we will restart from with
   // a no-op debugger "run" command.
@@ -240,4 +236,6 @@ private:
   std::map<int, Checkpoint> checkpoints;
 };
 
+} // namespace rr
+
 #endif /* RR_GDB_SERVER_H_ */
