@@ -5,6 +5,8 @@
 
 #include "FileMonitor.h"
 
+namespace rr {
+
 /**
  * A FileMonitor to track writes to RR_MAGIC_SAVE_DATA_FD.
  */
@@ -16,7 +18,10 @@ public:
    * During recording, record the written data.
    * During replay, check that the written data matches what was recorded.
    */
-  virtual void did_write(Task* t, const std::vector<Range>& ranges);
+  virtual void did_write(Task* t, const std::vector<Range>& ranges,
+                         int64_t offset);
 };
 
+} // namespace rr
+
 #endif /* RR_MAGIC_SAVE_DATA_MONITOR_H_ */
