@@ -16,12 +16,15 @@
 #include <algorithm>
 #include <string>
 
+#include "Flags.h"
 #include "kernel_metadata.h"
 #include "log.h"
 #include "util.h"
 
 using namespace std;
 
+namespace rr {
+
 static bool attributes_initialized;
 static struct perf_event_attr ticks_attr;
 static struct perf_event_attr page_faults_attr;
@@ -94,9 +97,8 @@ static CpuMicroarch get_cpu_microarch() {
     FATAL() << "Forced uarch " << Flags::get().forced_uarch << " isn't known.";
   }
 
-  unsigned int cpu_type, eax, ecx, edx;
-  cpuid(CPUID_GETFEATURES, 0, &eax, &ecx, &edx);
-  cpu_type = (eax & 0xF0FF0);
+  auto cpuid_data = cpuid(CPUID_GETFEATURES, 0);
+  unsigned int cpu_type = (cpuid_data.eax & 0xF0FF0);
   switch (cpu_type) {
     case 0x006F0:
     case 0x10660:
@@ -127,6 +129,7 @@ static CpuMicroarch get_cpu_microarch() {
     case 0x406F0:
     case 0x50660:
       return IntelBroadwell;
+    case 0x406e0:
     case 0x506e0:
       return IntelSkylake;
     default:
@@ -179,6 +182,11 @@ static void init_attributes() {
                        PERF_COUNT_SW_PAGE_FAULTS);
 }
 
+const struct perf_event_attr& PerfCounters::ticks_attr() {
+  init_attributes();
+  return rr::ticks_attr;
+}
+
 PerfCounters::PerfCounters(pid_t tid) : tid(tid), started(false) {
   init_attributes();
 }
@@ -208,7 +216,7 @@ void PerfCounters::reset(Ticks ticks_period) {
 
   stop();
 
-  struct perf_event_attr attr = ticks_attr;
+  struct perf_event_attr attr = rr::ticks_attr;
   attr.sample_period = ticks_period;
   fd_ticks = start_counter(tid, -1, &attr);
 
@@ -271,3 +279,5 @@ PerfCounters::Extra PerfCounters::read_extra() {
   }
   return extra;
 }
+
+} // namespace rr
