@@ -1,7 +1,5 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-//#define DEBUGTAG "registers"
-
 #include "Registers.h"
 
 #include <array>
@@ -11,12 +9,13 @@
 #include <assert.h>
 #include <string.h>
 
+#include "ReplayTask.h"
 #include "log.h"
-#include "task.h"
 
-using namespace rr;
 using namespace std;
 
+namespace rr {
+
 struct RegisterValue {
   // The name of this register.
   const char* name;
@@ -179,7 +178,6 @@ RegisterValue RegisterInfo<rr::X64Arch>::non_gdb_registers[2] = {
 
 // 32-bit format, 64-bit format for all of these.
 // format_index in RegisterPrinting depends on the ordering here.
-static const char* hex_format[] = { "%" PRIx32, "%" PRIx64 };
 static const char* hex_format_leading_0x[] = { "0x%" PRIx32, "0x%" PRIx64 };
 // static const char* decimal_format[] = { "%" PRId32, "%" PRId64 };
 
@@ -234,7 +232,7 @@ void Registers::print_register_file_arch(FILE* f, const char* formats[]) const {
 }
 
 void Registers::print_register_file(FILE* f) const {
-  RR_ARCH_FUNCTION(print_register_file_arch, arch(), f, hex_format);
+  RR_ARCH_FUNCTION(print_register_file_arch, arch(), f, hex_format_leading_0x);
 }
 
 template <typename Arch>
@@ -279,11 +277,6 @@ void Registers::print_register_file_for_trace_arch(
 }
 
 void Registers::print_register_file_compact(FILE* f) const {
-  RR_ARCH_FUNCTION(print_register_file_for_trace_arch, arch(), f, Annotated,
-                   hex_format);
-}
-
-void Registers::print_register_file_for_trace(FILE* f) const {
   RR_ARCH_FUNCTION(print_register_file_for_trace_arch, arch(), f, Annotated,
                    hex_format_leading_0x);
 }
@@ -414,15 +407,19 @@ template <>
 }
 
 /*static*/ bool Registers::compare_register_files(
-    Task* t, const char* name1, const Registers& reg1, const char* name2,
+    ReplayTask* t, const char* name1, const Registers& reg1, const char* name2,
     const Registers& reg2, MismatchBehavior mismatch_behavior) {
   bool bail_error = mismatch_behavior >= BAIL_ON_MISMATCH;
   bool match = compare_register_files_internal(name1, reg1, name2, reg2,
                                                mismatch_behavior);
 
-  ASSERT(t, !bail_error || match)
-      << "Fatal register mismatch (ticks/rec:" << t->tick_count() << "/"
-      << t->current_trace_frame().ticks() << ")";
+  if (t) {
+    ASSERT(t, !bail_error || match)
+        << "Fatal register mismatch (ticks/rec:" << t->tick_count() << "/"
+        << t->current_trace_frame().ticks() << ")";
+  } else {
+    assert(!bail_error || match);
+  }
 
   if (match && mismatch_behavior == LOG_MISMATCHES) {
     LOG(info) << "(register files are the same for " << name1 << " and "
@@ -435,9 +432,10 @@ template <>
 template <typename Arch>
 size_t Registers::read_register_arch(uint8_t* buf, GdbRegister regno,
                                      bool* defined) const {
-  assert(regno < total_registers());
-  // Make sure these two definitions don't get out of sync.
-  assert(array_length(RegisterInfo<Arch>::registers) == total_registers());
+  if (regno >= array_length(RegisterInfo<Arch>::registers)) {
+    *defined = false;
+    return 0;
+  }
 
   RegisterValue& rv = RegisterInfo<Arch>::registers[regno];
   if (rv.nbytes == 0) {
@@ -477,7 +475,7 @@ size_t Registers::read_register_by_user_offset(uint8_t* buf, uintptr_t offset,
 }
 
 template <typename Arch>
-void Registers::write_register_arch(GdbRegister regno, const uint8_t* value,
+void Registers::write_register_arch(GdbRegister regno, const void* value,
                                     size_t value_size) {
   RegisterValue& rv = RegisterInfo<Arch>::registers[regno];
 
@@ -493,25 +491,48 @@ void Registers::write_register_arch(GdbRegister regno, const uint8_t* value,
   }
 }
 
-void Registers::write_register(GdbRegister regno, const uint8_t* value,
+void Registers::write_register(GdbRegister regno, const void* value,
                                size_t value_size) {
   RR_ARCH_FUNCTION(write_register_arch, arch(), regno, value, value_size);
 }
 
-template <typename Arch> size_t Registers::total_registers_arch() const {
-  return RegisterInfo<Arch>::num_registers;
+template <typename Arch>
+void Registers::write_register_by_user_offset_arch(uintptr_t offset,
+                                                   uintptr_t value) {
+  for (size_t regno = 0; regno < RegisterInfo<Arch>::num_registers; ++regno) {
+    RegisterValue& rv = RegisterInfo<Arch>::registers[regno];
+    if (rv.offset == offset) {
+      assert(rv.nbytes <= sizeof(value));
+      memcpy(rv.pointer_into(&u), &value, rv.nbytes);
+      return;
+    }
+  }
 }
 
-size_t Registers::total_registers() const {
-  RR_ARCH_FUNCTION(total_registers_arch, arch());
+void Registers::write_register_by_user_offset(uintptr_t offset,
+                                              uintptr_t value) {
+  RR_ARCH_FUNCTION(write_register_by_user_offset_arch, arch(), offset, value);
 }
 
+// In theory it doesn't matter how 32-bit register values are sign extended
+// to 64 bits for PTRACE_SETREGS. However:
+// -- When setting up a signal handler frame, the kernel does some arithmetic
+// on the 64-bit SP value and validates that the result points to writeable
+// memory. This validation fails if SP has been sign-extended to point
+// outside the 32-bit address space.
+// -- Some kernels (e.g. 4.3.3-301.fc23.x86_64) with commmit
+// c5c46f59e4e7c1ab244b8d38f2b61d317df90bba have a bug where if you clear
+// the upper 32 bits of %rax while in the kernel, syscalls may fail to
+// restart. So sign-extension is necessary for %eax in this case. We may as
+// well sign-extend %eax in all cases.
+
 typedef void (*NarrowConversion)(int32_t& r32, uint64_t& r64);
 typedef void (*SameConversion)(int32_t& r32, uint32_t& r64);
-template <NarrowConversion narrow, SameConversion same>
+template <NarrowConversion narrow, NarrowConversion narrow_signed,
+          SameConversion same>
 void convert_x86(X86Arch::user_regs_struct& x86,
                  X64Arch::user_regs_struct& x64) {
-  narrow(x86.eax, x64.rax);
+  narrow_signed(x86.eax, x64.rax);
   narrow(x86.ebx, x64.rbx);
   narrow(x86.ecx, x64.rcx);
   narrow(x86.edx, x64.rdx);
@@ -532,15 +553,8 @@ void convert_x86(X86Arch::user_regs_struct& x86,
 
 void to_x86_narrow(int32_t& r32, uint64_t& r64) { r32 = r64; }
 void to_x86_same(int32_t& r32, uint32_t& r64) { r32 = r64; }
-void from_x86_narrow(int32_t& r32, uint64_t& r64) {
-  // We must zero-extend 32-bit register values to 64-bit values when we
-  // do a PTRACE_SETREGS. Most of the time the upper 32 bits are irrelevant for
-  // a 32-bit tracee, but when setting up a signal handler frame, the kernel
-  // does some arithmetic on the 64-bit SP value and validates that the
-  // result points to writeable memory. This validation fails if SP has been
-  // sign-extended to point outside the 32-bit address space.
-  r64 = (uint32_t)r32;
-}
+void from_x86_narrow(int32_t& r32, uint64_t& r64) { r64 = (uint32_t)r32; }
+void from_x86_narrow_signed(int32_t& r32, uint64_t& r64) { r64 = (int64_t)r32; }
 void from_x86_same(int32_t& r32, uint32_t& r64) { r64 = r32; }
 
 void Registers::set_from_ptrace(const struct user_regs_struct& ptrace_regs) {
@@ -550,7 +564,7 @@ void Registers::set_from_ptrace(const struct user_regs_struct& ptrace_regs) {
   }
 
   assert(arch() == x86 && NativeArch::arch() == x86_64);
-  convert_x86<to_x86_narrow, to_x86_same>(
+  convert_x86<to_x86_narrow, to_x86_narrow, to_x86_same>(
       u.x86regs, *reinterpret_cast<X64Arch::user_regs_struct*>(
                      const_cast<struct user_regs_struct*>(&ptrace_regs)));
 }
@@ -573,7 +587,7 @@ struct user_regs_struct Registers::get_ptrace() const {
 
   assert(arch() == x86 && NativeArch::arch() == x86_64);
   memset(&result, 0, sizeof(result));
-  convert_x86<from_x86_narrow, from_x86_same>(
+  convert_x86<from_x86_narrow, from_x86_narrow_signed, from_x86_same>(
       const_cast<Registers*>(this)->u.x86regs, result.x64arch_api);
   return result.linux_api;
 }
@@ -585,12 +599,12 @@ vector<uint8_t> Registers::get_ptrace_for_arch(SupportedArch arch) const {
   vector<uint8_t> result;
   switch (arch) {
     case x86:
-      result.resize(sizeof(u.x86regs));
-      memcpy(result.data(), &u.x86regs, result.size());
+      result.resize(sizeof(tmp_regs.u.x86regs));
+      memcpy(result.data(), &tmp_regs.u.x86regs, result.size());
       break;
     case x86_64:
-      result.resize(sizeof(u.x64regs));
-      memcpy(result.data(), &u.x64regs, result.size());
+      result.resize(sizeof(tmp_regs.u.x64regs));
+      memcpy(result.data(), &tmp_regs.u.x64regs, result.size());
       break;
     default:
       assert(0 && "Unknown arch");
@@ -599,6 +613,21 @@ vector<uint8_t> Registers::get_ptrace_for_arch(SupportedArch arch) const {
   return result;
 }
 
+void Registers::set_from_ptrace_for_arch(SupportedArch a, const void* data,
+                                         size_t size) {
+  if (a == NativeArch::arch()) {
+    assert(size == sizeof(struct user_regs_struct));
+    set_from_ptrace(*static_cast<const struct user_regs_struct*>(data));
+    return;
+  }
+
+  assert(a == x86 && NativeArch::arch() == x86_64);
+  // We don't support a 32-bit tracee trying to set registers of a 64-bit tracee
+  assert(arch() == x86);
+  assert(size == sizeof(u.x86regs));
+  memcpy(&u.x86regs, data, sizeof(u.x86regs));
+}
+
 uintptr_t Registers::flags() const {
   switch (arch()) {
     case x86:
@@ -626,9 +655,28 @@ void Registers::set_flags(uintptr_t value) {
   }
 }
 
+bool Registers::syscall_failed() const {
+  auto result = syscall_result_signed();
+  return -ERANGE <= result && result < 0;
+}
+
+bool Registers::syscall_may_restart() const {
+  switch (-syscall_result_signed()) {
+    case ERESTART_RESTARTBLOCK:
+    case ERESTARTNOINTR:
+    case ERESTARTNOHAND:
+    case ERESTARTSYS:
+      return true;
+    default:
+      return false;
+  }
+}
+
 ostream& operator<<(ostream& stream, const Registers& r) {
   stream << "{ args:(" << HEX(r.arg1()) << "," << HEX(r.arg2()) << ","
          << HEX(r.arg3()) << "," << HEX(r.arg4()) << "," << HEX(r.arg5()) << ","
          << r.arg6() << ") orig_syscall:" << r.original_syscallno() << " }";
   return stream;
 }
+
+} // namespace rr
