@@ -9,6 +9,8 @@
 #include <unordered_map>
 #include <vector>
 
+namespace rr {
+
 class Task;
 
 /**
@@ -41,4 +43,6 @@ private:
   std::vector<uint32_t> index_to_result;
 };
 
+} // namespace rr
+
 #endif // RR_SECCOMP_FILTER_REWRITER_H_
