@@ -1,7 +1,5 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-//#define DEBUGTAG "Trace"
-
 #include "TraceStream.h"
 
 #include <inttypes.h>
@@ -9,14 +7,21 @@
 #include <sysexits.h>
 
 #include <fstream>
-#include <string>
 #include <sstream>
+#include <string>
 
+#include "AddressSpace.h"
+#include "RecordSession.h"
+#include "Task.h"
+#include "TaskishUid.h"
+#include "kernel_supplement.h"
 #include "log.h"
 #include "util.h"
 
 using namespace std;
 
+namespace rr {
+
 //
 // This represents the format and layout of recorded traces.  This
 // version number doesn't track the rr version number, because changes
@@ -26,7 +31,7 @@ using namespace std;
 // MUST increment this version number.  Otherwise users' old traces
 // will become unreplayable and they won't know why.
 //
-#define TRACE_VERSION 40
+#define TRACE_VERSION 58
 
 struct SubstreamData {
   const char* name;
@@ -34,15 +39,16 @@ struct SubstreamData {
   int threads;
 };
 
-static const SubstreamData substreams[TraceStream::SUBSTREAM_COUNT] = {
-  { "events", 1024 * 1024, 1 },
-  { "data_header", 1024 * 1024, 1 },
-  { "data", 8 * 1024 * 1024, 3 },
-  { "mmaps", 64 * 1024, 1 },
-  { "tasks", 64 * 1024, 1 }
+static SubstreamData substreams[TraceStream::SUBSTREAM_COUNT] = {
+  { "events", 1024 * 1024, 1 }, { "data_header", 1024 * 1024, 1 },
+  { "data", 1024 * 1024, 0 },   { "mmaps", 64 * 1024, 1 },
+  { "tasks", 64 * 1024, 1 },    { "generic", 64 * 1024, 1 },
 };
 
 static const SubstreamData& substream(TraceStream::Substream s) {
+  if (!substreams[TraceStream::RAW_DATA].threads) {
+    substreams[TraceStream::RAW_DATA].threads = sysconf(_SC_NPROCESSORS_ONLN);
+  }
   return substreams[s];
 }
 
@@ -142,6 +148,12 @@ static void ensure_default_rr_trace_dir() {
   ensure_dir(default_rr_trace_dir(), S_IRWXU);
 }
 
+string TraceStream::file_data_clone_file_name(const TaskUid& tuid) {
+  stringstream ss;
+  ss << trace_dir << "/cloned_data_" << tuid.tid() << "_" << tuid.serial();
+  return ss.str();
+}
+
 string TraceStream::path(Substream s) {
   return trace_dir + "/" + substream(s).name;
 }
@@ -169,13 +181,14 @@ struct BasicInfo {
   pid_t tid_;
   EncodedEvent ev;
   Ticks ticks_;
+  double monotonic_sec;
 };
 
 void TraceWriter::write_frame(const TraceFrame& frame) {
   auto& events = writer(EVENTS);
 
   BasicInfo basic_info = { frame.time(), frame.tid(), frame.event().encode(),
-                           frame.ticks() };
+                           frame.ticks(), frame.monotonic_time() };
   events << basic_info;
   if (!events.good()) {
     FATAL() << "Tried to save " << sizeof(basic_info)
@@ -220,7 +233,8 @@ TraceFrame TraceReader::read_frame() {
   BasicInfo basic_info;
   events >> basic_info;
   TraceFrame frame(basic_info.global_time, basic_info.tid_,
-                   Event(basic_info.ev), basic_info.ticks_);
+                   Event(basic_info.ev), basic_info.ticks_,
+                   basic_info.monotonic_sec);
   if (frame.event().has_exec_info() == HAS_EXEC_INFO) {
     events >> frame.recorded_regs >> frame.extra_perf;
 
@@ -231,9 +245,8 @@ TraceFrame TraceReader::read_frame() {
       vector<uint8_t> data;
       data.resize(extra_reg_bytes);
       events.read((char*)data.data(), extra_reg_bytes);
-      frame.recorded_extra_regs.set_arch(frame.event().arch());
       frame.recorded_extra_regs.set_to_raw_data(
-          (ExtraRegisters::Format)extra_reg_format, data);
+          frame.event().arch(), (ExtraRegisters::Format)extra_reg_format, data);
     } else {
       assert(extra_reg_format == ExtraRegisters::NONE);
       frame.recorded_extra_regs = ExtraRegisters(frame.event().arch());
@@ -264,6 +277,7 @@ void TraceWriter::write_task_event(const TraceTaskEvent& event) {
       tasks << event.file_name() << event.cmd_line() << event.fds_to_close();
       break;
     case TraceTaskEvent::EXIT:
+      tasks << event.exit_status_;
       break;
     case TraceTaskEvent::NONE:
       assert(0 && "Writing NONE TraceTaskEvent");
@@ -286,6 +300,7 @@ TraceTaskEvent TraceReader::read_task_event() {
       tasks >> r.file_name_ >> r.cmd_line_ >> r.fds_to_close_;
       break;
     case TraceTaskEvent::EXIT:
+      tasks >> r.exit_status_;
       break;
     case TraceTaskEvent::NONE:
       // Should be EOF only
@@ -295,25 +310,57 @@ TraceTaskEvent TraceReader::read_task_event() {
   return r;
 }
 
+static string base_file_name(const string& file_name) {
+  size_t last_slash = file_name.rfind('/');
+  return (last_slash != file_name.npos) ? file_name.substr(last_slash + 1)
+                                        : file_name;
+}
+
 string TraceWriter::try_hardlink_file(const string& file_name) {
   char count_str[20];
   sprintf(count_str, "%d", mmap_count);
 
-  size_t last_slash = file_name.rfind('/');
-  string basename = (last_slash != file_name.npos)
-                        ? file_name.substr(last_slash + 1)
-                        : file_name;
-  string link_path = dir() + "/mmap_" + count_str + "_hardlink_" + basename;
-  int ret = link(file_name.c_str(), link_path.c_str());
+  string path =
+      string("mmap_hardlink_") + count_str + "_" + base_file_name(file_name);
+  int ret = link(file_name.c_str(), (dir() + "/" + path).c_str());
   if (ret < 0) {
     // maybe tried to link across filesystems?
     return file_name;
   }
-  return link_path;
+  return path;
+}
+
+bool TraceWriter::try_clone_file(const string& file_name, string* new_name) {
+  char count_str[20];
+  sprintf(count_str, "%d", mmap_count);
+
+  string path =
+      string("mmap_clone_") + count_str + "_" + base_file_name(file_name);
+
+  ScopedFd src(file_name.c_str(), O_RDONLY);
+  if (!src.is_open()) {
+    return false;
+  }
+  string dest_path = dir() + "/" + path;
+  ScopedFd dest(dest_path.c_str(), O_WRONLY | O_CREAT | O_EXCL, 0700);
+  if (!dest.is_open()) {
+    return false;
+  }
+
+  int ret = ioctl(dest, BTRFS_IOC_CLONE, src.get());
+  if (ret < 0) {
+    // maybe not on the same filesystem, or filesystem doesn't support clone?
+    unlink(dest_path.c_str());
+    return false;
+  }
+
+  *new_name = path;
+  return true;
 }
 
 TraceWriter::RecordInTrace TraceWriter::write_mapped_region(
-    const KernelMapping& km, const struct stat& stat, MappingOrigin origin) {
+    Task* t, const KernelMapping& km, const struct stat& stat,
+    MappingOrigin origin) {
   auto& mmaps = writer(MMAPS);
   TraceReader::MappedDataSource source;
   string backing_file_name;
@@ -322,6 +369,12 @@ TraceWriter::RecordInTrace TraceWriter::write_mapped_region(
   } else if (origin == SYSCALL_MAPPING &&
              (km.inode() == 0 || km.fsname() == "/dev/zero (deleted)")) {
     source = TraceReader::SOURCE_ZERO;
+  } else if (origin == RR_BUFFER_MAPPING) {
+    source = TraceReader::SOURCE_ZERO;
+  } else if ((km.flags() & MAP_PRIVATE) &&
+             t->session().as_record()->use_file_cloning() &&
+             try_clone_file(km.fsname(), &backing_file_name)) {
+    source = TraceReader::SOURCE_FILE;
   } else if (should_copy_mmap_region(km, stat) &&
              files_assumed_immutable.find(make_pair(
                  stat.st_dev, stat.st_ino)) == files_assumed_immutable.end()) {
@@ -346,7 +399,9 @@ TraceWriter::RecordInTrace TraceWriter::write_mapped_region(
                                              : DONT_RECORD_IN_TRACE;
 }
 
-KernelMapping TraceReader::read_mapped_region(MappedData* data, bool* found) {
+KernelMapping TraceReader::read_mapped_region(MappedData* data, bool* found,
+                                              ValidateSourceFile validate,
+                                              TimeConstraint time_constraint) {
   if (found) {
     *found = false;
   }
@@ -356,16 +411,19 @@ KernelMapping TraceReader::read_mapped_region(MappedData* data, bool* found) {
     return KernelMapping();
   }
 
-  mmaps.save_state();
   TraceFrame::Time time;
-  mmaps >> time;
-  mmaps.restore_state();
-  if (time != global_time) {
-    return KernelMapping();
+  if (time_constraint == CURRENT_TIME_ONLY) {
+    mmaps.save_state();
+    mmaps >> time;
+    mmaps.restore_state();
+    if (time != global_time) {
+      return KernelMapping();
+    }
   }
 
   string original_file_name;
   string backing_file_name;
+  MappedDataSource source;
   remote_ptr<void> start, end;
   dev_t device;
   ino_t inode;
@@ -373,30 +431,41 @@ KernelMapping TraceReader::read_mapped_region(MappedData* data, bool* found) {
   uint32_t uid, gid, mode;
   uint64_t file_offset_bytes;
   int64_t mtime, file_size;
-  mmaps >> time >> data->source >> start >> end >> original_file_name >>
-      device >> inode >> prot >> flags >> file_offset_bytes >>
-      backing_file_name >> mode >> uid >> gid >> file_size >> mtime;
-  assert(time == global_time);
-  if (data->source == SOURCE_FILE) {
-    if (backing_file_name[0] != '/') {
-      backing_file_name = dir() + "/" + backing_file_name;
-    }
-    struct stat backing_stat;
-    if (stat(backing_file_name.c_str(), &backing_stat)) {
-      FATAL() << "Failed to stat " << backing_file_name
-              << ": replay is impossible";
-    }
-    if (backing_stat.st_ino != inode || backing_stat.st_mode != mode ||
-        backing_stat.st_uid != uid || backing_stat.st_gid != gid ||
-        backing_stat.st_size != file_size || backing_stat.st_mtime != mtime) {
-      LOG(error)
-          << "Metadata of " << original_file_name
-          << " changed: replay divergence likely, but continuing anyway ...";
+  mmaps >> time >> source >> start >> end >> original_file_name >> device >>
+      inode >> prot >> flags >> file_offset_bytes >> backing_file_name >>
+      mode >> uid >> gid >> file_size >> mtime;
+  assert(time_constraint == ANY_TIME || time == global_time);
+  if (data) {
+    data->source = source;
+    if (data->source == SOURCE_FILE) {
+      static const string clone_prefix("mmap_clone_");
+      bool is_clone =
+          backing_file_name.substr(0, clone_prefix.size()) == clone_prefix;
+      if (backing_file_name[0] != '/') {
+        backing_file_name = dir() + "/" + backing_file_name;
+      }
+      if (!is_clone && validate == VALIDATE) {
+        struct stat backing_stat;
+        if (stat(backing_file_name.c_str(), &backing_stat)) {
+          FATAL() << "Failed to stat " << backing_file_name
+                  << ": replay is impossible";
+        }
+        if (backing_stat.st_ino != inode || backing_stat.st_mode != mode ||
+            backing_stat.st_uid != uid || backing_stat.st_gid != gid ||
+            backing_stat.st_size != file_size ||
+            backing_stat.st_mtime != mtime) {
+          LOG(error) << "Metadata of " << original_file_name
+                     << " changed: replay divergence likely, but continuing "
+                        "anyway ...";
+        }
+      }
+      data->file_name = backing_file_name;
+      data->data_offset_bytes = file_offset_bytes;
+    } else {
+      data->data_offset_bytes = 0;
     }
+    data->file_size_bytes = file_size;
   }
-  data->file_name = backing_file_name;
-  data->file_data_offset_bytes = file_offset_bytes;
-  data->file_size_bytes = file_size;
   if (found) {
     *found = true;
   }
@@ -404,26 +473,6 @@ KernelMapping TraceReader::read_mapped_region(MappedData* data, bool* found) {
                        flags, file_offset_bytes);
 }
 
-static ostream& operator<<(ostream& out, const vector<string>& vs) {
-  out << vs.size() << endl;
-  for (auto& v : vs) {
-    out << v << '\0';
-  }
-  return out;
-}
-
-static istream& operator>>(istream& in, vector<string>& vs) {
-  size_t len;
-  in >> len;
-  in.ignore(1);
-  for (size_t i = 0; i < len; ++i) {
-    char buf[PATH_MAX];
-    in.getline(buf, sizeof(buf), '\0');
-    vs.push_back(buf);
-  }
-  return in;
-}
-
 void TraceWriter::write_raw(const void* d, size_t len, remote_ptr<void> addr) {
   auto& data = writer(RAW_DATA);
   auto& data_header = writer(RAW_DATA_HEADER);
@@ -461,6 +510,40 @@ bool TraceReader::read_raw_data_for_frame(const TraceFrame& frame, RawData& d) {
   return true;
 }
 
+void TraceWriter::write_generic(const void* d, size_t len) {
+  auto& generic = writer(GENERIC);
+  generic << global_time << len;
+  generic.write(d, len);
+}
+
+void TraceReader::read_generic(vector<uint8_t>& out) {
+  auto& generic = reader(GENERIC);
+  TraceFrame::Time time;
+  size_t num_bytes;
+  generic >> time >> num_bytes;
+  assert(time == global_time);
+  out.resize(num_bytes);
+  generic.read((char*)out.data(), num_bytes);
+}
+
+bool TraceReader::read_generic_for_frame(const TraceFrame& frame,
+                                         vector<uint8_t>& out) {
+  auto& generic = reader(GENERIC);
+  if (generic.at_end()) {
+    return false;
+  }
+  TraceFrame::Time time;
+  generic.save_state();
+  generic >> time;
+  generic.restore_state();
+  assert(time >= frame.time());
+  if (time > frame.time()) {
+    return false;
+  }
+  read_generic(out);
+  return true;
+}
+
 void TraceWriter::close() {
   for (auto& w : writers) {
     w->close();
@@ -489,16 +572,13 @@ static string make_trace_dir(const string& exe_path) {
   return dir;
 }
 
-TraceWriter::TraceWriter(const vector<string>& argv, const vector<string>& envp,
-                         const string& cwd, int bind_to_cpu)
-    : TraceStream(make_trace_dir(argv[0]),
+TraceWriter::TraceWriter(const std::string& file_name, int bind_to_cpu)
+    : TraceStream(make_trace_dir(file_name),
                   // Somewhat arbitrarily start the
                   // global time from 1.
                   1),
-      mmap_count(0) {
-  this->argv = argv;
-  this->envp = envp;
-  this->cwd = cwd;
+      mmap_count(0),
+      supports_file_data_cloning_(false) {
   this->bind_to_cpu = bind_to_cpu;
 
   for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
@@ -507,12 +587,43 @@ TraceWriter::TraceWriter(const vector<string>& argv, const vector<string>& envp,
   }
 
   string ver_path = version_path();
-  fstream version(ver_path.c_str(), fstream::out);
-  if (!version.good()) {
+  ScopedFd version_fd(ver_path.c_str(), O_RDWR | O_CREAT, 0600);
+  if (!version_fd.is_open()) {
     FATAL() << "Unable to create " << ver_path;
   }
-  version << TRACE_VERSION << endl;
+  char buf[10];
+  sprintf(buf, "%d\n", TRACE_VERSION);
+  ssize_t buf_len = strlen(buf);
+  if (write(version_fd, buf, buf_len) != buf_len) {
+    FATAL() << "Unable to write " << ver_path;
+  }
 
+  // Test if file data cloning is supported
+  string version_clone_path = trace_dir + "/tmp_clone";
+  ScopedFd version_clone_fd(version_clone_path.c_str(), O_WRONLY | O_CREAT,
+                            0600);
+  if (!version_clone_fd.is_open()) {
+    FATAL() << "Unable to create " << version_clone_path;
+  }
+  btrfs_ioctl_clone_range_args clone_args;
+  clone_args.src_fd = version_fd;
+  clone_args.src_offset = 0;
+  clone_args.src_length = buf_len;
+  clone_args.dest_offset = 0;
+  if (ioctl(version_clone_fd, BTRFS_IOC_CLONE_RANGE, &clone_args) == 0) {
+    supports_file_data_cloning_ = true;
+  }
+  unlink(version_clone_path.c_str());
+
+  if (!probably_not_interactive(STDOUT_FILENO)) {
+    printf("rr: Saving execution to trace directory `%s'.\n",
+           trace_dir.c_str());
+  }
+
+  write_generic(&bind_to_cpu, sizeof(bind_to_cpu));
+}
+
+void TraceWriter::make_latest_trace() {
   string link_name = latest_trace_symlink();
   // Try to update the symlink to |this|.  We only try attempt
   // to set the symlink once.  If the link is re-created after
@@ -525,18 +636,6 @@ TraceWriter::TraceWriter(const vector<string>& argv, const vector<string>& envp,
     FATAL() << "Failed to update symlink `" << link_name << "' to `"
             << trace_dir << "'.";
   }
-
-  if (!probably_not_interactive(STDOUT_FILENO)) {
-    printf("rr: Saving the execution of `%s' to trace directory `%s'.\n",
-           argv[0].c_str(), trace_dir.c_str());
-  }
-
-  ofstream out(args_env_path());
-  out << cwd << '\0';
-  out << argv;
-  out << envp;
-  out << bind_to_cpu;
-  assert(out.good());
 }
 
 TraceFrame TraceReader::peek_frame() {
@@ -552,26 +651,6 @@ TraceFrame TraceReader::peek_frame() {
   return frame;
 }
 
-TraceFrame TraceReader::peek_to(pid_t pid, EventType type, SyscallState state) {
-  auto& events = reader(EVENTS);
-  TraceFrame frame;
-  events.save_state();
-  auto saved_time = global_time;
-  while (good() && !at_end()) {
-    frame = read_frame();
-    if (frame.tid() == pid && frame.event().type() == type &&
-        (!frame.event().is_syscall_event() ||
-         frame.event().Syscall().state == state)) {
-      events.restore_state();
-      global_time = saved_time;
-      return frame;
-    }
-  }
-  FATAL() << "Unable to find requested frame in stream";
-  // Unreachable
-  return frame;
-}
-
 void TraceReader::rewind() {
   for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
     reader(s).rewind();
@@ -581,12 +660,7 @@ void TraceReader::rewind() {
 }
 
 TraceReader::TraceReader(const string& dir)
-    : TraceStream(dir.empty() ? latest_trace_symlink() : dir,
-                  // Initialize the global time at 0, so
-                  // that when we tick it when reading
-                  // the first trace, it matches the
-                  // initial global time at recording, 1.
-                  0) {
+    : TraceStream(dir.empty() ? latest_trace_symlink() : dir, 1) {
   for (Substream s = SUBSTREAM_FIRST; s < SUBSTREAM_COUNT; ++s) {
     readers[s] = unique_ptr<CompressedReader>(new CompressedReader(path(s)));
   }
@@ -623,14 +697,14 @@ TraceReader::TraceReader(const string& dir)
     exit(EX_DATAERR);
   }
 
-  ifstream in(args_env_path());
-  assert(in.good());
-  char buf[PATH_MAX];
-  in.getline(buf, sizeof(buf), '\0');
-  cwd = buf;
-  in >> argv;
-  in >> envp;
-  in >> bind_to_cpu;
+  vector<uint8_t> bind_to_cpu_bytes;
+  read_generic(bind_to_cpu_bytes);
+  assert(bind_to_cpu_bytes.size() == sizeof(bind_to_cpu));
+  memcpy(&bind_to_cpu, bind_to_cpu_bytes.data(), sizeof(bind_to_cpu));
+
+  // Set the global time at 0, so that when we tick it for the first
+  // event, it matches the initial global time at recording, 1.
+  global_time = 0;
 }
 
 /**
@@ -645,9 +719,6 @@ TraceReader::TraceReader(const TraceReader& other)
         unique_ptr<CompressedReader>(new CompressedReader(other.reader(s)));
   }
 
-  argv = other.argv;
-  envp = other.envp;
-  cwd = other.cwd;
   bind_to_cpu = other.bind_to_cpu;
 }
 
@@ -666,3 +737,5 @@ uint64_t TraceReader::compressed_bytes() const {
   }
   return total;
 }
+
+} // namespace rr
