@@ -3,59 +3,20 @@
 #ifndef RR_KERNEL_ABI_H
 #define RR_KERNEL_ABI_H
 
-// Include remote_ptr.h first since it (indirectly) requires a definition of
-// ERANGE, which other headers below #undef :-(
-#include "remote_ptr.h"
-
-// Get all the kernel definitions so we can verify our alternative versions.
-#include <arpa/inet.h>
-#include <asm/ldt.h>
-#include <elf.h>
-#include <fcntl.h>
-#include <linux/capability.h>
-#include <linux/ethtool.h>
-#include <linux/filter.h>
-#include <linux/futex.h>
-#include <linux/ipc.h>
-#include <linux/msg.h>
-#include <linux/net.h>
-#include <linux/sem.h>
-#include <linux/shm.h>
-#include <linux/sockios.h>
-#include <linux/sysctl.h>
-#include <linux/videodev2.h>
-#include <linux/wireless.h>
-#include <poll.h>
-#include <signal.h>
-#include <stddef.h>
-#include <stdint.h>
-#include <sound/asound.h>
-#include <sys/epoll.h>
-#include <sys/ioctl.h>
-#include <sys/quota.h>
-#include <sys/resource.h>
-#include <sys/socket.h>
-#include <sys/stat.h>
-#include <sys/time.h>
-#include <sys/times.h>
-#include <sys/un.h>
-#include <sys/user.h>
-#include <sys/utsname.h>
-#include <sys/vfs.h>
-#include <sys/sysinfo.h>
-#include <termios.h>
-
 #include <assert.h>
+#include <signal.h>
 
 #include <vector>
 
+#include "remote_ptr.h"
+
+namespace rr {
+
 class remote_code_ptr;
 class Task;
 
 enum SupportedArch { x86, x86_64, SupportedArch_MAX = x86_64 };
 
-namespace rr {
-
 #if defined(__i386__)
 const SupportedArch RR_NATIVE_ARCH = SupportedArch::x86;
 #elif defined(__x86_64__)
@@ -80,16 +41,15 @@ template <typename T> struct Verifier<RR_NATIVE_ARCH, T, T> {
   // with itself or (unlikely) the system's structure with itself.
 };
 
-#define RR_VERIFY_TYPE_ARCH(arch_, system_type_, rr_type_)                     \
-  static_assert(Verifier<arch_, system_type_, rr_type_>::same_size,            \
-                "type " #system_type_ " not correctly defined");
-
-// For instances where the system type and the rr type are named differently.
-#define RR_VERIFY_TYPE_EXPLICIT(system_type_, rr_type_)                        \
-  RR_VERIFY_TYPE_ARCH(arch_, system_type_, rr_type_)
-
-// For instances where the system type and the rr type are named identically.
-#define RR_VERIFY_TYPE(type_) RR_VERIFY_TYPE_EXPLICIT(::type_, type_)
+// We want verify that the types have the same size as their
+// counterparts in the system header. To avoid having to include
+// all system headers here, we instead make the verification macros
+// a no-op unless inlcuded from kernel_abi.cc.
+#ifndef RR_VERIFY_TYPE
+#define RR_VERIFY_TYPE_ARCH(arch_, system_type_, rr_type_) // no-op
+#define RR_VERIFY_TYPE_EXPLICIT(system_type_, rr_type_)    // no-op
+#define RR_VERIFY_TYPE(x)                                  // no-op
+#endif
 
 struct KernelConstants {
   static const ::size_t SIGINFO_MAX_SIZE = 128;
@@ -137,9 +97,18 @@ struct FcntlConstants {
   };
 };
 
-struct WordSize32Defs : public KernelConstants {
+// Various ELF constants we use. These are verified to be the same
+// as those in the system headers by kernel_abi.cc
+enum ELFCLASS { CLASSNONE = 0, CLASS32 = 1, CLASS64 = 2 };
+enum ELFENDIAN { DATA2LSB = 1 };
+enum EM {
+  I386 = 3,
+  X86_64 = 62,
+};
+
+struct WordSize32Defs {
   static const ::size_t SIGINFO_PAD_SIZE =
-      (SIGINFO_MAX_SIZE / sizeof(int32_t)) - 3;
+      (KernelConstants::SIGINFO_MAX_SIZE / sizeof(int32_t)) - 3;
 
   typedef int16_t signed_short;
   typedef uint16_t unsigned_short;
@@ -163,15 +132,56 @@ struct WordSize32Defs : public KernelConstants {
   typedef int32_t sigchld_clock_t;
   typedef uint32_t __statfs_word;
 
-  static const size_t elfclass = ELFCLASS32;
-  typedef Elf32_Ehdr ElfEhdr;
-  typedef Elf32_Shdr ElfShdr;
-  typedef Elf32_Sym ElfSym;
+  static const size_t elfclass = ELFCLASS::CLASS32;
+  typedef struct {
+    uint8_t e_ident[16];
+    uint16_t e_type;
+    uint16_t e_machine;
+    uint32_t e_version;
+    uint32_t e_entry;
+    uint32_t e_phoff;
+    uint32_t e_shoff;
+    uint32_t e_flags;
+    uint16_t e_ehsize;
+    uint16_t e_phentsize;
+    uint16_t e_phnum;
+    uint16_t e_shentsize;
+    uint16_t e_shnum;
+    uint16_t e_shstrndx;
+  } ElfEhdr;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf32_Ehdr, ElfEhdr);
+  typedef struct {
+    uint32_t sh_name;
+    uint32_t sh_type;
+    uint32_t sh_flags;
+    uint32_t sh_addr;
+    uint32_t sh_offset;
+    uint32_t sh_size;
+    uint32_t sh_link;
+    uint32_t sh_info;
+    uint32_t sh_addralign;
+    uint32_t sh_entsize;
+  } ElfShdr;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf32_Shdr, ElfShdr);
+  typedef struct {
+    uint32_t st_name;
+    uint32_t st_value;
+    uint32_t st_size;
+    uint8_t st_info;
+    uint8_t st_other;
+    uint16_t st_shndx;
+  } ElfSym;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf32_Sym, ElfSym);
+  typedef struct {
+    int32_t d_tag;
+    uint32_t d_val;
+  } ElfDyn;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf32_Dyn, ElfDyn);
 };
 
-struct WordSize64Defs : public KernelConstants {
+struct WordSize64Defs {
   static const ::size_t SIGINFO_PAD_SIZE =
-      (SIGINFO_MAX_SIZE / sizeof(int32_t)) - 4;
+      (KernelConstants::SIGINFO_MAX_SIZE / sizeof(int32_t)) - 4;
 
   typedef int16_t signed_short;
   typedef uint16_t unsigned_short;
@@ -195,10 +205,51 @@ struct WordSize64Defs : public KernelConstants {
   typedef int64_t sigchld_clock_t;
   typedef signed_long __statfs_word;
 
-  static const size_t elfclass = ELFCLASS64;
-  typedef Elf64_Ehdr ElfEhdr;
-  typedef Elf64_Shdr ElfShdr;
-  typedef Elf64_Sym ElfSym;
+  static const size_t elfclass = ELFCLASS::CLASS64;
+  typedef struct {
+    uint8_t e_ident[16];
+    uint16_t e_type;
+    uint16_t e_machine;
+    uint32_t e_version;
+    uint64_t e_entry;
+    uint64_t e_phoff;
+    uint64_t e_shoff;
+    uint32_t e_flags;
+    uint16_t e_ehsize;
+    uint16_t e_phentsize;
+    uint16_t e_phnum;
+    uint16_t e_shentsize;
+    uint16_t e_shnum;
+    uint16_t e_shstrndx;
+  } ElfEhdr;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf64_Ehdr, ElfEhdr);
+  typedef struct {
+    uint32_t sh_name;
+    uint32_t sh_type;
+    uint64_t sh_flags;
+    uint64_t sh_addr;
+    uint64_t sh_offset;
+    uint64_t sh_size;
+    uint32_t sh_link;
+    uint32_t sh_info;
+    uint64_t sh_addralign;
+    uint64_t sh_entsize;
+  } ElfShdr;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf64_Shdr, ElfShdr);
+  typedef struct {
+    uint32_t st_name;
+    uint8_t st_info;
+    uint8_t st_other;
+    uint16_t st_shndx;
+    uint64_t st_value;
+    uint64_t st_size;
+  } ElfSym;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf64_Sym, ElfSym);
+  typedef struct {
+    uint64_t d_tag;
+    uint64_t d_val;
+  } ElfDyn;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf64_Dyn, ElfDyn);
 };
 
 /**
@@ -207,7 +258,9 @@ struct WordSize64Defs : public KernelConstants {
  * the tracee.
  */
 template <SupportedArch arch_, typename wordsize>
-struct BaseArch : public wordsize, public FcntlConstants {
+struct BaseArch : public wordsize,
+                  public FcntlConstants,
+                  public KernelConstants {
   static SupportedArch arch() { return arch_; }
 
   typedef typename wordsize::syscall_slong_t syscall_slong_t;
@@ -450,6 +503,16 @@ struct BaseArch : public wordsize, public FcntlConstants {
   };
   RR_VERIFY_TYPE(termios);
 
+  struct termio {
+    unsigned_short c_iflag;
+    unsigned_short c_oflag;
+    unsigned_short c_cflag;
+    unsigned_short c_lflag;
+    unsigned char c_line;
+    unsigned char c_cc[8];
+  };
+  RR_VERIFY_TYPE(termio);
+
   struct winsize {
     unsigned_short ws_row;
     unsigned_short ws_col;
@@ -476,7 +539,7 @@ struct BaseArch : public wordsize, public FcntlConstants {
   struct msqid64_ds {
     ipc64_perm msg_perm;
     // These msg*time fields are really __kernel_time_t plus
-    // appropiate padding.  We don't touch the fields, though.
+    // appropriate padding.  We don't touch the fields, though.
     //
     // We do, however, suffix them with _only_little_endian to
     // urge anybody who does touch them to make sure the right
@@ -1180,11 +1243,56 @@ struct BaseArch : public wordsize, public FcntlConstants {
     unsigned char components[128];
   };
   RR_VERIFY_TYPE(snd_ctl_card_info);
+
+  struct usbdevfs_iso_packet_desc {
+    unsigned int length;
+    unsigned int actual_length;
+    unsigned int status;
+  };
+  RR_VERIFY_TYPE(usbdevfs_iso_packet_desc);
+
+  struct usbdevfs_urb {
+    unsigned char type;
+    unsigned char endpoint;
+    int status;
+    unsigned int flags;
+    ptr<void> buffer;
+    int buffer_length;
+    int actual_length;
+    int start_frame;
+    union {
+      int number_of_packets;
+      unsigned int stream_id;
+    };
+    int error_count;
+    unsigned int signr;
+    ptr<void> usercontext;
+    struct usbdevfs_iso_packet_desc iso_frame_desc[0];
+  };
+  RR_VERIFY_TYPE(usbdevfs_urb);
+
+  struct usbdevfs_ioctl {
+    int ifno;
+    int ioctl_code;
+    ptr<void> data;
+  };
+  RR_VERIFY_TYPE(usbdevfs_ioctl);
+
+  struct usbdevfs_ctrltransfer {
+    uint8_t bRequestType;
+    uint8_t bRequest;
+    uint16_t wValue;
+    uint16_t wIndex;
+    uint16_t wLength;
+    uint32_t timeout;
+    ptr<void> data;
+  };
+  RR_VERIFY_TYPE(usbdevfs_ctrltransfer);
 };
 
 struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
-  static const size_t elfmachine = EM_386;
-  static const size_t elfendian = ELFDATA2LSB;
+  static const size_t elfmachine = EM::I386;
+  static const size_t elfendian = ELFENDIAN::DATA2LSB;
 
   static const MmapCallingSemantics mmap_semantics = StructArguments;
   static const CloneTLSType clone_tls_type = UserDescPointer;
@@ -1321,8 +1429,8 @@ struct X86Arch : public BaseArch<SupportedArch::x86, WordSize32Defs> {
 };
 
 struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
-  static const size_t elfmachine = EM_X86_64;
-  static const size_t elfendian = ELFDATA2LSB;
+  static const size_t elfmachine = EM::X86_64;
+  static const size_t elfendian = ELFENDIAN::DATA2LSB;
 
   static const MmapCallingSemantics mmap_semantics = RegisterArguments;
   static const CloneTLSType clone_tls_type = PthreadStructurePointer;
@@ -1494,6 +1602,9 @@ std::vector<uint8_t> syscall_instruction(SupportedArch arch);
  */
 ssize_t syscall_instruction_length(SupportedArch arch);
 
+void set_arch_siginfo(const siginfo_t& siginfo, SupportedArch a, void* dest,
+                      size_t dest_size);
+
 #if defined(__i386__)
 typedef X86Arch NativeArch;
 #elif defined(__x86_64__)
