@@ -12,11 +12,13 @@
 
 #include "Command.h"
 #include "Flags.h"
-#include "log.h"
 #include "RecordCommand.h"
+#include "log.h"
 
 using namespace std;
 
+namespace rr {
+
 // Show version and quit.
 static bool show_version = false;
 
@@ -106,7 +108,9 @@ void print_usage(FILE* out) {
       "detection\n"
       "                             says otherwise.  NAME should be a string "
       "like\n"
-      "                             'Ivy Bridge'.\n"
+      "                             'Ivy Bridge'. Note that rr will not work "
+      "with\n"
+      "                             Intel Merom or Penryn microarchitectures.\n"
       "  -C, --checksum={on-syscalls,on-all-events}|FROM_TIME\n"
       "                             compute and store (during recording) or\n"
       "                             read and verify (during replay) checksums\n"
@@ -144,13 +148,17 @@ void print_usage(FILE* out) {
       "  -V, --verbose              log messages that may not be urgently \n"
       "                             critical to the user\n"
       "  -W, --wait-secs=<NUM_SECS> wait NUM_SECS seconds just after startup,\n"
-      "                             before initiating recording or replaying\n",
+      "                             before initiating recording or replaying\n"
+      "\n"
+      "Use RR_LOG to control logging; e.g. RR_LOG=all:warn,Task:debug\n",
       out);
 }
 
 static void init_random() {
   // Not very good, but good enough for our non-security-sensitive needs.
-  srandom(time(nullptr) ^ getpid());
+  int key = time(nullptr) ^ getpid();
+  srandom(key);
+  srand(key);
 }
 
 bool parse_global_option(std::vector<std::string>& args) {
@@ -192,7 +200,11 @@ bool parse_global_option(std::vector<std::string>& args) {
       }
       break;
     case 'D':
-      flags.dump_on = atoi(opt.value.c_str());
+      if (opt.value == "RDTSC") {
+        flags.dump_on = Flags::DUMP_ON_RDTSC;
+      } else {
+        flags.dump_on = atoi(opt.value.c_str());
+      }
       break;
     case 'E':
       flags.fatal_errors_and_warnings = true;
@@ -212,9 +224,6 @@ bool parse_global_option(std::vector<std::string>& args) {
     case 'T':
       flags.dump_at = atoi(opt.value.c_str());
       break;
-    case 'V':
-      flags.verbose = true;
-      break;
     case 'N':
       show_version = true;
       break;
@@ -224,6 +233,10 @@ bool parse_global_option(std::vector<std::string>& args) {
   return true;
 }
 
+} // namespace rr
+
+using namespace rr;
+
 int main(int argc, char* argv[]) {
   init_random();
 
