@@ -1,7 +1,5 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-//#define DEBUGTAG "Signal"
-
 #include "record_signal.h"
 
 #include <assert.h>
@@ -10,29 +8,30 @@
 #include <sched.h>
 #include <stdlib.h>
 #include <string.h>
-#include <syscall.h>
 #include <sys/mman.h>
+#include <sys/prctl.h>
+#include <sys/resource.h>
 #include <sys/user.h>
+#include <syscall.h>
 #include <x86intrin.h>
 
 #include "preload/preload_interface.h"
 
 #include "AutoRemoteSyscalls.h"
 #include "Flags.h"
-#include "kernel_metadata.h"
-#include "log.h"
 #include "PerfCounters.h"
 #include "RecordSession.h"
-#include "task.h"
+#include "RecordTask.h"
 #include "TraceStream.h"
+#include "kernel_metadata.h"
+#include "log.h"
 #include "util.h"
 
-using namespace rr;
 using namespace std;
 
-static __inline__ unsigned long long rdtsc(void) { return __rdtsc(); }
+namespace rr {
 
-static const int STOPSIG_SYSCALL = 0x80 | SIGTRAP;
+static __inline__ unsigned long long rdtsc(void) { return __rdtsc(); }
 
 template <typename Arch> static size_t sigaction_sigset_size_arch() {
   return Arch::sigaction_sigset_size;
@@ -46,7 +45,7 @@ static size_t sigaction_sigset_size(SupportedArch arch) {
  * Restore the blocked-ness and sigaction for SIGSEGV from |t|'s local
  * copy.
  */
-static void restore_sigsegv_state(Task* t) {
+static void restore_sigsegv_state(RecordTask* t) {
   const vector<uint8_t>& sa = t->signal_action(SIGSEGV);
   AutoRemoteSyscalls remote(t);
   {
@@ -64,7 +63,7 @@ static void restore_sigsegv_state(Task* t) {
 
 /** Return true iff |t->ip()| points at a RDTSC instruction. */
 static const uint8_t rdtsc_insn[] = { 0x0f, 0x31 };
-static bool is_ip_rdtsc(Task* t) {
+static bool is_ip_rdtsc(RecordTask* t) {
   uint8_t insn[sizeof(rdtsc_insn)];
   if (sizeof(insn) !=
       t->read_bytes_fallible(t->ip().to_data_ptr<uint8_t>(), sizeof(insn),
@@ -78,10 +77,10 @@ static bool is_ip_rdtsc(Task* t) {
  * Return true if |t| was stopped because of a SIGSEGV resulting
  * from a rdtsc and |t| was updated appropriately, false otherwise.
  */
-static bool try_handle_rdtsc(Task* t, siginfo_t* si) {
+static bool try_handle_rdtsc(RecordTask* t, siginfo_t* si) {
   ASSERT(t, si->si_signo == SIGSEGV);
 
-  if (!is_ip_rdtsc(t)) {
+  if (!is_ip_rdtsc(t) || t->tsc_mode == PR_TSC_SIGSEGV) {
     return false;
   }
 
@@ -100,69 +99,91 @@ static bool try_handle_rdtsc(Task* t, siginfo_t* si) {
  * Return true if |t| was stopped because of a SIGSEGV and we want to retry
  * the instruction after emulating MAP_GROWSDOWN.
  */
-static bool try_grow_map(Task* t, siginfo_t* si) {
+static bool try_grow_map(RecordTask* t, siginfo_t* si) {
   ASSERT(t, si->si_signo == SIGSEGV);
 
   // Use kernel_abi to avoid odd inconsistencies between distros
   auto arch_si = reinterpret_cast<NativeArch::siginfo_t*>(si);
   auto addr = arch_si->_sifields._sigfault.si_addr_.rptr();
 
-  auto maps = t->vm()->maps_starting_at(floor_page_size(addr));
-  auto it = maps.begin();
-  if (it == maps.end() || addr >= it->map.start() ||
-      !(it->map.flags() & MAP_GROWSDOWN)) {
+  if (t->vm()->has_mapping(addr)) {
+    LOG(debug) << "try_grow_map " << addr << ": address already mapped";
     return false;
   }
-
-  auto new_start = floor_page_size(addr);
-  static const uintptr_t grow_size = 0x10000;
-  if (it->map.start().as_int() >= grow_size) {
-    auto possible_new_start = std::min(new_start, it->map.start() - grow_size);
-    auto earlier_maps = t->vm()->maps_starting_at(possible_new_start);
-    if (earlier_maps.begin()->map.start() == it->map.start()) {
-      // No intervening map
-      new_start = possible_new_start;
-    }
+  auto maps = t->vm()->maps_starting_at(floor_page_size(addr));
+  auto it = maps.begin();
+  if (it == maps.end()) {
+    LOG(debug) << "try_grow_map " << addr << ": no later map to grow downward";
+    return false;
+  }
+  if (!(it->map.flags() & MAP_GROWSDOWN)) {
+    LOG(debug) << "try_grow_map " << addr << ": map is not MAP_GROWSDOWN ("
+               << it->map << ")";
+    return false;
+  }
+  if (addr >= page_size() && t->vm()->has_mapping(addr - page_size())) {
+    LOG(debug) << "try_grow_map " << addr << ": address would be in guard page";
+    return false;
   }
-
   struct rlimit stack_limit;
+  remote_ptr<void> limit_bottom;
   int ret = prlimit(t->tid, RLIMIT_STACK, NULL, &stack_limit);
   if (ret >= 0 && stack_limit.rlim_cur != RLIM_INFINITY) {
-    new_start = std::max(new_start,
-                         ceil_page_size(it->map.end() - stack_limit.rlim_cur));
-    if (new_start > addr) {
+    limit_bottom = ceil_page_size(it->map.end() - stack_limit.rlim_cur);
+    if (limit_bottom > addr) {
+      LOG(debug) << "try_grow_map " << addr << ": RLIMIT_STACK exceeded";
       return false;
     }
   }
 
+  // Try to grow by 64K at a time to reduce signal frequency.
+  auto new_start = floor_page_size(addr);
+  static const uintptr_t grow_size = 0x10000;
+  if (it->map.start().as_int() >= grow_size) {
+    auto possible_new_start = std::max(
+        limit_bottom, std::min(new_start, it->map.start() - grow_size));
+    // Ensure that no mapping exists between possible_new_start - page_size()
+    // and new_start. If there is, possible_new_start is not valid, in which
+    // case we just abandon the optimization.
+    if (possible_new_start >= page_size() &&
+        !t->vm()->has_mapping(possible_new_start - page_size()) &&
+        t->vm()->maps_starting_at(possible_new_start - page_size())
+                .begin()
+                ->map.start() == it->map.start()) {
+      new_start = possible_new_start;
+    }
+  }
+  LOG(debug) << "try_grow_map " << addr << ": trying to grow map " << it->map;
+
   {
     AutoRemoteSyscalls remote(t, AutoRemoteSyscalls::DISABLE_MEMORY_PARAMS);
     remote.infallible_mmap_syscall(
         new_start, it->map.start() - new_start, it->map.prot(),
-        (it->map.flags() & ~MAP_GROWSDOWN) | MAP_ANONYMOUS, -1, 0);
+        (it->map.flags() & ~MAP_GROWSDOWN) | MAP_ANONYMOUS | MAP_FIXED, -1, 0);
   }
 
   KernelMapping km =
       t->vm()->map(new_start, it->map.start() - new_start, it->map.prot(),
                    it->map.flags() | MAP_ANONYMOUS, 0, string(),
                    KernelMapping::NO_DEVICE, KernelMapping::NO_INODE);
-  t->trace_writer().write_mapped_region(km, km.fake_stat());
+  t->trace_writer().write_mapped_region(t, km, km.fake_stat());
   // No need to flush syscallbuf here. It's safe to map these pages "early"
   // before they're really needed.
   t->record_event(Event(EV_GROW_MAP, NO_EXEC_INFO, t->arch()),
-                  Task::DONT_FLUSH_SYSCALLBUF);
+                  RecordTask::DONT_FLUSH_SYSCALLBUF);
   t->push_event(Event::noop(t->arch()));
-  LOG(debug) << "  trapped for MAP_GROWSDOWN";
+  LOG(debug) << "try_grow_map " << addr << ": extended map "
+             << t->vm()->mapping_of(addr).map;
   return true;
 }
 
-void disarm_desched_event(Task* t) {
+void disarm_desched_event(RecordTask* t) {
   if (ioctl(t->desched_fd, PERF_EVENT_IOC_DISABLE, 0)) {
     FATAL() << "Failed to disarm desched event";
   }
 }
 
-void arm_desched_event(Task* t) {
+void arm_desched_event(RecordTask* t) {
   if (ioctl(t->desched_fd, PERF_EVENT_IOC_ENABLE, 0)) {
     FATAL() << "Failed to disarm desched event";
   }
@@ -173,7 +194,7 @@ void arm_desched_event(Task* t) {
  * The tracee's execution may be advanced, and if so |regs| is updated
  * to the tracee's latest state.
  */
-static void handle_desched_event(Task* t, const siginfo_t* si) {
+static void handle_desched_event(RecordTask* t, const siginfo_t* si) {
   ASSERT(t, (SYSCALLBUF_DESCHED_SIGNAL == si->si_signo &&
              si->si_code == POLL_IN && si->si_fd == t->desched_fd_child))
       << "Tracee is using SIGPWR??? (code=" << si->si_code
@@ -192,7 +213,8 @@ static void handle_desched_event(Task* t, const siginfo_t* si) {
    * the desched_signal_may_be_relevant was set by the outermost syscallbuf
    * invocation.
    */
-  if (!t->syscallbuf_hdr->desched_signal_may_be_relevant ||
+  if (!t->read_mem(REMOTE_PTR_FIELD(t->syscallbuf_child,
+                                    desched_signal_may_be_relevant)) ||
       t->running_inside_desched()) {
     LOG(debug) << "  (not entering may-block syscall; resuming)";
     /* We have to disarm the event just in case the tracee
@@ -273,14 +295,14 @@ static void handle_desched_event(Task* t, const siginfo_t* si) {
     disarm_desched_event(t);
 
     t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_UNLIMITED_TICKS);
-    int sig = t->stop_sig();
 
-    if (STOPSIG_SYSCALL == sig) {
+    if (t->status().is_syscall()) {
       if (t->is_arm_desched_event_syscall()) {
         continue;
       }
       break;
     }
+
     // Completely ignore spurious desched signals and
     // signals that aren't going to be delivered to the
     // tracee.
@@ -295,6 +317,8 @@ static void handle_desched_event(Task* t, const siginfo_t* si) {
     // TODO: it's theoretically possible for this to
     // happen an unbounded number of consecutive times
     // and the tracee never switched out.
+    int sig = t->stop_sig();
+    ASSERT(t, sig);
     if (SYSCALLBUF_DESCHED_SIGNAL == sig ||
         PerfCounters::TIME_SLICE_SIGNAL == sig || t->is_sig_ignored(sig)) {
       LOG(debug) << "  dropping ignored " << signal_name(sig);
@@ -327,10 +351,10 @@ static void handle_desched_event(Task* t, const siginfo_t* si) {
      * away this breadcrumb so that we can figure out what syscall
      * the tracee was in, and how much "scratch" space it carved
      * off the syscallbuf, if needed. */
-    const struct syscallbuf_record* desched_rec =
-        next_record(t->syscallbuf_hdr);
+    remote_ptr<const struct syscallbuf_record> desched_rec =
+        t->next_syscallbuf_record();
     t->push_event(DeschedEvent(desched_rec, t->arch()));
-    int call = t->desched_rec()->syscallno;
+    int call = t->read_mem(REMOTE_PTR_FIELD(t->desched_rec(), syscallno));
 
     /* The descheduled syscall was interrupted by a signal, like
      * all other may-restart syscalls, with the exception that
@@ -349,21 +373,22 @@ static void handle_desched_event(Task* t, const siginfo_t* si) {
    * otherwise we'll get a divergence during replay, which will not
    * encounter this problem.
    */
-  int call = t->desched_rec()->syscallno;
+  int call = t->read_mem(REMOTE_PTR_FIELD(t->desched_rec(), syscallno));
   ev.regs.set_original_syscallno(call);
   t->set_regs(ev.regs);
-  ev.state = EXITING_SYSCALL;
+  // runnable_state_changed will observe us entering this syscall and change
+  // state to ENTERING_SYSCALL
 
   LOG(debug) << "  resuming (and probably switching out) blocked `"
              << t->syscall_name(call) << "'";
 }
 
-static bool is_safe_to_deliver_signal(Task* t) {
-  struct syscallbuf_hdr* hdr = t->syscallbuf_hdr;
-
-  if (!hdr) {
+static bool is_safe_to_deliver_signal(RecordTask* t) {
+  if (!t->syscallbuf_child) {
     /* Can't be in critical section because the lock
      * doesn't exist yet! */
+    LOG(debug) << "Safe to deliver signal at " << t->ip()
+               << " because no syscallbuf_child";
     return true;
   }
 
@@ -373,28 +398,38 @@ static bool is_safe_to_deliver_signal(Task* t) {
      * syscallbuf critical sections.  The
      * exception is signal handlers "re-entering"
      * desched'd syscalls, which are OK. */
+    LOG(debug) << "Safe to deliver signal at " << t->ip()
+               << " because not in syscallbuf";
     return true;
   }
 
   if (t->is_in_traced_syscall()) {
-    LOG(debug) << "  tracee at traced syscallbuf syscall";
+    LOG(debug) << "Safe to deliver signal at " << t->ip()
+               << " because in traced syscall";
     return true;
   }
 
   if (t->is_in_untraced_syscall() && t->desched_rec()) {
-    LOG(debug) << "  tracee interrupted by desched of "
-               << t->syscall_name(t->desched_rec()->syscallno);
+    LOG(debug) << "Safe to deliver signal at " << t->ip()
+               << " because tracee interrupted by desched of "
+               << t->syscall_name(t->read_mem(
+                      REMOTE_PTR_FIELD(t->desched_rec(), syscallno)));
     return true;
   }
 
   // Our emulation of SYS_rrcall_notify_syscall_hook_exit clears this flag.
-  hdr->notify_on_syscall_hook_exit = true;
+  t->write_mem(
+      REMOTE_PTR_FIELD(t->syscallbuf_child, notify_on_syscall_hook_exit),
+      (uint8_t)1);
+  LOG(debug) << "Not safe to deliver signal at " << t->ip();
   return false;
 }
 
-SignalHandled handle_signal(Task* t, siginfo_t* si) {
+SignalHandled handle_signal(RecordTask* t, siginfo_t* si,
+                            SignalDeterministic deterministic) {
   LOG(debug) << t->tid << ": handling signal " << signal_name(si->si_signo)
-             << " (pevent: " << t->ptrace_event() << ", event: " << t->ev();
+             << " (pevent: " << ptrace_event_name(t->ptrace_event())
+             << ", event: " << t->ev();
 
   /* We have to check for a desched event first, because for
    * those we *do not* want to (and cannot, most of the time)
@@ -448,20 +483,31 @@ SignalHandled handle_signal(Task* t, siginfo_t* si) {
   /* This signal was generated by the program or an external
    * source, record it normally. */
 
-  if (t->emulate_ptrace_stop((si->si_signo << 8) | 0x7f,
-                             SIGNAL_DELIVERY_STOP)) {
-    t->save_ptrace_signal_siginfo(*si);
-    // Record a SCHED event so that replay progresses the tracee to the
+  if (t->emulate_ptrace_stop(WaitStatus::for_stop_sig(si->si_signo), si)) {
+    // Record an event so that replay progresses the tracee to the
     // current point before we notify the tracer.
-    t->push_event(Event(EV_SCHED, HAS_EXEC_INFO, t->arch()));
-    t->record_current_event();
-    t->pop_event(EV_SCHED);
+    // If the signal is deterministic, record it as an EV_SIGNAL so that
+    // we replay it using the deterministic-signal replay path. This is
+    // more efficient than emulate_async_signal. Also emulate_async_signal
+    // currently assumes it won't encounter a deterministic SIGTRAP (due to
+    // a hardcoded breakpoint in the tracee).
+    if (deterministic == DETERMINISTIC_SIG) {
+      t->push_event(SignalEvent(*si, deterministic, t));
+      t->record_current_event();
+      t->pop_event(EV_SIGNAL);
+    } else {
+      t->push_event(Event(EV_SCHED, HAS_EXEC_INFO, t->arch()));
+      t->record_current_event();
+      t->pop_event(EV_SCHED);
+    }
     // ptracer has been notified, so don't deliver the signal now.
     // The signal won't be delivered for real until the ptracer calls
     // PTRACE_CONT with the signal number (which we don't support yet!).
     return SIGNAL_PTRACE_STOP;
   }
 
-  t->push_event(SignalEvent(*si, t->arch()));
+  t->push_event(SignalEvent(*si, deterministic, t));
   return SIGNAL_HANDLED;
 }
+
+} // namespace rr
