@@ -6,7 +6,7 @@ send_gdb('c')
 expect_gdb('Breakpoint 1')
 
 send_gdb('checkpoint')
-expect_gdb('= 1')
+expect_gdb('Checkpoint 1 at')
 send_gdb('next')
 
 send_gdb('restart 1')
