@@ -10,7 +10,7 @@ static void sighandler(int sig) {
   _exit(0);
 }
 
-int main(int argc, char* argv[]) {
+int main(void) {
   size_t page_size = sysconf(_SC_PAGESIZE);
   int fd = open(DUMMY_FILE, O_CREAT | O_EXCL | O_RDWR, 0600);
   int one = 1;
