@@ -28,7 +28,8 @@ static void find_home_device(void) {
       sscanf(mount_line, "%*d %*d %d:%d %*s %*s %*s - %*s %1000s %*s", &maj,
              &min, home_device);
     }
-    if (maj == major(home_stat.st_dev) && min == minor(home_stat.st_dev)) {
+    if (maj == (int)major(home_stat.st_dev) &&
+        min == (int)minor(home_stat.st_dev) && maj != 0) {
       atomic_printf("%s (%d:%d) is on device special file %s\n", home, maj, min,
                     home_device);
       return;
@@ -41,10 +42,24 @@ static void find_home_device(void) {
   exit(0);
 }
 
-int main(int argc, char* argv[]) {
+int main(void) {
   struct dqblk dq;
   int ret;
 
+  /* ENOTBLK returned before checking cmd */
+  ret = quotactl(0x12345678, "/", getuid(), (caddr_t)&dq);
+  test_assert(ret < 0 && errno == ENOTBLK);
+
+  /* ENOEND returned before checking cmd */
+  ret =
+      quotactl(0x12345678, "/asdfjlafdlkk289892389pkj", getuid(), (caddr_t)&dq);
+  test_assert(ret < 0 && errno == ENOENT);
+
+  /* ENODEV returned before checking cmd when the device is a Btrfs volume */
+  ret = quotactl(0x12345678, "/dev/dm-0", getuid(), (caddr_t)&dq);
+  test_assert(ret < 0 && (errno == ENOTBLK || errno == ENOENT ||
+                          errno == ENODEV || errno == EINVAL));
+
   find_home_device();
   ret =
       quotactl(QCMD(Q_GETQUOTA, USRQUOTA), home_device, getuid(), (caddr_t)&dq);
