@@ -6,10 +6,15 @@
 #define _GNU_SOURCE 1
 #define _POSIX_C_SOURCE 2
 
+/* btrfs needs NULL but doesn't #include it */
+#include <stdlib.h>
+
 #include <arpa/inet.h>
 #include <asm/prctl.h>
 #include <assert.h>
 #include <dirent.h>
+#include <dlfcn.h>
+#include <elf.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <inttypes.h>
@@ -20,6 +25,7 @@
 #include <linux/futex.h>
 #include <linux/if.h>
 #include <linux/limits.h>
+#include <linux/mtio.h>
 #include <linux/perf_event.h>
 #include <linux/random.h>
 #include <linux/seccomp.h>
@@ -30,6 +36,7 @@
 #include <poll.h>
 #include <pthread.h>
 #include <pty.h>
+#include <pwd.h>
 #include <sched.h>
 #include <signal.h>
 #include <sound/asound.h>
@@ -37,12 +44,13 @@
 #include <stddef.h>
 #include <stdint.h>
 #include <stdio.h>
-#include <stdlib.h>
 #include <string.h>
-#include <syscall.h>
-#include <sys/file.h>
-#include <sys/xattr.h>
 #include <sys/epoll.h>
+#include <sys/eventfd.h>
+#include <sys/fanotify.h>
+#include <sys/file.h>
+#include <sys/fsuid.h>
+#include <sys/inotify.h>
 #include <sys/ioctl.h>
 #include <sys/ipc.h>
 #include <sys/mman.h>
@@ -61,8 +69,8 @@
 #include <sys/stat.h>
 #include <sys/sysinfo.h>
 #include <sys/time.h>
-#include <sys/times.h>
 #include <sys/timerfd.h>
+#include <sys/times.h>
 #include <sys/types.h>
 #include <sys/ucontext.h>
 #include <sys/uio.h>
@@ -71,6 +79,8 @@
 #include <sys/utsname.h>
 #include <sys/vfs.h>
 #include <sys/wait.h>
+#include <sys/xattr.h>
+#include <syscall.h>
 #include <termios.h>
 #include <time.h>
 #include <ucontext.h>
@@ -150,6 +160,11 @@ inline static void check_data(void* buf, size_t len) {
  */
 inline static uint64_t rdtsc(void) { return __rdtsc(); }
 
+/**
+ * Perform some syscall that writes an event, i.e. is not syscall-buffered.
+ */
+inline static void event_syscall(void) { syscall(-1); }
+
 static uint64_t GUARD_VALUE = 0xdeadbeeff00dbaad;
 
 /**
@@ -185,6 +200,8 @@ inline static void free_guard(size_t size, void* p) {
   free((char*)p - sizeof(GUARD_VALUE));
 }
 
+inline static void crash_null_deref(void) { *(volatile int*)NULL = 0; }
+
 #define ALLOCATE_GUARD(p, v) p = allocate_guard(sizeof(*p), v)
 #define VERIFY_GUARD(p) verify_guard(sizeof(*p), p)
 #define FREE_GUARD(p) free_guard(sizeof(*p), p)
@@ -196,4 +213,19 @@ inline static void free_guard(size_t size, void* p) {
 #define SECCOMP_SET_MODE_FILTER 1
 #endif
 
+/* Old systems don't have linux/kcmp.h */
+#define RR_KCMP_FILE 0
+#define RR_KCMP_FILES 2
+
+/* Old systems don't have these */
+#ifndef TIOCGPKT
+#define TIOCGPKT _IOR('T', 0x38, int)
+#endif
+#ifndef TIOCGPTLCK
+#define TIOCGPTLCK _IOR('T', 0x39, int)
+#endif
+#ifndef TIOCGEXCL
+#define TIOCGEXCL _IOR('T', 0x40, int)
+#endif
+
 #endif /* RRUTIL_H */
