@@ -6,7 +6,7 @@ expect_gdb('Breakpoint 1')
 send_gdb('c')
 expect_gdb('Breakpoint 1')
 send_gdb('check')
-expect_gdb('= 1')
+expect_gdb('Checkpoint 1 at')
 send_gdb('c')
 expect_gdb('xited normally')
 send_gdb('restart 1')
