@@ -10,7 +10,7 @@ send_gdb('c');
 expect_gdb('Breakpoint 1, before_writing')
 
 send_gdb('checkpoint');
-expect_gdb('= 1');
+expect_gdb('Checkpoint 1 at');
 
 send_gdb('c');
 expect_gdb('Breakpoint 2, after_writing')
