@@ -8,7 +8,7 @@ expect_gdb('Program received signal SIGILL')
 expect_gdb('ud2')
 
 send_gdb('checkpoint')
-expect_gdb('= 1')
+expect_gdb('Checkpoint 1 at')
 
 send_gdb('c')
 expect_gdb('Breakpoint 1, sighandler')
