@@ -1,7 +1,5 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil; -*- */
 
-//#define DEBUGTAG "Util"
-
 //#define FIRST_INTERESTING_EVENT 10700
 //#define LAST_INTERESTING_EVENT 10900
 
@@ -16,8 +14,8 @@
 #include <limits.h>
 #include <linux/magic.h>
 #include <linux/prctl.h>
-#include <string.h>
 #include <stdlib.h>
+#include <string.h>
 #include <sys/vfs.h>
 #include <unistd.h>
 
@@ -26,16 +24,17 @@
 #include "AddressSpace.h"
 #include "AutoRemoteSyscalls.h"
 #include "Flags.h"
+#include "ReplaySession.h"
+#include "ReplayTask.h"
+#include "TraceStream.h"
 #include "kernel_abi.h"
 #include "kernel_metadata.h"
 #include "log.h"
-#include "ReplaySession.h"
 #include "seccomp-bpf.h"
-#include "task.h"
-#include "TraceStream.h"
 
 using namespace std;
-using namespace rr;
+
+namespace rr {
 
 // FIXME this function assumes that there's only one address space.
 // Should instead only look at the address space of the task in
@@ -60,7 +59,7 @@ bool probably_not_interactive(int fd) {
 
 int clone_flags_to_task_flags(int flags_arg) {
   int flags = CLONE_SHARE_NOTHING;
-  // See task.h for description of the flags.
+  // See Task.h for description of the flags.
   flags |= (CLONE_CHILD_CLEARTID & flags_arg) ? CLONE_CLEARTID : 0;
   flags |= (CLONE_SETTLS & flags_arg) ? CLONE_SET_TLS : 0;
   flags |= (CLONE_SIGHAND & flags_arg) ? CLONE_SHARE_SIGHANDLERS : 0;
@@ -125,7 +124,7 @@ void format_dump_filename(Task* t, TraceFrame::Time global_time,
            t->rec_tid, global_time, tag);
 }
 
-bool should_dump_memory(Task* t, const TraceFrame& f) {
+bool should_dump_memory(const TraceFrame& f) {
   const Flags* flags = &Flags::get();
 
 #if defined(FIRST_INTERESTING_EVENT)
@@ -140,6 +139,12 @@ bool should_dump_memory(Task* t, const TraceFrame& f) {
   }
 #endif
   return flags->dump_on == Flags::DUMP_ON_ALL ||
+         (f.event().is_syscall_event() &&
+          f.event().Syscall().number == flags->dump_on) ||
+         (f.event().is_signal_event() &&
+          f.event().Signal().siginfo.si_signo == -flags->dump_on) ||
+         (flags->dump_on == Flags::DUMP_ON_RDTSC &&
+          f.event().type() == EV_SEGV_RDTSC) ||
          flags->dump_at == int(f.time());
 }
 
@@ -169,7 +174,7 @@ void dump_process_memory(Task* t, TraceFrame::Time global_time,
   fclose(dump_file);
 }
 
-static void notify_checksum_error(Task* t, TraceFrame::Time global_time,
+static void notify_checksum_error(ReplayTask* t, TraceFrame::Time global_time,
                                   unsigned checksum, unsigned rec_checksum,
                                   const string& raw_map_line) {
   char cur_dump[PATH_MAX];
@@ -223,6 +228,10 @@ static bool checksum_segment_filter(const AddressSpace::Mapping& m) {
   struct stat st;
   int may_diverge;
 
+  if (m.map.fsname() == "[vsyscall]") {
+    // This can't be read/checksummed.
+    return false;
+  }
   if (stat(m.map.fsname().c_str(), &st)) {
     /* If there's no persistent resource backing this
      * mapping, we should expect it to change. */
@@ -233,13 +242,27 @@ static bool checksum_segment_filter(const AddressSpace::Mapping& m) {
    * immutable, skip checksumming, it's a waste of time.  Except
    * if the mapping is mutable, for example the rw data segment
    * of a system library, then it's interesting. */
+  static const char mmap_clone[] = "mmap_clone_";
   may_diverge =
-      should_copy_mmap_region(m.map, st) || (PROT_WRITE & m.map.prot());
+      m.map.fsname().substr(0, array_length(mmap_clone) - 1) != mmap_clone &&
+      (should_copy_mmap_region(m.map, st) || (PROT_WRITE & m.map.prot()));
   LOG(debug) << (may_diverge ? "CHECKSUMMING" : "  skipping") << " '"
              << m.map.fsname() << "'";
   return may_diverge;
 }
 
+static uint32_t compute_checksum(void* data, size_t len) {
+  uint32_t checksum = len;
+  size_t words = len / sizeof(uint32_t);
+  uint32_t* buf = static_cast<uint32_t*>(data);
+  for (size_t i = 0; i < words; ++i) {
+    checksum = (checksum << 4) + checksum + buf[i];
+  }
+  return checksum;
+}
+
+static const uint32_t ignored_checksum = 0x98765432;
+
 /**
  * Either create and store checksums for each segment mapped in |t|'s
  * address space, or validate an existing computed checksum.  Behavior
@@ -261,23 +284,63 @@ static void iterate_checksums(Task* t, ChecksumMode mode,
     FATAL() << "Failed to open checksum file " << filename;
   }
 
-  const AddressSpace& as = *(t->vm());
-  for (auto m : as.maps()) {
-    vector<uint8_t> mem;
-    ssize_t valid_mem_len = 0;
+  remote_ptr<unsigned char> in_replay_flag;
+  unsigned char in_replay = 0;
+  if (t->preload_globals) {
+    in_replay_flag = REMOTE_PTR_FIELD(t->preload_globals, in_replay);
+    in_replay = t->read_mem(in_replay_flag);
+    t->write_mem(in_replay_flag, (unsigned char)0);
+  }
 
-    if (checksum_segment_filter(m)) {
-      mem.resize(m.map.size());
-      valid_mem_len =
-          t->read_bytes_fallible(m.map.start(), m.map.size(), mem.data());
-      valid_mem_len = max(ssize_t(0), valid_mem_len);
+  const AddressSpace& as = *t->vm();
+  for (auto m : as.maps()) {
+    string raw_map_line = m.map.str();
+    uint32_t rec_checksum = 0;
+
+    if (VALIDATE_CHECKSUMS == mode) {
+      char line[1024];
+      fgets(line, sizeof(line), c.checksums_file);
+      unsigned long rec_start;
+      unsigned long rec_end;
+      unsigned tmp_checksum;
+      int nparsed =
+          sscanf(line, "(%x) %lx-%lx", &tmp_checksum, &rec_start, &rec_end);
+      rec_checksum = tmp_checksum;
+      remote_ptr<void> rec_start_addr = rec_start;
+      remote_ptr<void> rec_end_addr = rec_end;
+      ASSERT(t, 3 == nparsed) << "Parsed " << nparsed << " items";
+      ASSERT(t, rec_start_addr == m.map.start() && rec_end_addr == m.map.end())
+          << "Segment " << rec_start_addr << "-" << rec_end_addr
+          << " changed to " << m.map << "??";
+      if (is_start_of_scratch_region(t, rec_start_addr)) {
+        /* Replay doesn't touch scratch regions, so
+         * their contents are allowed to diverge.
+         * Tracees can't observe those segments unless
+         * they do something sneaky (or disastrously
+         * buggy). */
+        LOG(debug) << "Not validating scratch starting at " << rec_start_addr;
+        continue;
+      }
+      if (rec_checksum == ignored_checksum) {
+        LOG(debug) << "Checksum not computed during recording";
+        continue;
+      }
+    } else {
+      if (!checksum_segment_filter(m)) {
+        fprintf(c.checksums_file, "(%x) %s\n", ignored_checksum,
+                raw_map_line.c_str());
+        continue;
+      }
     }
 
-    unsigned* buf = (unsigned*)mem.data();
-    unsigned checksum = 0;
-    int i;
+    vector<uint8_t> mem;
+    mem.resize(m.map.size());
+    ssize_t valid_mem_len =
+        t->read_bytes_fallible(m.map.start(), m.map.size(), mem.data());
+    ASSERT(t, valid_mem_len >= 0);
+    mem.resize(valid_mem_len);
 
-    if (m.map.fsname().find(SYSCALLBUF_SHMEM_PATH_PREFIX) == 0) {
+    if (m.flags & AddressSpace::Mapping::IS_SYSCALLBUF) {
       /* The syscallbuf consists of a region that's written
       * deterministically wrt the trace events, and a
       * region that's written nondeterministically in the
@@ -292,57 +355,44 @@ static void iterate_checksums(Task* t, ChecksumMode mode,
       * the deterministic region. */
       auto child_hdr = m.map.start().cast<struct syscallbuf_hdr>();
       auto hdr = t->read_mem(child_hdr);
-      valid_mem_len = !buf ? 0 : sizeof(hdr) + hdr.num_rec_bytes +
-                                     sizeof(struct syscallbuf_record);
+      mem.resize(sizeof(hdr) + hdr.num_rec_bytes +
+                 sizeof(struct syscallbuf_record));
     }
 
-    ASSERT(t, buf || valid_mem_len == 0);
-    for (i = 0; i < ssize_t(valid_mem_len / sizeof(*buf)); ++i) {
-      checksum += buf[i];
-    }
+    uint32_t checksum = compute_checksum(mem.data(), mem.size());
 
-    string raw_map_line = m.map.str();
-    if (STORE_CHECKSUMS == c.mode) {
+    if (STORE_CHECKSUMS == mode) {
       fprintf(c.checksums_file, "(%x) %s\n", checksum, raw_map_line.c_str());
     } else {
-      char line[1024];
-      unsigned rec_checksum;
-      unsigned long rec_start;
-      unsigned long rec_end;
-      int nparsed;
+      ASSERT(t, t->session().is_replaying());
+      auto rt = static_cast<ReplayTask*>(t);
 
-      fgets(line, sizeof(line), c.checksums_file);
-      nparsed =
-          sscanf(line, "(%x) %lx-%lx", &rec_checksum, &rec_start, &rec_end);
-      remote_ptr<void> rec_start_addr = rec_start;
-      remote_ptr<void> rec_end_addr = rec_end;
-      ASSERT(t, 3 == nparsed) << "Only parsed " << nparsed << " items";
-
-      ASSERT(t, rec_start_addr == m.map.start() && rec_end_addr == m.map.end())
-          << "Segment " << rec_start_addr << "-" << rec_end_addr
-          << " changed to " << m.map << "??";
-
-      if (is_start_of_scratch_region(t, rec_start_addr)) {
-        /* Replay doesn't touch scratch regions, so
-         * their contents are allowed to diverge.
-         * Tracees can't observe those segments unless
-         * they do something sneaky (or disastrously
-         * buggy). */
-        LOG(debug) << "Not validating scratch starting at 0x" << hex
-                   << rec_start_addr << dec;
-        continue;
-      }
+      // Ignore checksums when valid_mem_len == 0
       if (checksum != rec_checksum) {
-        notify_checksum_error(t, c.global_time, checksum, rec_checksum,
+        notify_checksum_error(rt, c.global_time, checksum, rec_checksum,
                               raw_map_line.c_str());
       }
     }
   }
 
+  if (in_replay_flag) {
+    t->write_mem(in_replay_flag, in_replay);
+  }
+
   fclose(c.checksums_file);
 }
 
-bool should_checksum(Task* t, const TraceFrame& f) {
+bool should_checksum(const TraceFrame& f) {
+  if (f.event().type() == EV_EXIT || f.event().type() == EV_UNSTABLE_EXIT) {
+    // Task is dead, or at least detached, and we can't read its memory safely.
+    return false;
+  }
+  if (f.event().has_ticks_slop()) {
+    // We may not be at the same point during recording and replay, so don't
+    // compute checksums.
+    return false;
+  }
+
   int checksum = Flags::get().checksum;
   bool is_syscall_exit = EV_SYSCALL == f.event().type() &&
                          EXITING_SYSCALL == f.event().Syscall().state;
@@ -428,12 +478,12 @@ signal_action default_action(int sig) {
   }
 }
 
-SignalDeterministic is_deterministic_signal(const siginfo_t& si) {
+SignalDeterministic is_deterministic_signal(Task* t) {
+  const siginfo_t& si = t->get_siginfo();
   switch (si.si_signo) {
     /* These signals may be delivered deterministically;
      * we'll check for sure below. */
     case SIGILL:
-    case SIGTRAP:
     case SIGBUS:
     case SIGFPE:
     case SIGSEGV:
@@ -446,6 +496,14 @@ SignalDeterministic is_deterministic_signal(const siginfo_t& si) {
        * kernel delivered it, then it must have been
        * delivered deterministically. */
       return si.si_code > 0 ? DETERMINISTIC_SIG : NONDETERMINISTIC_SIG;
+    case SIGTRAP: {
+      // The kernel code is wrong about this one. It treats singlestep
+      // traps as deterministic, but they aren't. PTRACE_ATTACH traps aren't
+      // really deterministic either.
+      auto reasons = t->compute_trap_reasons();
+      return reasons.breakpoint || reasons.watchpoint ? DETERMINISTIC_SIG
+                                                      : NONDETERMINISTIC_SIG;
+    }
     default:
       /* All other signals can never be delivered
        * deterministically (to the approximation required by
@@ -504,6 +562,14 @@ bool should_copy_mmap_region(const KernelMapping& mapping,
     LOG(debug) << "  (no copy for private mapping of +x " << file_name << ")";
     return false;
   }
+  bool can_read_file = (0 == access(file_name.c_str(), R_OK));
+  if (!can_read_file) {
+    // It's possible for a tracee to mmap a file it doesn't have permission
+    // to read, e.g. if a daemon opened the file and passed the fd over a
+    // socket. We should copy the data now because we won't be able to read
+    // it later. nscd does this.
+    return true;
+  }
 
   // TODO: using "can the euid of the rr process write this
   // file" as an approximation of whether the tracee can write
@@ -566,58 +632,47 @@ void resize_shmem_segment(ScopedFd& fd, uint64_t num_bytes) {
   }
 }
 
-void cpuid(int code, int subrequest, unsigned int* a, unsigned int* c,
-           unsigned int* d) {
+CPUIDData cpuid(int code, int subrequest) {
+  CPUIDData result;
   asm volatile("cpuid"
-               : "=a"(*a), "=c"(*c), "=d"(*d)
-               : "a"(code), "c"(subrequest)
-               : "ebx");
+               : "=a"(result.eax), "=b"(result.ebx), "=c"(result.ecx),
+                 "=d"(result.edx)
+               : "a"(code), "c"(subrequest));
+  return result;
 }
 
 template <typename Arch>
-static void extract_clone_parameters_arch(const Registers& regs,
-                                          remote_ptr<void>* stack,
-                                          remote_ptr<int>* parent_tid,
-                                          remote_ptr<void>* tls,
-                                          remote_ptr<int>* child_tid) {
+static CloneParameters extract_clone_parameters_arch(const Registers& regs) {
+  CloneParameters result;
+  result.stack = regs.arg2();
+  result.ptid = regs.arg3();
   switch (Arch::clone_parameter_ordering) {
     case Arch::FlagsStackParentTLSChild:
-      if (stack) {
-        *stack = regs.arg2();
-      }
-      if (parent_tid) {
-        *parent_tid = regs.arg3();
-      }
-      if (tls) {
-        *tls = regs.arg4();
-      }
-      if (child_tid) {
-        *child_tid = regs.arg5();
-      }
+      result.tls = regs.arg4();
+      result.ctid = regs.arg5();
       break;
     case Arch::FlagsStackParentChildTLS:
-      if (stack) {
-        *stack = regs.arg2();
-      }
-      if (parent_tid) {
-        *parent_tid = regs.arg3();
-      }
-      if (child_tid) {
-        *child_tid = regs.arg4();
-      }
-      if (tls) {
-        *tls = regs.arg5();
-      }
+      result.tls = regs.arg5();
+      result.ctid = regs.arg4();
       break;
   }
+  int flags = (int)regs.arg1();
+  // If these flags aren't set, the corresponding clone parameters may be
+  // invalid pointers, so make sure they're ignored.
+  if (!(flags & CLONE_PARENT_SETTID)) {
+    result.ptid = nullptr;
+  }
+  if (!(flags & (CLONE_CHILD_SETTID | CLONE_CHILD_CLEARTID))) {
+    result.ctid = nullptr;
+  }
+  if (!(flags & CLONE_SETTLS)) {
+    result.tls = nullptr;
+  }
+  return result;
 }
 
-void extract_clone_parameters(Task* t, remote_ptr<void>* stack,
-                              remote_ptr<int>* parent_tid,
-                              remote_ptr<void>* tls,
-                              remote_ptr<int>* child_tid) {
-  RR_ARCH_FUNCTION(extract_clone_parameters_arch, t->arch(), t->regs(), stack,
-                   parent_tid, tls, child_tid);
+CloneParameters extract_clone_parameters(Task* t) {
+  RR_ARCH_FUNCTION(extract_clone_parameters_arch, t->arch(), t->regs());
 }
 
 int read_elf_class(const string& filename) {
@@ -669,8 +724,10 @@ string real_path(const string& path) {
   return path;
 }
 
-string exe_directory() {
-  string exe_path = real_path("/proc/self/exe");
+static string read_exe_dir() {
+  KernelMapping km =
+      AddressSpace::read_local_kernel_mapping((uint8_t*)&read_exe_dir);
+  string exe_path = km.fsname();
   int end = exe_path.length();
   // Chop off the filename
   while (end > 0 && exe_path[end - 1] != '/') {
@@ -679,3 +736,69 @@ string exe_directory() {
   exe_path.erase(end);
   return exe_path;
 }
+
+string exe_directory() {
+  static string exe_path = read_exe_dir();
+  return exe_path;
+}
+
+/**
+ * Get the current time from the preferred monotonic clock in units of
+ * seconds, relative to an unspecific point in the past.
+ */
+double monotonic_now_sec() {
+  struct timespec tp;
+  clock_gettime(CLOCK_MONOTONIC, &tp);
+  return (double)tp.tv_sec + (double)tp.tv_nsec / 1e9;
+}
+
+bool running_under_rr() { return getenv("RUNNING_UNDER_RR") != NULL; }
+
+vector<string> read_proc_status_fields(pid_t tid, const char* name,
+                                       const char* name2) {
+  vector<string> result;
+  char buf[1000];
+  sprintf(buf, "/proc/%d/status", tid);
+  FILE* f = fopen(buf, "r");
+  if (!f) {
+    return result;
+  }
+  vector<string> matches;
+  matches.push_back(string(name) + ":");
+  if (name2) {
+    matches.push_back(string(name2) + ":");
+  }
+  for (auto& m : matches) {
+    while (true) {
+      if (!fgets(buf, sizeof(buf), f)) {
+        break;
+      }
+      if (strncmp(buf, m.c_str(), m.size()) == 0) {
+        char* b = buf + m.size();
+        while (*b == ' ' || *b == '\t') {
+          ++b;
+        }
+        char* e = b;
+        while (*e && *e != '\n') {
+          ++e;
+        }
+        result.push_back(string(b, e - b));
+        break;
+      }
+    }
+  }
+  fclose(f);
+  return result;
+}
+
+static bool check_for_pax_kernel() {
+  auto results = read_proc_status_fields(getpid(), "PaX");
+  return !results.empty();
+}
+
+bool uses_invisible_guard_page() {
+  static bool is_pax_kernel = check_for_pax_kernel();
+  return !is_pax_kernel;
+}
+
+} // namespace rr
