@@ -7,11 +7,14 @@
 #include <map>
 #include <set>
 #include <string>
+#include <vector>
 
 #include "Event.h"
-#include "remote_ptr.h"
 #include "ScopedFd.h"
 #include "TraceFrame.h"
+#include "remote_ptr.h"
+
+namespace rr {
 
 /*
  * This file is a dumping ground for functionality that needs to be shared but
@@ -25,12 +28,12 @@ class KernelMapping;
 class Task;
 class TraceFrame;
 
-template <typename T, size_t N> constexpr size_t array_length(T(&array)[N]) {
+template <typename T, size_t N> constexpr size_t array_length(T (&)[N]) {
   return N;
 }
 
 template <typename T, size_t N>
-constexpr size_t array_length(std::array<T, N>& array) {
+constexpr size_t array_length(std::array<T, N>&) {
   return N;
 }
 
@@ -58,15 +61,26 @@ template <typename T> bool type_has_no_holes() {
 
 #define SHMEM_FS "/dev/shm"
 
-/* The syscallbuf shared with tracees is created with this prefix
- * followed by the tracee tid, then immediately unlinked and shared
- * anonymously. */
-#define SYSCALLBUF_SHMEM_PATH_PREFIX "/tmp/rr-tracee-shmem-"
-
 #define PREFIX_FOR_EMPTY_MMAPED_REGIONS "/tmp/rr-emptyfile-"
 
 enum Completion { COMPLETE, INCOMPLETE };
 
+/**
+ * During recording, sometimes we need to ensure that an iteration of
+ * RecordSession::record_step schedules the same task as in the previous
+ * iteration. The PREVENT_SWITCH value indicates that this is required.
+ * For example, the futex operation FUTEX_WAKE_OP modifies userspace
+ * memory; those changes are only recorded after the system call completes;
+ * and they must be replayed before we allow a context switch to a woken-up
+ * task (because the kernel guarantees those effects are seen by woken-up
+ * tasks).
+ * Entering a potentially blocking system call must use ALLOW_SWITCH, or
+ * we risk deadlock. Most non-blocking system calls could use PREVENT_SWITCH
+ * or ALLOW_SWITCH; for simplicity we use ALLOW_SWITCH to indicate a call could
+ * block and PREVENT_SWITCH otherwise.
+ * Note that even if a system call uses PREVENT_SWITCH, as soon as we've
+ * recorded the completion of the system call, we can switch to another task.
+ */
 enum Switchable { PREVENT_SWITCH, ALLOW_SWITCH };
 
 /**
@@ -94,9 +108,9 @@ void format_dump_filename(Task* t, TraceFrame::Time global_time,
                           size_t filename_size);
 
 /**
- * Return true if the user requested memory be dumped for |t| at |f|.
+ * Return true if the user requested memory be dumped at |f|.
  */
-bool should_dump_memory(Task* t, const TraceFrame& f);
+bool should_dump_memory(const TraceFrame& f);
 /**
  * Dump all of the memory in |t|'s address to the file
  * "[trace_dir]/[t->tid]_[global_time]_[tag]".
@@ -108,7 +122,7 @@ void dump_process_memory(Task* t, TraceFrame::Time global_time,
  * Return true if the user has requested |t|'s memory be
  * checksummed at |f|.
  */
-bool should_checksum(Task* t, const TraceFrame& f);
+bool should_checksum(const TraceFrame& f);
 /**
  * Write a checksum of each mapped region in |t|'s address space to a
  * special log, where it can be read by |validate_process_memory()|
@@ -155,7 +169,7 @@ size_t page_size();
 enum signal_action { DUMP_CORE, TERMINATE, CONTINUE, STOP, IGNORE };
 signal_action default_action(int sig);
 
-SignalDeterministic is_deterministic_signal(const siginfo_t& si);
+SignalDeterministic is_deterministic_signal(Task* t);
 
 /**
  * Return nonzero if a mapping of |filename| with metadata |stat|,
@@ -187,22 +201,30 @@ enum cpuid_requests {
   CPUID_INTELBRANDSTRINGEND,
 };
 
+const int OSXSAVE_FEATURE_FLAG = 1 << 27;
+const int AVX_FEATURE_FLAG = 1 << 28;
+
 /** issue a single request to CPUID. Fits 'intel features', for instance
  *  note that even if only "eax" and "edx" are of interest, other registers
  *  will be modified by the operation, so we need to tell the compiler about it.
  *  'code' is placed in EAX. 'subrequest' is placed in ECX.
  *  *a, *c and *d receive EAX, ECX and EDX respectively.
  */
-void cpuid(int code, int subrequest, unsigned int* a, unsigned int* c,
-           unsigned int* d);
+struct CPUIDData {
+  unsigned int eax, ebx, ecx, edx;
+};
+CPUIDData cpuid(int code, int subrequest);
 
+struct CloneParameters {
+  remote_ptr<void> stack;
+  remote_ptr<int> ptid;
+  remote_ptr<void> tls;
+  remote_ptr<int> ctid;
+};
 /**
  * Extract various clone(2) parameters out of the given Task's registers.
- * Each remote_ptr parameter may be nullptr.
  */
-void extract_clone_parameters(Task* t, remote_ptr<void>* stack,
-                              remote_ptr<int>* ptid, remote_ptr<void>* tls,
-                              remote_ptr<int>* ctid);
+CloneParameters extract_clone_parameters(Task* t);
 
 /**
  * Read the ELF CLASS from the given filename. If it's unable to be read,
@@ -223,4 +245,23 @@ std::string real_path(const std::string& path);
 
 std::string exe_directory();
 
+/**
+ * Get the current time from the preferred monotonic clock in units of
+ * seconds, relative to an unspecific point in the past.
+ */
+double monotonic_now_sec();
+
+bool running_under_rr();
+
+std::vector<std::string> read_proc_status_fields(pid_t tid, const char* name,
+                                                 const char* name2 = nullptr);
+
+/**
+ * Mainline Linux kernels use an invisible (to /proc/<pid>/maps) guard page
+ * for stacks. grsecurity kernels don't.
+ */
+bool uses_invisible_guard_page();
+
+} // namespace rr
+
 #endif /* RR_UTIL_H_ */
